sqlglot.dialects.tsql API documentation

Edit on GitHub

# sqlglot.dialects.tsql

View Source

       1from __future__ import annotations
       2
       3import datetime
       4import re
       5import typing as t
       6
       7from sqlglot import exp, generator, parser, tokens, transforms
       8from sqlglot.dialects.dialect import (
       9    Dialect,
      10    NormalizationStrategy,
      11    any_value_to_max_sql,
      12    date_delta_sql,
      13    datestrtodate_sql,
      14    generatedasidentitycolumnconstraint_sql,
      15    max_or_greatest,
      16    min_or_least,
      17    build_date_delta,
      18    rename_func,
      19    timestrtotime_sql,
      20    trim_sql,
      21)
      22from sqlglot.helper import seq_get
      23from sqlglot.time import format_time
      24from sqlglot.tokens import TokenType
      25
      26if t.TYPE_CHECKING:
      27    from sqlglot._typing import E
      28
      29FULL_FORMAT_TIME_MAPPING = {
      30    "weekday": "%A",
      31    "dw": "%A",
      32    "w": "%A",
      33    "month": "%B",
      34    "mm": "%B",
      35    "m": "%B",
      36}
      37
      38DATE_DELTA_INTERVAL = {
      39    "year": "year",
      40    "yyyy": "year",
      41    "yy": "year",
      42    "quarter": "quarter",
      43    "qq": "quarter",
      44    "q": "quarter",
      45    "month": "month",
      46    "mm": "month",
      47    "m": "month",
      48    "week": "week",
      49    "ww": "week",
      50    "wk": "week",
      51    "day": "day",
      52    "dd": "day",
      53    "d": "day",
      54}
      55
      56
      57DATE_FMT_RE = re.compile("([dD]{1,2})|([mM]{1,2})|([yY]{1,4})|([hH]{1,2})|([sS]{1,2})")
      58
      59# N = Numeric, C=Currency
      60TRANSPILE_SAFE_NUMBER_FMT = {"N", "C"}
      61
      62DEFAULT_START_DATE = datetime.date(1900, 1, 1)
      63
      64BIT_TYPES = {exp.EQ, exp.NEQ, exp.Is, exp.In, exp.Select, exp.Alias}
      65
      66# Unsupported options:
      67# - OPTIMIZE FOR ( @variable_name { UNKNOWN | = <literal_constant> } [ , ...n ] )
      68# - TABLE HINT
      69OPTIONS: parser.OPTIONS_TYPE = {
      70    **dict.fromkeys(
      71        (
      72            "DISABLE_OPTIMIZED_PLAN_FORCING",
      73            "FAST",
      74            "IGNORE_NONCLUSTERED_COLUMNSTORE_INDEX",
      75            "LABEL",
      76            "MAXDOP",
      77            "MAXRECURSION",
      78            "MAX_GRANT_PERCENT",
      79            "MIN_GRANT_PERCENT",
      80            "NO_PERFORMANCE_SPOOL",
      81            "QUERYTRACEON",
      82            "RECOMPILE",
      83        ),
      84        tuple(),
      85    ),
      86    "CONCAT": ("UNION",),
      87    "DISABLE": ("EXTERNALPUSHDOWN", "SCALEOUTEXECUTION"),
      88    "EXPAND": ("VIEWS",),
      89    "FORCE": ("EXTERNALPUSHDOWN", "ORDER", "SCALEOUTEXECUTION"),
      90    "HASH": ("GROUP", "JOIN", "UNION"),
      91    "KEEP": ("PLAN",),
      92    "KEEPFIXED": ("PLAN",),
      93    "LOOP": ("JOIN",),
      94    "MERGE": ("JOIN", "UNION"),
      95    "OPTIMIZE": (("FOR", "UNKNOWN"),),
      96    "ORDER": ("GROUP",),
      97    "PARAMETERIZATION": ("FORCED", "SIMPLE"),
      98    "ROBUST": ("PLAN",),
      99    "USE": ("PLAN",),
     100}
     101
     102OPTIONS_THAT_REQUIRE_EQUAL = ("MAX_GRANT_PERCENT", "MIN_GRANT_PERCENT", "LABEL")
     103
     104
     105def _build_formatted_time(
     106    exp_class: t.Type[E], full_format_mapping: t.Optional[bool] = None
     107) -> t.Callable[[t.List], E]:
     108    def _builder(args: t.List) -> E:
     109        assert len(args) == 2
     110
     111        return exp_class(
     112            this=exp.cast(args[1], "datetime"),
     113            format=exp.Literal.string(
     114                format_time(
     115                    args[0].name.lower(),
     116                    (
     117                        {**TSQL.TIME_MAPPING, **FULL_FORMAT_TIME_MAPPING}
     118                        if full_format_mapping
     119                        else TSQL.TIME_MAPPING
     120                    ),
     121                )
     122            ),
     123        )
     124
     125    return _builder
     126
     127
     128def _build_format(args: t.List) -> exp.NumberToStr | exp.TimeToStr:
     129    this = seq_get(args, 0)
     130    fmt = seq_get(args, 1)
     131    culture = seq_get(args, 2)
     132
     133    number_fmt = fmt and (fmt.name in TRANSPILE_SAFE_NUMBER_FMT or not DATE_FMT_RE.search(fmt.name))
     134
     135    if number_fmt:
     136        return exp.NumberToStr(this=this, format=fmt, culture=culture)
     137
     138    if fmt:
     139        fmt = exp.Literal.string(
     140            format_time(fmt.name, TSQL.FORMAT_TIME_MAPPING)
     141            if len(fmt.name) == 1
     142            else format_time(fmt.name, TSQL.TIME_MAPPING)
     143        )
     144
     145    return exp.TimeToStr(this=this, format=fmt, culture=culture)
     146
     147
     148def _build_eomonth(args: t.List) -> exp.LastDay:
     149    date = exp.TsOrDsToDate(this=seq_get(args, 0))
     150    month_lag = seq_get(args, 1)
     151
     152    if month_lag is None:
     153        this: exp.Expression = date
     154    else:
     155        unit = DATE_DELTA_INTERVAL.get("month")
     156        this = exp.DateAdd(this=date, expression=month_lag, unit=unit and exp.var(unit))
     157
     158    return exp.LastDay(this=this)
     159
     160
     161def _build_hashbytes(args: t.List) -> exp.Expression:
     162    kind, data = args
     163    kind = kind.name.upper() if kind.is_string else ""
     164
     165    if kind == "MD5":
     166        args.pop(0)
     167        return exp.MD5(this=data)
     168    if kind in ("SHA", "SHA1"):
     169        args.pop(0)
     170        return exp.SHA(this=data)
     171    if kind == "SHA2_256":
     172        return exp.SHA2(this=data, length=exp.Literal.number(256))
     173    if kind == "SHA2_512":
     174        return exp.SHA2(this=data, length=exp.Literal.number(512))
     175
     176    return exp.func("HASHBYTES", *args)
     177
     178
     179DATEPART_ONLY_FORMATS = {"DW", "HOUR", "QUARTER"}
     180
     181
     182def _format_sql(self: TSQL.Generator, expression: exp.NumberToStr | exp.TimeToStr) -> str:
     183    fmt = expression.args["format"]
     184
     185    if not isinstance(expression, exp.NumberToStr):
     186        if fmt.is_string:
     187            mapped_fmt = format_time(fmt.name, TSQL.INVERSE_TIME_MAPPING)
     188
     189            name = (mapped_fmt or "").upper()
     190            if name in DATEPART_ONLY_FORMATS:
     191                return self.func("DATEPART", name, expression.this)
     192
     193            fmt_sql = self.sql(exp.Literal.string(mapped_fmt))
     194        else:
     195            fmt_sql = self.format_time(expression) or self.sql(fmt)
     196    else:
     197        fmt_sql = self.sql(fmt)
     198
     199    return self.func("FORMAT", expression.this, fmt_sql, expression.args.get("culture"))
     200
     201
     202def _string_agg_sql(self: TSQL.Generator, expression: exp.GroupConcat) -> str:
     203    this = expression.this
     204    distinct = expression.find(exp.Distinct)
     205    if distinct:
     206        # exp.Distinct can appear below an exp.Order or an exp.GroupConcat expression
     207        self.unsupported("T-SQL STRING_AGG doesn't support DISTINCT.")
     208        this = distinct.pop().expressions[0]
     209
     210    order = ""
     211    if isinstance(expression.this, exp.Order):
     212        if expression.this.this:
     213            this = expression.this.this.pop()
     214        order = f" WITHIN GROUP ({self.sql(expression.this)[1:]})"  # Order has a leading space
     215
     216    separator = expression.args.get("separator") or exp.Literal.string(",")
     217    return f"STRING_AGG({self.format_args(this, separator)}){order}"
     218
     219
     220def _build_date_delta(
     221    exp_class: t.Type[E], unit_mapping: t.Optional[t.Dict[str, str]] = None
     222) -> t.Callable[[t.List], E]:
     223    def _builder(args: t.List) -> E:
     224        unit = seq_get(args, 0)
     225        if unit and unit_mapping:
     226            unit = exp.var(unit_mapping.get(unit.name.lower(), unit.name))
     227
     228        start_date = seq_get(args, 1)
     229        if start_date and start_date.is_number:
     230            # Numeric types are valid DATETIME values
     231            if start_date.is_int:
     232                adds = DEFAULT_START_DATE + datetime.timedelta(days=int(start_date.this))
     233                start_date = exp.Literal.string(adds.strftime("%F"))
     234            else:
     235                # We currently don't handle float values, i.e. they're not converted to equivalent DATETIMEs.
     236                # This is not a problem when generating T-SQL code, it is when transpiling to other dialects.
     237                return exp_class(this=seq_get(args, 2), expression=start_date, unit=unit)
     238
     239        return exp_class(
     240            this=exp.TimeStrToTime(this=seq_get(args, 2)),
     241            expression=exp.TimeStrToTime(this=start_date),
     242            unit=unit,
     243        )
     244
     245    return _builder
     246
     247
     248def qualify_derived_table_outputs(expression: exp.Expression) -> exp.Expression:
     249    """Ensures all (unnamed) output columns are aliased for CTEs and Subqueries."""
     250    alias = expression.args.get("alias")
     251
     252    if (
     253        isinstance(expression, (exp.CTE, exp.Subquery))
     254        and isinstance(alias, exp.TableAlias)
     255        and not alias.columns
     256    ):
     257        from sqlglot.optimizer.qualify_columns import qualify_outputs
     258
     259        # We keep track of the unaliased column projection indexes instead of the expressions
     260        # themselves, because the latter are going to be replaced by new nodes when the aliases
     261        # are added and hence we won't be able to reach these newly added Alias parents
     262        query = expression.this
     263        unaliased_column_indexes = (
     264            i for i, c in enumerate(query.selects) if isinstance(c, exp.Column) and not c.alias
     265        )
     266
     267        qualify_outputs(query)
     268
     269        # Preserve the quoting information of columns for newly added Alias nodes
     270        query_selects = query.selects
     271        for select_index in unaliased_column_indexes:
     272            alias = query_selects[select_index]
     273            column = alias.this
     274            if isinstance(column.this, exp.Identifier):
     275                alias.args["alias"].set("quoted", column.this.quoted)
     276
     277    return expression
     278
     279
     280# https://learn.microsoft.com/en-us/sql/t-sql/functions/datetimefromparts-transact-sql?view=sql-server-ver16#syntax
     281def _build_datetimefromparts(args: t.List) -> exp.TimestampFromParts:
     282    return exp.TimestampFromParts(
     283        year=seq_get(args, 0),
     284        month=seq_get(args, 1),
     285        day=seq_get(args, 2),
     286        hour=seq_get(args, 3),
     287        min=seq_get(args, 4),
     288        sec=seq_get(args, 5),
     289        milli=seq_get(args, 6),
     290    )
     291
     292
     293# https://learn.microsoft.com/en-us/sql/t-sql/functions/timefromparts-transact-sql?view=sql-server-ver16#syntax
     294def _build_timefromparts(args: t.List) -> exp.TimeFromParts:
     295    return exp.TimeFromParts(
     296        hour=seq_get(args, 0),
     297        min=seq_get(args, 1),
     298        sec=seq_get(args, 2),
     299        fractions=seq_get(args, 3),
     300        precision=seq_get(args, 4),
     301    )
     302
     303
     304def _build_with_arg_as_text(
     305    klass: t.Type[exp.Expression],
     306) -> t.Callable[[t.List[exp.Expression]], exp.Expression]:
     307    def _parse(args: t.List[exp.Expression]) -> exp.Expression:
     308        this = seq_get(args, 0)
     309
     310        if this and not this.is_string:
     311            this = exp.cast(this, exp.DataType.Type.TEXT)
     312
     313        expression = seq_get(args, 1)
     314        kwargs = {"this": this}
     315
     316        if expression:
     317            kwargs["expression"] = expression
     318
     319        return klass(**kwargs)
     320
     321    return _parse
     322
     323
     324def _json_extract_sql(
     325    self: TSQL.Generator, expression: exp.JSONExtract | exp.JSONExtractScalar
     326) -> str:
     327    json_query = self.func("JSON_QUERY", expression.this, expression.expression)
     328    json_value = self.func("JSON_VALUE", expression.this, expression.expression)
     329    return self.func("ISNULL", json_query, json_value)
     330
     331
     332class TSQL(Dialect):
     333    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE
     334    TIME_FORMAT = "'yyyy-mm-dd hh:mm:ss'"
     335    SUPPORTS_SEMI_ANTI_JOIN = False
     336    LOG_BASE_FIRST = False
     337    TYPED_DIVISION = True
     338    CONCAT_COALESCE = True
     339
     340    TIME_MAPPING = {
     341        "year": "%Y",
     342        "dayofyear": "%j",
     343        "day": "%d",
     344        "dy": "%d",
     345        "y": "%Y",
     346        "week": "%W",
     347        "ww": "%W",
     348        "wk": "%W",
     349        "hour": "%h",
     350        "hh": "%I",
     351        "minute": "%M",
     352        "mi": "%M",
     353        "n": "%M",
     354        "second": "%S",
     355        "ss": "%S",
     356        "s": "%-S",
     357        "millisecond": "%f",
     358        "ms": "%f",
     359        "weekday": "%W",
     360        "dw": "%W",
     361        "month": "%m",
     362        "mm": "%M",
     363        "m": "%-M",
     364        "Y": "%Y",
     365        "YYYY": "%Y",
     366        "YY": "%y",
     367        "MMMM": "%B",
     368        "MMM": "%b",
     369        "MM": "%m",
     370        "M": "%-m",
     371        "dddd": "%A",
     372        "dd": "%d",
     373        "d": "%-d",
     374        "HH": "%H",
     375        "H": "%-H",
     376        "h": "%-I",
     377        "S": "%f",
     378        "yyyy": "%Y",
     379        "yy": "%y",
     380    }
     381
     382    CONVERT_FORMAT_MAPPING = {
     383        "0": "%b %d %Y %-I:%M%p",
     384        "1": "%m/%d/%y",
     385        "2": "%y.%m.%d",
     386        "3": "%d/%m/%y",
     387        "4": "%d.%m.%y",
     388        "5": "%d-%m-%y",
     389        "6": "%d %b %y",
     390        "7": "%b %d, %y",
     391        "8": "%H:%M:%S",
     392        "9": "%b %d %Y %-I:%M:%S:%f%p",
     393        "10": "mm-dd-yy",
     394        "11": "yy/mm/dd",
     395        "12": "yymmdd",
     396        "13": "%d %b %Y %H:%M:ss:%f",
     397        "14": "%H:%M:%S:%f",
     398        "20": "%Y-%m-%d %H:%M:%S",
     399        "21": "%Y-%m-%d %H:%M:%S.%f",
     400        "22": "%m/%d/%y %-I:%M:%S %p",
     401        "23": "%Y-%m-%d",
     402        "24": "%H:%M:%S",
     403        "25": "%Y-%m-%d %H:%M:%S.%f",
     404        "100": "%b %d %Y %-I:%M%p",
     405        "101": "%m/%d/%Y",
     406        "102": "%Y.%m.%d",
     407        "103": "%d/%m/%Y",
     408        "104": "%d.%m.%Y",
     409        "105": "%d-%m-%Y",
     410        "106": "%d %b %Y",
     411        "107": "%b %d, %Y",
     412        "108": "%H:%M:%S",
     413        "109": "%b %d %Y %-I:%M:%S:%f%p",
     414        "110": "%m-%d-%Y",
     415        "111": "%Y/%m/%d",
     416        "112": "%Y%m%d",
     417        "113": "%d %b %Y %H:%M:%S:%f",
     418        "114": "%H:%M:%S:%f",
     419        "120": "%Y-%m-%d %H:%M:%S",
     420        "121": "%Y-%m-%d %H:%M:%S.%f",
     421    }
     422
     423    FORMAT_TIME_MAPPING = {
     424        "y": "%B %Y",
     425        "d": "%m/%d/%Y",
     426        "H": "%-H",
     427        "h": "%-I",
     428        "s": "%Y-%m-%d %H:%M:%S",
     429        "D": "%A,%B,%Y",
     430        "f": "%A,%B,%Y %-I:%M %p",
     431        "F": "%A,%B,%Y %-I:%M:%S %p",
     432        "g": "%m/%d/%Y %-I:%M %p",
     433        "G": "%m/%d/%Y %-I:%M:%S %p",
     434        "M": "%B %-d",
     435        "m": "%B %-d",
     436        "O": "%Y-%m-%dT%H:%M:%S",
     437        "u": "%Y-%M-%D %H:%M:%S%z",
     438        "U": "%A, %B %D, %Y %H:%M:%S%z",
     439        "T": "%-I:%M:%S %p",
     440        "t": "%-I:%M",
     441        "Y": "%a %Y",
     442    }
     443
     444    class Tokenizer(tokens.Tokenizer):
     445        IDENTIFIERS = [("[", "]"), '"']
     446        QUOTES = ["'", '"']
     447        HEX_STRINGS = [("0x", ""), ("0X", "")]
     448        VAR_SINGLE_TOKENS = {"@", "$", "#"}
     449
     450        KEYWORDS = {
     451            **tokens.Tokenizer.KEYWORDS,
     452            "DATETIME2": TokenType.DATETIME,
     453            "DATETIMEOFFSET": TokenType.TIMESTAMPTZ,
     454            "DECLARE": TokenType.COMMAND,
     455            "EXEC": TokenType.COMMAND,
     456            "IMAGE": TokenType.IMAGE,
     457            "MONEY": TokenType.MONEY,
     458            "NTEXT": TokenType.TEXT,
     459            "PRINT": TokenType.COMMAND,
     460            "PROC": TokenType.PROCEDURE,
     461            "REAL": TokenType.FLOAT,
     462            "ROWVERSION": TokenType.ROWVERSION,
     463            "SMALLDATETIME": TokenType.DATETIME,
     464            "SMALLMONEY": TokenType.SMALLMONEY,
     465            "SQL_VARIANT": TokenType.VARIANT,
     466            "TOP": TokenType.TOP,
     467            "UNIQUEIDENTIFIER": TokenType.UNIQUEIDENTIFIER,
     468            "UPDATE STATISTICS": TokenType.COMMAND,
     469            "XML": TokenType.XML,
     470            "OUTPUT": TokenType.RETURNING,
     471            "SYSTEM_USER": TokenType.CURRENT_USER,
     472            "FOR SYSTEM_TIME": TokenType.TIMESTAMP_SNAPSHOT,
     473            "OPTION": TokenType.OPTION,
     474        }
     475
     476    class Parser(parser.Parser):
     477        SET_REQUIRES_ASSIGNMENT_DELIMITER = False
     478        LOG_DEFAULTS_TO_LN = True
     479        ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN = False
     480        STRING_ALIASES = True
     481        NO_PAREN_IF_COMMANDS = False
     482
     483        QUERY_MODIFIER_PARSERS = {
     484            **parser.Parser.QUERY_MODIFIER_PARSERS,
     485            TokenType.OPTION: lambda self: ("options", self._parse_options()),
     486        }
     487
     488        FUNCTIONS = {
     489            **parser.Parser.FUNCTIONS,
     490            "CHARINDEX": lambda args: exp.StrPosition(
     491                this=seq_get(args, 1),
     492                substr=seq_get(args, 0),
     493                position=seq_get(args, 2),
     494            ),
     495            "DATEADD": build_date_delta(exp.DateAdd, unit_mapping=DATE_DELTA_INTERVAL),
     496            "DATEDIFF": _build_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
     497            "DATENAME": _build_formatted_time(exp.TimeToStr, full_format_mapping=True),
     498            "DATEPART": _build_formatted_time(exp.TimeToStr),
     499            "DATETIMEFROMPARTS": _build_datetimefromparts,
     500            "EOMONTH": _build_eomonth,
     501            "FORMAT": _build_format,
     502            "GETDATE": exp.CurrentTimestamp.from_arg_list,
     503            "HASHBYTES": _build_hashbytes,
     504            "ISNULL": exp.Coalesce.from_arg_list,
     505            "JSON_QUERY": parser.build_extract_json_with_path(exp.JSONExtract),
     506            "JSON_VALUE": parser.build_extract_json_with_path(exp.JSONExtractScalar),
     507            "LEN": _build_with_arg_as_text(exp.Length),
     508            "LEFT": _build_with_arg_as_text(exp.Left),
     509            "RIGHT": _build_with_arg_as_text(exp.Right),
     510            "REPLICATE": exp.Repeat.from_arg_list,
     511            "SQUARE": lambda args: exp.Pow(this=seq_get(args, 0), expression=exp.Literal.number(2)),
     512            "SYSDATETIME": exp.CurrentTimestamp.from_arg_list,
     513            "SUSER_NAME": exp.CurrentUser.from_arg_list,
     514            "SUSER_SNAME": exp.CurrentUser.from_arg_list,
     515            "SYSTEM_USER": exp.CurrentUser.from_arg_list,
     516            "TIMEFROMPARTS": _build_timefromparts,
     517        }
     518
     519        JOIN_HINTS = {"LOOP", "HASH", "MERGE", "REMOTE"}
     520
     521        RETURNS_TABLE_TOKENS = parser.Parser.ID_VAR_TOKENS - {
     522            TokenType.TABLE,
     523            *parser.Parser.TYPE_TOKENS,
     524        }
     525
     526        STATEMENT_PARSERS = {
     527            **parser.Parser.STATEMENT_PARSERS,
     528            TokenType.END: lambda self: self._parse_command(),
     529        }
     530
     531        def _parse_options(self) -> t.Optional[t.List[exp.Expression]]:
     532            if not self._match(TokenType.OPTION):
     533                return None
     534
     535            def _parse_option() -> t.Optional[exp.Expression]:
     536                option = self._parse_var_from_options(OPTIONS)
     537                if not option:
     538                    return None
     539
     540                self._match(TokenType.EQ)
     541                return self.expression(
     542                    exp.QueryOption, this=option, expression=self._parse_primary_or_var()
     543                )
     544
     545            return self._parse_wrapped_csv(_parse_option)
     546
     547        def _parse_projections(self) -> t.List[exp.Expression]:
     548            """
     549            T-SQL supports the syntax alias = expression in the SELECT's projection list,
     550            so we transform all parsed Selects to convert their EQ projections into Aliases.
     551
     552            See: https://learn.microsoft.com/en-us/sql/t-sql/queries/select-clause-transact-sql?view=sql-server-ver16#syntax
     553            """
     554            return [
     555                (
     556                    exp.alias_(projection.expression, projection.this.this, copy=False)
     557                    if isinstance(projection, exp.EQ) and isinstance(projection.this, exp.Column)
     558                    else projection
     559                )
     560                for projection in super()._parse_projections()
     561            ]
     562
     563        def _parse_commit_or_rollback(self) -> exp.Commit | exp.Rollback:
     564            """Applies to SQL Server and Azure SQL Database
     565            COMMIT [ { TRAN | TRANSACTION }
     566                [ transaction_name | @tran_name_variable ] ]
     567                [ WITH ( DELAYED_DURABILITY = { OFF | ON } ) ]
     568
     569            ROLLBACK { TRAN | TRANSACTION }
     570                [ transaction_name | @tran_name_variable
     571                | savepoint_name | @savepoint_variable ]
     572            """
     573            rollback = self._prev.token_type == TokenType.ROLLBACK
     574
     575            self._match_texts(("TRAN", "TRANSACTION"))
     576            this = self._parse_id_var()
     577
     578            if rollback:
     579                return self.expression(exp.Rollback, this=this)
     580
     581            durability = None
     582            if self._match_pair(TokenType.WITH, TokenType.L_PAREN):
     583                self._match_text_seq("DELAYED_DURABILITY")
     584                self._match(TokenType.EQ)
     585
     586                if self._match_text_seq("OFF"):
     587                    durability = False
     588                else:
     589                    self._match(TokenType.ON)
     590                    durability = True
     591
     592                self._match_r_paren()
     593
     594            return self.expression(exp.Commit, this=this, durability=durability)
     595
     596        def _parse_transaction(self) -> exp.Transaction | exp.Command:
     597            """Applies to SQL Server and Azure SQL Database
     598            BEGIN { TRAN | TRANSACTION }
     599            [ { transaction_name | @tran_name_variable }
     600            [ WITH MARK [ 'description' ] ]
     601            ]
     602            """
     603            if self._match_texts(("TRAN", "TRANSACTION")):
     604                transaction = self.expression(exp.Transaction, this=self._parse_id_var())
     605                if self._match_text_seq("WITH", "MARK"):
     606                    transaction.set("mark", self._parse_string())
     607
     608                return transaction
     609
     610            return self._parse_as_command(self._prev)
     611
     612        def _parse_returns(self) -> exp.ReturnsProperty:
     613            table = self._parse_id_var(any_token=False, tokens=self.RETURNS_TABLE_TOKENS)
     614            returns = super()._parse_returns()
     615            returns.set("table", table)
     616            return returns
     617
     618        def _parse_convert(
     619            self, strict: bool, safe: t.Optional[bool] = None
     620        ) -> t.Optional[exp.Expression]:
     621            this = self._parse_types()
     622            self._match(TokenType.COMMA)
     623            args = [this, *self._parse_csv(self._parse_conjunction)]
     624            convert = exp.Convert.from_arg_list(args)
     625            convert.set("safe", safe)
     626            convert.set("strict", strict)
     627            return convert
     628
     629        def _parse_user_defined_function(
     630            self, kind: t.Optional[TokenType] = None
     631        ) -> t.Optional[exp.Expression]:
     632            this = super()._parse_user_defined_function(kind=kind)
     633
     634            if (
     635                kind == TokenType.FUNCTION
     636                or isinstance(this, exp.UserDefinedFunction)
     637                or self._match(TokenType.ALIAS, advance=False)
     638            ):
     639                return this
     640
     641            expressions = self._parse_csv(self._parse_function_parameter)
     642            return self.expression(exp.UserDefinedFunction, this=this, expressions=expressions)
     643
     644        def _parse_id_var(
     645            self,
     646            any_token: bool = True,
     647            tokens: t.Optional[t.Collection[TokenType]] = None,
     648        ) -> t.Optional[exp.Expression]:
     649            is_temporary = self._match(TokenType.HASH)
     650            is_global = is_temporary and self._match(TokenType.HASH)
     651
     652            this = super()._parse_id_var(any_token=any_token, tokens=tokens)
     653            if this:
     654                if is_global:
     655                    this.set("global", True)
     656                elif is_temporary:
     657                    this.set("temporary", True)
     658
     659            return this
     660
     661        def _parse_create(self) -> exp.Create | exp.Command:
     662            create = super()._parse_create()
     663
     664            if isinstance(create, exp.Create):
     665                table = create.this.this if isinstance(create.this, exp.Schema) else create.this
     666                if isinstance(table, exp.Table) and table.this.args.get("temporary"):
     667                    if not create.args.get("properties"):
     668                        create.set("properties", exp.Properties(expressions=[]))
     669
     670                    create.args["properties"].append("expressions", exp.TemporaryProperty())
     671
     672            return create
     673
     674        def _parse_if(self) -> t.Optional[exp.Expression]:
     675            index = self._index
     676
     677            if self._match_text_seq("OBJECT_ID"):
     678                self._parse_wrapped_csv(self._parse_string)
     679                if self._match_text_seq("IS", "NOT", "NULL") and self._match(TokenType.DROP):
     680                    return self._parse_drop(exists=True)
     681                self._retreat(index)
     682
     683            return super()._parse_if()
     684
     685        def _parse_unique(self) -> exp.UniqueColumnConstraint:
     686            if self._match_texts(("CLUSTERED", "NONCLUSTERED")):
     687                this = self.CONSTRAINT_PARSERS[self._prev.text.upper()](self)
     688            else:
     689                this = self._parse_schema(self._parse_id_var(any_token=False))
     690
     691            return self.expression(exp.UniqueColumnConstraint, this=this)
     692
     693        def _parse_partition(self) -> t.Optional[exp.Partition]:
     694            if not self._match_text_seq("WITH", "(", "PARTITIONS"):
     695                return None
     696
     697            def parse_range():
     698                low = self._parse_bitwise()
     699                high = self._parse_bitwise() if self._match_text_seq("TO") else None
     700
     701                return (
     702                    self.expression(exp.PartitionRange, this=low, expression=high) if high else low
     703                )
     704
     705            partition = self.expression(
     706                exp.Partition, expressions=self._parse_wrapped_csv(parse_range)
     707            )
     708
     709            self._match_r_paren()
     710
     711            return partition
     712
     713    class Generator(generator.Generator):
     714        LIMIT_IS_TOP = True
     715        QUERY_HINTS = False
     716        RETURNING_END = False
     717        NVL2_SUPPORTED = False
     718        ALTER_TABLE_INCLUDE_COLUMN_KEYWORD = False
     719        LIMIT_FETCH = "FETCH"
     720        COMPUTED_COLUMN_WITH_TYPE = False
     721        CTE_RECURSIVE_KEYWORD_REQUIRED = False
     722        ENSURE_BOOLS = True
     723        NULL_ORDERING_SUPPORTED = None
     724        SUPPORTS_SINGLE_ARG_CONCAT = False
     725        TABLESAMPLE_SEED_KEYWORD = "REPEATABLE"
     726        SUPPORTS_SELECT_INTO = True
     727        JSON_PATH_BRACKETED_KEY_SUPPORTED = False
     728        SUPPORTS_TO_NUMBER = False
     729
     730        EXPRESSIONS_WITHOUT_NESTED_CTES = {
     731            exp.Delete,
     732            exp.Insert,
     733            exp.Merge,
     734            exp.Select,
     735            exp.Subquery,
     736            exp.Union,
     737            exp.Update,
     738        }
     739
     740        SUPPORTED_JSON_PATH_PARTS = {
     741            exp.JSONPathKey,
     742            exp.JSONPathRoot,
     743            exp.JSONPathSubscript,
     744        }
     745
     746        TYPE_MAPPING = {
     747            **generator.Generator.TYPE_MAPPING,
     748            exp.DataType.Type.BOOLEAN: "BIT",
     749            exp.DataType.Type.DECIMAL: "NUMERIC",
     750            exp.DataType.Type.DATETIME: "DATETIME2",
     751            exp.DataType.Type.DOUBLE: "FLOAT",
     752            exp.DataType.Type.INT: "INTEGER",
     753            exp.DataType.Type.TEXT: "VARCHAR(MAX)",
     754            exp.DataType.Type.TIMESTAMP: "DATETIME2",
     755            exp.DataType.Type.TIMESTAMPTZ: "DATETIMEOFFSET",
     756            exp.DataType.Type.VARIANT: "SQL_VARIANT",
     757        }
     758
     759        TYPE_MAPPING.pop(exp.DataType.Type.NCHAR)
     760        TYPE_MAPPING.pop(exp.DataType.Type.NVARCHAR)
     761
     762        TRANSFORMS = {
     763            **generator.Generator.TRANSFORMS,
     764            exp.AnyValue: any_value_to_max_sql,
     765            exp.ArrayToString: rename_func("STRING_AGG"),
     766            exp.AutoIncrementColumnConstraint: lambda *_: "IDENTITY",
     767            exp.DateAdd: date_delta_sql("DATEADD"),
     768            exp.DateDiff: date_delta_sql("DATEDIFF"),
     769            exp.CTE: transforms.preprocess([qualify_derived_table_outputs]),
     770            exp.CurrentDate: rename_func("GETDATE"),
     771            exp.CurrentTimestamp: rename_func("GETDATE"),
     772            exp.DateStrToDate: datestrtodate_sql,
     773            exp.Extract: rename_func("DATEPART"),
     774            exp.GeneratedAsIdentityColumnConstraint: generatedasidentitycolumnconstraint_sql,
     775            exp.GroupConcat: _string_agg_sql,
     776            exp.If: rename_func("IIF"),
     777            exp.JSONExtract: _json_extract_sql,
     778            exp.JSONExtractScalar: _json_extract_sql,
     779            exp.LastDay: lambda self, e: self.func("EOMONTH", e.this),
     780            exp.Max: max_or_greatest,
     781            exp.MD5: lambda self, e: self.func("HASHBYTES", exp.Literal.string("MD5"), e.this),
     782            exp.Min: min_or_least,
     783            exp.NumberToStr: _format_sql,
     784            exp.ParseJSON: lambda self, e: self.sql(e, "this"),
     785            exp.Select: transforms.preprocess(
     786                [
     787                    transforms.eliminate_distinct_on,
     788                    transforms.eliminate_semi_and_anti_joins,
     789                    transforms.eliminate_qualify,
     790                ]
     791            ),
     792            exp.StrPosition: lambda self, e: self.func(
     793                "CHARINDEX", e.args.get("substr"), e.this, e.args.get("position")
     794            ),
     795            exp.Subquery: transforms.preprocess([qualify_derived_table_outputs]),
     796            exp.SHA: lambda self, e: self.func("HASHBYTES", exp.Literal.string("SHA1"), e.this),
     797            exp.SHA2: lambda self, e: self.func(
     798                "HASHBYTES", exp.Literal.string(f"SHA2_{e.args.get('length', 256)}"), e.this
     799            ),
     800            exp.TemporaryProperty: lambda self, e: "",
     801            exp.TimeStrToTime: timestrtotime_sql,
     802            exp.TimeToStr: _format_sql,
     803            exp.Trim: trim_sql,
     804            exp.TsOrDsAdd: date_delta_sql("DATEADD", cast=True),
     805            exp.TsOrDsDiff: date_delta_sql("DATEDIFF"),
     806        }
     807
     808        TRANSFORMS.pop(exp.ReturnsProperty)
     809
     810        PROPERTIES_LOCATION = {
     811            **generator.Generator.PROPERTIES_LOCATION,
     812            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
     813        }
     814
     815        def select_sql(self, expression: exp.Select) -> str:
     816            if expression.args.get("offset"):
     817                if not expression.args.get("order"):
     818                    # ORDER BY is required in order to use OFFSET in a query, so we use
     819                    # a noop order by, since we don't really care about the order.
     820                    # See: https://www.microsoftpressstore.com/articles/article.aspx?p=2314819
     821                    expression.order_by(exp.select(exp.null()).subquery(), copy=False)
     822
     823                limit = expression.args.get("limit")
     824                if isinstance(limit, exp.Limit):
     825                    # TOP and OFFSET can't be combined, we need use FETCH instead of TOP
     826                    # we replace here because otherwise TOP would be generated in select_sql
     827                    limit.replace(exp.Fetch(direction="FIRST", count=limit.expression))
     828
     829            return super().select_sql(expression)
     830
     831        def convert_sql(self, expression: exp.Convert) -> str:
     832            name = "TRY_CONVERT" if expression.args.get("safe") else "CONVERT"
     833            return self.func(
     834                name, expression.this, expression.expression, expression.args.get("style")
     835            )
     836
     837        def queryoption_sql(self, expression: exp.QueryOption) -> str:
     838            option = self.sql(expression, "this")
     839            value = self.sql(expression, "expression")
     840            if value:
     841                optional_equal_sign = "= " if option in OPTIONS_THAT_REQUIRE_EQUAL else ""
     842                return f"{option} {optional_equal_sign}{value}"
     843            return option
     844
     845        def lateral_op(self, expression: exp.Lateral) -> str:
     846            cross_apply = expression.args.get("cross_apply")
     847            if cross_apply is True:
     848                return "CROSS APPLY"
     849            if cross_apply is False:
     850                return "OUTER APPLY"
     851
     852            # TODO: perhaps we can check if the parent is a Join and transpile it appropriately
     853            self.unsupported("LATERAL clause is not supported.")
     854            return "LATERAL"
     855
     856        def timefromparts_sql(self, expression: exp.TimeFromParts) -> str:
     857            nano = expression.args.get("nano")
     858            if nano is not None:
     859                nano.pop()
     860                self.unsupported("Specifying nanoseconds is not supported in TIMEFROMPARTS.")
     861
     862            if expression.args.get("fractions") is None:
     863                expression.set("fractions", exp.Literal.number(0))
     864            if expression.args.get("precision") is None:
     865                expression.set("precision", exp.Literal.number(0))
     866
     867            return rename_func("TIMEFROMPARTS")(self, expression)
     868
     869        def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
     870            zone = expression.args.get("zone")
     871            if zone is not None:
     872                zone.pop()
     873                self.unsupported("Time zone is not supported in DATETIMEFROMPARTS.")
     874
     875            nano = expression.args.get("nano")
     876            if nano is not None:
     877                nano.pop()
     878                self.unsupported("Specifying nanoseconds is not supported in DATETIMEFROMPARTS.")
     879
     880            if expression.args.get("milli") is None:
     881                expression.set("milli", exp.Literal.number(0))
     882
     883            return rename_func("DATETIMEFROMPARTS")(self, expression)
     884
     885        def set_operations(self, expression: exp.Union) -> str:
     886            limit = expression.args.get("limit")
     887            if limit:
     888                return self.sql(expression.limit(limit.pop(), copy=False))
     889
     890            return super().set_operations(expression)
     891
     892        def setitem_sql(self, expression: exp.SetItem) -> str:
     893            this = expression.this
     894            if isinstance(this, exp.EQ) and not isinstance(this.left, exp.Parameter):
     895                # T-SQL does not use '=' in SET command, except when the LHS is a variable.
     896                return f"{self.sql(this.left)} {self.sql(this.right)}"
     897
     898            return super().setitem_sql(expression)
     899
     900        def boolean_sql(self, expression: exp.Boolean) -> str:
     901            if type(expression.parent) in BIT_TYPES:
     902                return "1" if expression.this else "0"
     903
     904            return "(1 = 1)" if expression.this else "(1 = 0)"
     905
     906        def is_sql(self, expression: exp.Is) -> str:
     907            if isinstance(expression.expression, exp.Boolean):
     908                return self.binary(expression, "=")
     909            return self.binary(expression, "IS")
     910
     911        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
     912            sql = self.sql(expression, "this")
     913            properties = expression.args.get("properties")
     914
     915            if sql[:1] != "#" and any(
     916                isinstance(prop, exp.TemporaryProperty)
     917                for prop in (properties.expressions if properties else [])
     918            ):
     919                sql = f"#{sql}"
     920
     921            return sql
     922
     923        def create_sql(self, expression: exp.Create) -> str:
     924            kind = expression.kind
     925            exists = expression.args.pop("exists", None)
     926            sql = super().create_sql(expression)
     927
     928            like_property = expression.find(exp.LikeProperty)
     929            if like_property:
     930                ctas_expression = like_property.this
     931            else:
     932                ctas_expression = expression.expression
     933
     934            table = expression.find(exp.Table)
     935
     936            # Convert CTAS statement to SELECT .. INTO ..
     937            if kind == "TABLE" and ctas_expression:
     938                ctas_with = ctas_expression.args.get("with")
     939                if ctas_with:
     940                    ctas_with = ctas_with.pop()
     941
     942                if isinstance(ctas_expression, exp.UNWRAPPED_QUERIES):
     943                    ctas_expression = ctas_expression.subquery()
     944
     945                select_into = exp.select("*").from_(exp.alias_(ctas_expression, "temp", table=True))
     946                select_into.set("into", exp.Into(this=table))
     947                select_into.set("with", ctas_with)
     948
     949                if like_property:
     950                    select_into.limit(0, copy=False)
     951
     952                sql = self.sql(select_into)
     953
     954            if exists:
     955                identifier = self.sql(exp.Literal.string(exp.table_name(table) if table else ""))
     956                sql = self.sql(exp.Literal.string(sql))
     957                if kind == "SCHEMA":
     958                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.schemata WHERE schema_name = {identifier}) EXEC({sql})"""
     959                elif kind == "TABLE":
     960                    assert table
     961                    where = exp.and_(
     962                        exp.column("table_name").eq(table.name),
     963                        exp.column("table_schema").eq(table.db) if table.db else None,
     964                        exp.column("table_catalog").eq(table.catalog) if table.catalog else None,
     965                    )
     966                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.tables WHERE {where}) EXEC({sql})"""
     967                elif kind == "INDEX":
     968                    index = self.sql(exp.Literal.string(expression.this.text("this")))
     969                    sql = f"""IF NOT EXISTS (SELECT * FROM sys.indexes WHERE object_id = object_id({identifier}) AND name = {index}) EXEC({sql})"""
     970            elif expression.args.get("replace"):
     971                sql = sql.replace("CREATE OR REPLACE ", "CREATE OR ALTER ", 1)
     972
     973            return self.prepend_ctes(expression, sql)
     974
     975        def offset_sql(self, expression: exp.Offset) -> str:
     976            return f"{super().offset_sql(expression)} ROWS"
     977
     978        def version_sql(self, expression: exp.Version) -> str:
     979            name = "SYSTEM_TIME" if expression.name == "TIMESTAMP" else expression.name
     980            this = f"FOR {name}"
     981            expr = expression.expression
     982            kind = expression.text("kind")
     983            if kind in ("FROM", "BETWEEN"):
     984                args = expr.expressions
     985                sep = "TO" if kind == "FROM" else "AND"
     986                expr_sql = f"{self.sql(seq_get(args, 0))} {sep} {self.sql(seq_get(args, 1))}"
     987            else:
     988                expr_sql = self.sql(expr)
     989
     990            expr_sql = f" {expr_sql}" if expr_sql else ""
     991            return f"{this} {kind}{expr_sql}"
     992
     993        def returnsproperty_sql(self, expression: exp.ReturnsProperty) -> str:
     994            table = expression.args.get("table")
     995            table = f"{table} " if table else ""
     996            return f"RETURNS {table}{self.sql(expression, 'this')}"
     997
     998        def returning_sql(self, expression: exp.Returning) -> str:
     999            into = self.sql(expression, "into")
    1000            into = self.seg(f"INTO {into}") if into else ""
    1001            return f"{self.seg('OUTPUT')} {self.expressions(expression, flat=True)}{into}"
    1002
    1003        def transaction_sql(self, expression: exp.Transaction) -> str:
    1004            this = self.sql(expression, "this")
    1005            this = f" {this}" if this else ""
    1006            mark = self.sql(expression, "mark")
    1007            mark = f" WITH MARK {mark}" if mark else ""
    1008            return f"BEGIN TRANSACTION{this}{mark}"
    1009
    1010        def commit_sql(self, expression: exp.Commit) -> str:
    1011            this = self.sql(expression, "this")
    1012            this = f" {this}" if this else ""
    1013            durability = expression.args.get("durability")
    1014            durability = (
    1015                f" WITH (DELAYED_DURABILITY = {'ON' if durability else 'OFF'})"
    1016                if durability is not None
    1017                else ""
    1018            )
    1019            return f"COMMIT TRANSACTION{this}{durability}"
    1020
    1021        def rollback_sql(self, expression: exp.Rollback) -> str:
    1022            this = self.sql(expression, "this")
    1023            this = f" {this}" if this else ""
    1024            return f"ROLLBACK TRANSACTION{this}"
    1025
    1026        def identifier_sql(self, expression: exp.Identifier) -> str:
    1027            identifier = super().identifier_sql(expression)
    1028
    1029            if expression.args.get("global"):
    1030                identifier = f"##{identifier}"
    1031            elif expression.args.get("temporary"):
    1032                identifier = f"#{identifier}"
    1033
    1034            return identifier
    1035
    1036        def constraint_sql(self, expression: exp.Constraint) -> str:
    1037            this = self.sql(expression, "this")
    1038            expressions = self.expressions(expression, flat=True, sep=" ")
    1039            return f"CONSTRAINT {this} {expressions}"
    1040
    1041        def length_sql(self, expression: exp.Length) -> str:
    1042            return self._uncast_text(expression, "LEN")
    1043
    1044        def right_sql(self, expression: exp.Right) -> str:
    1045            return self._uncast_text(expression, "RIGHT")
    1046
    1047        def left_sql(self, expression: exp.Left) -> str:
    1048            return self._uncast_text(expression, "LEFT")
    1049
    1050        def _uncast_text(self, expression: exp.Expression, name: str) -> str:
    1051            this = expression.this
    1052            if isinstance(this, exp.Cast) and this.is_type(exp.DataType.Type.TEXT):
    1053                this_sql = self.sql(this, "this")
    1054            else:
    1055                this_sql = self.sql(this)
    1056            expression_sql = self.sql(expression, "expression")
    1057            return self.func(name, this_sql, expression_sql if expression_sql else None)
    1058
    1059        def partition_sql(self, expression: exp.Partition) -> str:
    1060            return f"WITH (PARTITIONS({self.expressions(expression, flat=True)}))"

FULL_FORMAT_TIME_MAPPING = {'weekday': '%A', 'dw': '%A', 'w': '%A', 'month': '%B', 'mm': '%B', 'm': '%B'}
DATE_DELTA_INTERVAL = {'year': 'year', 'yyyy': 'year', 'yy': 'year', 'quarter': 'quarter', 'qq': 'quarter', 'q': 'quarter', 'month': 'month', 'mm': 'month', 'm': 'month', 'week': 'week', 'ww': 'week', 'wk': 'week', 'day': 'day', 'dd': 'day', 'd': 'day'}
DATE_FMT_RE = re.compile('([dD]{1,2})|([mM]{1,2})|([yY]{1,4})|([hH]{1,2})|([sS]{1,2})')
TRANSPILE_SAFE_NUMBER_FMT = {'C', 'N'}
DEFAULT_START_DATE = datetime.date(1900, 1, 1)
BIT_TYPES = {<class 'sqlglot.expressions.In'>, <class 'sqlglot.expressions.Is'>, <class 'sqlglot.expressions.Alias'>, <class 'sqlglot.expressions.Select'>, <class 'sqlglot.expressions.EQ'>, <class 'sqlglot.expressions.NEQ'>}
OPTIONS: Dict[str, Sequence[Union[Sequence[str], str]]] = {'DISABLE_OPTIMIZED_PLAN_FORCING': (), 'FAST': (), 'IGNORE_NONCLUSTERED_COLUMNSTORE_INDEX': (), 'LABEL': (), 'MAXDOP': (), 'MAXRECURSION': (), 'MAX_GRANT_PERCENT': (), 'MIN_GRANT_PERCENT': (), 'NO_PERFORMANCE_SPOOL': (), 'QUERYTRACEON': (), 'RECOMPILE': (), 'CONCAT': ('UNION',), 'DISABLE': ('EXTERNALPUSHDOWN', 'SCALEOUTEXECUTION'), 'EXPAND': ('VIEWS',), 'FORCE': ('EXTERNALPUSHDOWN', 'ORDER', 'SCALEOUTEXECUTION'), 'HASH': ('GROUP', 'JOIN', 'UNION'), 'KEEP': ('PLAN',), 'KEEPFIXED': ('PLAN',), 'LOOP': ('JOIN',), 'MERGE': ('JOIN', 'UNION'), 'OPTIMIZE': (('FOR', 'UNKNOWN'),), 'ORDER': ('GROUP',), 'PARAMETERIZATION': ('FORCED', 'SIMPLE'), 'ROBUST': ('PLAN',), 'USE': ('PLAN',)}
OPTIONS_THAT_REQUIRE_EQUAL = ('MAX_GRANT_PERCENT', 'MIN_GRANT_PERCENT', 'LABEL')
DATEPART_ONLY_FORMATS = {'HOUR', 'QUARTER', 'DW'}
def qualify_derived_table_outputs( expression: sqlglot.expressions.Expression) -> sqlglot.expressions.Expression: View Source

    249def qualify_derived_table_outputs(expression: exp.Expression) -> exp.Expression:
    250    """Ensures all (unnamed) output columns are aliased for CTEs and Subqueries."""
    251    alias = expression.args.get("alias")
    252
    253    if (
    254        isinstance(expression, (exp.CTE, exp.Subquery))
    255        and isinstance(alias, exp.TableAlias)
    256        and not alias.columns
    257    ):
    258        from sqlglot.optimizer.qualify_columns import qualify_outputs
    259
    260        # We keep track of the unaliased column projection indexes instead of the expressions
    261        # themselves, because the latter are going to be replaced by new nodes when the aliases
    262        # are added and hence we won't be able to reach these newly added Alias parents
    263        query = expression.this
    264        unaliased_column_indexes = (
    265            i for i, c in enumerate(query.selects) if isinstance(c, exp.Column) and not c.alias
    266        )
    267
    268        qualify_outputs(query)
    269
    270        # Preserve the quoting information of columns for newly added Alias nodes
    271        query_selects = query.selects
    272        for select_index in unaliased_column_indexes:
    273            alias = query_selects[select_index]
    274            column = alias.this
    275            if isinstance(column.this, exp.Identifier):
    276                alias.args["alias"].set("quoted", column.this.quoted)
    277
    278    return expression

Ensures all (unnamed) output columns are aliased for CTEs and Subqueries.

class TSQL(sqlglot.dialects.dialect.Dialect): View Source

     333class TSQL(Dialect):
     334    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE
     335    TIME_FORMAT = "'yyyy-mm-dd hh:mm:ss'"
     336    SUPPORTS_SEMI_ANTI_JOIN = False
     337    LOG_BASE_FIRST = False
     338    TYPED_DIVISION = True
     339    CONCAT_COALESCE = True
     340
     341    TIME_MAPPING = {
     342        "year": "%Y",
     343        "dayofyear": "%j",
     344        "day": "%d",
     345        "dy": "%d",
     346        "y": "%Y",
     347        "week": "%W",
     348        "ww": "%W",
     349        "wk": "%W",
     350        "hour": "%h",
     351        "hh": "%I",
     352        "minute": "%M",
     353        "mi": "%M",
     354        "n": "%M",
     355        "second": "%S",
     356        "ss": "%S",
     357        "s": "%-S",
     358        "millisecond": "%f",
     359        "ms": "%f",
     360        "weekday": "%W",
     361        "dw": "%W",
     362        "month": "%m",
     363        "mm": "%M",
     364        "m": "%-M",
     365        "Y": "%Y",
     366        "YYYY": "%Y",
     367        "YY": "%y",
     368        "MMMM": "%B",
     369        "MMM": "%b",
     370        "MM": "%m",
     371        "M": "%-m",
     372        "dddd": "%A",
     373        "dd": "%d",
     374        "d": "%-d",
     375        "HH": "%H",
     376        "H": "%-H",
     377        "h": "%-I",
     378        "S": "%f",
     379        "yyyy": "%Y",
     380        "yy": "%y",
     381    }
     382
     383    CONVERT_FORMAT_MAPPING = {
     384        "0": "%b %d %Y %-I:%M%p",
     385        "1": "%m/%d/%y",
     386        "2": "%y.%m.%d",
     387        "3": "%d/%m/%y",
     388        "4": "%d.%m.%y",
     389        "5": "%d-%m-%y",
     390        "6": "%d %b %y",
     391        "7": "%b %d, %y",
     392        "8": "%H:%M:%S",
     393        "9": "%b %d %Y %-I:%M:%S:%f%p",
     394        "10": "mm-dd-yy",
     395        "11": "yy/mm/dd",
     396        "12": "yymmdd",
     397        "13": "%d %b %Y %H:%M:ss:%f",
     398        "14": "%H:%M:%S:%f",
     399        "20": "%Y-%m-%d %H:%M:%S",
     400        "21": "%Y-%m-%d %H:%M:%S.%f",
     401        "22": "%m/%d/%y %-I:%M:%S %p",
     402        "23": "%Y-%m-%d",
     403        "24": "%H:%M:%S",
     404        "25": "%Y-%m-%d %H:%M:%S.%f",
     405        "100": "%b %d %Y %-I:%M%p",
     406        "101": "%m/%d/%Y",
     407        "102": "%Y.%m.%d",
     408        "103": "%d/%m/%Y",
     409        "104": "%d.%m.%Y",
     410        "105": "%d-%m-%Y",
     411        "106": "%d %b %Y",
     412        "107": "%b %d, %Y",
     413        "108": "%H:%M:%S",
     414        "109": "%b %d %Y %-I:%M:%S:%f%p",
     415        "110": "%m-%d-%Y",
     416        "111": "%Y/%m/%d",
     417        "112": "%Y%m%d",
     418        "113": "%d %b %Y %H:%M:%S:%f",
     419        "114": "%H:%M:%S:%f",
     420        "120": "%Y-%m-%d %H:%M:%S",
     421        "121": "%Y-%m-%d %H:%M:%S.%f",
     422    }
     423
     424    FORMAT_TIME_MAPPING = {
     425        "y": "%B %Y",
     426        "d": "%m/%d/%Y",
     427        "H": "%-H",
     428        "h": "%-I",
     429        "s": "%Y-%m-%d %H:%M:%S",
     430        "D": "%A,%B,%Y",
     431        "f": "%A,%B,%Y %-I:%M %p",
     432        "F": "%A,%B,%Y %-I:%M:%S %p",
     433        "g": "%m/%d/%Y %-I:%M %p",
     434        "G": "%m/%d/%Y %-I:%M:%S %p",
     435        "M": "%B %-d",
     436        "m": "%B %-d",
     437        "O": "%Y-%m-%dT%H:%M:%S",
     438        "u": "%Y-%M-%D %H:%M:%S%z",
     439        "U": "%A, %B %D, %Y %H:%M:%S%z",
     440        "T": "%-I:%M:%S %p",
     441        "t": "%-I:%M",
     442        "Y": "%a %Y",
     443    }
     444
     445    class Tokenizer(tokens.Tokenizer):
     446        IDENTIFIERS = [("[", "]"), '"']
     447        QUOTES = ["'", '"']
     448        HEX_STRINGS = [("0x", ""), ("0X", "")]
     449        VAR_SINGLE_TOKENS = {"@", "$", "#"}
     450
     451        KEYWORDS = {
     452            **tokens.Tokenizer.KEYWORDS,
     453            "DATETIME2": TokenType.DATETIME,
     454            "DATETIMEOFFSET": TokenType.TIMESTAMPTZ,
     455            "DECLARE": TokenType.COMMAND,
     456            "EXEC": TokenType.COMMAND,
     457            "IMAGE": TokenType.IMAGE,
     458            "MONEY": TokenType.MONEY,
     459            "NTEXT": TokenType.TEXT,
     460            "PRINT": TokenType.COMMAND,
     461            "PROC": TokenType.PROCEDURE,
     462            "REAL": TokenType.FLOAT,
     463            "ROWVERSION": TokenType.ROWVERSION,
     464            "SMALLDATETIME": TokenType.DATETIME,
     465            "SMALLMONEY": TokenType.SMALLMONEY,
     466            "SQL_VARIANT": TokenType.VARIANT,
     467            "TOP": TokenType.TOP,
     468            "UNIQUEIDENTIFIER": TokenType.UNIQUEIDENTIFIER,
     469            "UPDATE STATISTICS": TokenType.COMMAND,
     470            "XML": TokenType.XML,
     471            "OUTPUT": TokenType.RETURNING,
     472            "SYSTEM_USER": TokenType.CURRENT_USER,
     473            "FOR SYSTEM_TIME": TokenType.TIMESTAMP_SNAPSHOT,
     474            "OPTION": TokenType.OPTION,
     475        }
     476
     477    class Parser(parser.Parser):
     478        SET_REQUIRES_ASSIGNMENT_DELIMITER = False
     479        LOG_DEFAULTS_TO_LN = True
     480        ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN = False
     481        STRING_ALIASES = True
     482        NO_PAREN_IF_COMMANDS = False
     483
     484        QUERY_MODIFIER_PARSERS = {
     485            **parser.Parser.QUERY_MODIFIER_PARSERS,
     486            TokenType.OPTION: lambda self: ("options", self._parse_options()),
     487        }
     488
     489        FUNCTIONS = {
     490            **parser.Parser.FUNCTIONS,
     491            "CHARINDEX": lambda args: exp.StrPosition(
     492                this=seq_get(args, 1),
     493                substr=seq_get(args, 0),
     494                position=seq_get(args, 2),
     495            ),
     496            "DATEADD": build_date_delta(exp.DateAdd, unit_mapping=DATE_DELTA_INTERVAL),
     497            "DATEDIFF": _build_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
     498            "DATENAME": _build_formatted_time(exp.TimeToStr, full_format_mapping=True),
     499            "DATEPART": _build_formatted_time(exp.TimeToStr),
     500            "DATETIMEFROMPARTS": _build_datetimefromparts,
     501            "EOMONTH": _build_eomonth,
     502            "FORMAT": _build_format,
     503            "GETDATE": exp.CurrentTimestamp.from_arg_list,
     504            "HASHBYTES": _build_hashbytes,
     505            "ISNULL": exp.Coalesce.from_arg_list,
     506            "JSON_QUERY": parser.build_extract_json_with_path(exp.JSONExtract),
     507            "JSON_VALUE": parser.build_extract_json_with_path(exp.JSONExtractScalar),
     508            "LEN": _build_with_arg_as_text(exp.Length),
     509            "LEFT": _build_with_arg_as_text(exp.Left),
     510            "RIGHT": _build_with_arg_as_text(exp.Right),
     511            "REPLICATE": exp.Repeat.from_arg_list,
     512            "SQUARE": lambda args: exp.Pow(this=seq_get(args, 0), expression=exp.Literal.number(2)),
     513            "SYSDATETIME": exp.CurrentTimestamp.from_arg_list,
     514            "SUSER_NAME": exp.CurrentUser.from_arg_list,
     515            "SUSER_SNAME": exp.CurrentUser.from_arg_list,
     516            "SYSTEM_USER": exp.CurrentUser.from_arg_list,
     517            "TIMEFROMPARTS": _build_timefromparts,
     518        }
     519
     520        JOIN_HINTS = {"LOOP", "HASH", "MERGE", "REMOTE"}
     521
     522        RETURNS_TABLE_TOKENS = parser.Parser.ID_VAR_TOKENS - {
     523            TokenType.TABLE,
     524            *parser.Parser.TYPE_TOKENS,
     525        }
     526
     527        STATEMENT_PARSERS = {
     528            **parser.Parser.STATEMENT_PARSERS,
     529            TokenType.END: lambda self: self._parse_command(),
     530        }
     531
     532        def _parse_options(self) -> t.Optional[t.List[exp.Expression]]:
     533            if not self._match(TokenType.OPTION):
     534                return None
     535
     536            def _parse_option() -> t.Optional[exp.Expression]:
     537                option = self._parse_var_from_options(OPTIONS)
     538                if not option:
     539                    return None
     540
     541                self._match(TokenType.EQ)
     542                return self.expression(
     543                    exp.QueryOption, this=option, expression=self._parse_primary_or_var()
     544                )
     545
     546            return self._parse_wrapped_csv(_parse_option)
     547
     548        def _parse_projections(self) -> t.List[exp.Expression]:
     549            """
     550            T-SQL supports the syntax alias = expression in the SELECT's projection list,
     551            so we transform all parsed Selects to convert their EQ projections into Aliases.
     552
     553            See: https://learn.microsoft.com/en-us/sql/t-sql/queries/select-clause-transact-sql?view=sql-server-ver16#syntax
     554            """
     555            return [
     556                (
     557                    exp.alias_(projection.expression, projection.this.this, copy=False)
     558                    if isinstance(projection, exp.EQ) and isinstance(projection.this, exp.Column)
     559                    else projection
     560                )
     561                for projection in super()._parse_projections()
     562            ]
     563
     564        def _parse_commit_or_rollback(self) -> exp.Commit | exp.Rollback:
     565            """Applies to SQL Server and Azure SQL Database
     566            COMMIT [ { TRAN | TRANSACTION }
     567                [ transaction_name | @tran_name_variable ] ]
     568                [ WITH ( DELAYED_DURABILITY = { OFF | ON } ) ]
     569
     570            ROLLBACK { TRAN | TRANSACTION }
     571                [ transaction_name | @tran_name_variable
     572                | savepoint_name | @savepoint_variable ]
     573            """
     574            rollback = self._prev.token_type == TokenType.ROLLBACK
     575
     576            self._match_texts(("TRAN", "TRANSACTION"))
     577            this = self._parse_id_var()
     578
     579            if rollback:
     580                return self.expression(exp.Rollback, this=this)
     581
     582            durability = None
     583            if self._match_pair(TokenType.WITH, TokenType.L_PAREN):
     584                self._match_text_seq("DELAYED_DURABILITY")
     585                self._match(TokenType.EQ)
     586
     587                if self._match_text_seq("OFF"):
     588                    durability = False
     589                else:
     590                    self._match(TokenType.ON)
     591                    durability = True
     592
     593                self._match_r_paren()
     594
     595            return self.expression(exp.Commit, this=this, durability=durability)
     596
     597        def _parse_transaction(self) -> exp.Transaction | exp.Command:
     598            """Applies to SQL Server and Azure SQL Database
     599            BEGIN { TRAN | TRANSACTION }
     600            [ { transaction_name | @tran_name_variable }
     601            [ WITH MARK [ 'description' ] ]
     602            ]
     603            """
     604            if self._match_texts(("TRAN", "TRANSACTION")):
     605                transaction = self.expression(exp.Transaction, this=self._parse_id_var())
     606                if self._match_text_seq("WITH", "MARK"):
     607                    transaction.set("mark", self._parse_string())
     608
     609                return transaction
     610
     611            return self._parse_as_command(self._prev)
     612
     613        def _parse_returns(self) -> exp.ReturnsProperty:
     614            table = self._parse_id_var(any_token=False, tokens=self.RETURNS_TABLE_TOKENS)
     615            returns = super()._parse_returns()
     616            returns.set("table", table)
     617            return returns
     618
     619        def _parse_convert(
     620            self, strict: bool, safe: t.Optional[bool] = None
     621        ) -> t.Optional[exp.Expression]:
     622            this = self._parse_types()
     623            self._match(TokenType.COMMA)
     624            args = [this, *self._parse_csv(self._parse_conjunction)]
     625            convert = exp.Convert.from_arg_list(args)
     626            convert.set("safe", safe)
     627            convert.set("strict", strict)
     628            return convert
     629
     630        def _parse_user_defined_function(
     631            self, kind: t.Optional[TokenType] = None
     632        ) -> t.Optional[exp.Expression]:
     633            this = super()._parse_user_defined_function(kind=kind)
     634
     635            if (
     636                kind == TokenType.FUNCTION
     637                or isinstance(this, exp.UserDefinedFunction)
     638                or self._match(TokenType.ALIAS, advance=False)
     639            ):
     640                return this
     641
     642            expressions = self._parse_csv(self._parse_function_parameter)
     643            return self.expression(exp.UserDefinedFunction, this=this, expressions=expressions)
     644
     645        def _parse_id_var(
     646            self,
     647            any_token: bool = True,
     648            tokens: t.Optional[t.Collection[TokenType]] = None,
     649        ) -> t.Optional[exp.Expression]:
     650            is_temporary = self._match(TokenType.HASH)
     651            is_global = is_temporary and self._match(TokenType.HASH)
     652
     653            this = super()._parse_id_var(any_token=any_token, tokens=tokens)
     654            if this:
     655                if is_global:
     656                    this.set("global", True)
     657                elif is_temporary:
     658                    this.set("temporary", True)
     659
     660            return this
     661
     662        def _parse_create(self) -> exp.Create | exp.Command:
     663            create = super()._parse_create()
     664
     665            if isinstance(create, exp.Create):
     666                table = create.this.this if isinstance(create.this, exp.Schema) else create.this
     667                if isinstance(table, exp.Table) and table.this.args.get("temporary"):
     668                    if not create.args.get("properties"):
     669                        create.set("properties", exp.Properties(expressions=[]))
     670
     671                    create.args["properties"].append("expressions", exp.TemporaryProperty())
     672
     673            return create
     674
     675        def _parse_if(self) -> t.Optional[exp.Expression]:
     676            index = self._index
     677
     678            if self._match_text_seq("OBJECT_ID"):
     679                self._parse_wrapped_csv(self._parse_string)
     680                if self._match_text_seq("IS", "NOT", "NULL") and self._match(TokenType.DROP):
     681                    return self._parse_drop(exists=True)
     682                self._retreat(index)
     683
     684            return super()._parse_if()
     685
     686        def _parse_unique(self) -> exp.UniqueColumnConstraint:
     687            if self._match_texts(("CLUSTERED", "NONCLUSTERED")):
     688                this = self.CONSTRAINT_PARSERS[self._prev.text.upper()](self)
     689            else:
     690                this = self._parse_schema(self._parse_id_var(any_token=False))
     691
     692            return self.expression(exp.UniqueColumnConstraint, this=this)
     693
     694        def _parse_partition(self) -> t.Optional[exp.Partition]:
     695            if not self._match_text_seq("WITH", "(", "PARTITIONS"):
     696                return None
     697
     698            def parse_range():
     699                low = self._parse_bitwise()
     700                high = self._parse_bitwise() if self._match_text_seq("TO") else None
     701
     702                return (
     703                    self.expression(exp.PartitionRange, this=low, expression=high) if high else low
     704                )
     705
     706            partition = self.expression(
     707                exp.Partition, expressions=self._parse_wrapped_csv(parse_range)
     708            )
     709
     710            self._match_r_paren()
     711
     712            return partition
     713
     714    class Generator(generator.Generator):
     715        LIMIT_IS_TOP = True
     716        QUERY_HINTS = False
     717        RETURNING_END = False
     718        NVL2_SUPPORTED = False
     719        ALTER_TABLE_INCLUDE_COLUMN_KEYWORD = False
     720        LIMIT_FETCH = "FETCH"
     721        COMPUTED_COLUMN_WITH_TYPE = False
     722        CTE_RECURSIVE_KEYWORD_REQUIRED = False
     723        ENSURE_BOOLS = True
     724        NULL_ORDERING_SUPPORTED = None
     725        SUPPORTS_SINGLE_ARG_CONCAT = False
     726        TABLESAMPLE_SEED_KEYWORD = "REPEATABLE"
     727        SUPPORTS_SELECT_INTO = True
     728        JSON_PATH_BRACKETED_KEY_SUPPORTED = False
     729        SUPPORTS_TO_NUMBER = False
     730
     731        EXPRESSIONS_WITHOUT_NESTED_CTES = {
     732            exp.Delete,
     733            exp.Insert,
     734            exp.Merge,
     735            exp.Select,
     736            exp.Subquery,
     737            exp.Union,
     738            exp.Update,
     739        }
     740
     741        SUPPORTED_JSON_PATH_PARTS = {
     742            exp.JSONPathKey,
     743            exp.JSONPathRoot,
     744            exp.JSONPathSubscript,
     745        }
     746
     747        TYPE_MAPPING = {
     748            **generator.Generator.TYPE_MAPPING,
     749            exp.DataType.Type.BOOLEAN: "BIT",
     750            exp.DataType.Type.DECIMAL: "NUMERIC",
     751            exp.DataType.Type.DATETIME: "DATETIME2",
     752            exp.DataType.Type.DOUBLE: "FLOAT",
     753            exp.DataType.Type.INT: "INTEGER",
     754            exp.DataType.Type.TEXT: "VARCHAR(MAX)",
     755            exp.DataType.Type.TIMESTAMP: "DATETIME2",
     756            exp.DataType.Type.TIMESTAMPTZ: "DATETIMEOFFSET",
     757            exp.DataType.Type.VARIANT: "SQL_VARIANT",
     758        }
     759
     760        TYPE_MAPPING.pop(exp.DataType.Type.NCHAR)
     761        TYPE_MAPPING.pop(exp.DataType.Type.NVARCHAR)
     762
     763        TRANSFORMS = {
     764            **generator.Generator.TRANSFORMS,
     765            exp.AnyValue: any_value_to_max_sql,
     766            exp.ArrayToString: rename_func("STRING_AGG"),
     767            exp.AutoIncrementColumnConstraint: lambda *_: "IDENTITY",
     768            exp.DateAdd: date_delta_sql("DATEADD"),
     769            exp.DateDiff: date_delta_sql("DATEDIFF"),
     770            exp.CTE: transforms.preprocess([qualify_derived_table_outputs]),
     771            exp.CurrentDate: rename_func("GETDATE"),
     772            exp.CurrentTimestamp: rename_func("GETDATE"),
     773            exp.DateStrToDate: datestrtodate_sql,
     774            exp.Extract: rename_func("DATEPART"),
     775            exp.GeneratedAsIdentityColumnConstraint: generatedasidentitycolumnconstraint_sql,
     776            exp.GroupConcat: _string_agg_sql,
     777            exp.If: rename_func("IIF"),
     778            exp.JSONExtract: _json_extract_sql,
     779            exp.JSONExtractScalar: _json_extract_sql,
     780            exp.LastDay: lambda self, e: self.func("EOMONTH", e.this),
     781            exp.Max: max_or_greatest,
     782            exp.MD5: lambda self, e: self.func("HASHBYTES", exp.Literal.string("MD5"), e.this),
     783            exp.Min: min_or_least,
     784            exp.NumberToStr: _format_sql,
     785            exp.ParseJSON: lambda self, e: self.sql(e, "this"),
     786            exp.Select: transforms.preprocess(
     787                [
     788                    transforms.eliminate_distinct_on,
     789                    transforms.eliminate_semi_and_anti_joins,
     790                    transforms.eliminate_qualify,
     791                ]
     792            ),
     793            exp.StrPosition: lambda self, e: self.func(
     794                "CHARINDEX", e.args.get("substr"), e.this, e.args.get("position")
     795            ),
     796            exp.Subquery: transforms.preprocess([qualify_derived_table_outputs]),
     797            exp.SHA: lambda self, e: self.func("HASHBYTES", exp.Literal.string("SHA1"), e.this),
     798            exp.SHA2: lambda self, e: self.func(
     799                "HASHBYTES", exp.Literal.string(f"SHA2_{e.args.get('length', 256)}"), e.this
     800            ),
     801            exp.TemporaryProperty: lambda self, e: "",
     802            exp.TimeStrToTime: timestrtotime_sql,
     803            exp.TimeToStr: _format_sql,
     804            exp.Trim: trim_sql,
     805            exp.TsOrDsAdd: date_delta_sql("DATEADD", cast=True),
     806            exp.TsOrDsDiff: date_delta_sql("DATEDIFF"),
     807        }
     808
     809        TRANSFORMS.pop(exp.ReturnsProperty)
     810
     811        PROPERTIES_LOCATION = {
     812            **generator.Generator.PROPERTIES_LOCATION,
     813            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
     814        }
     815
     816        def select_sql(self, expression: exp.Select) -> str:
     817            if expression.args.get("offset"):
     818                if not expression.args.get("order"):
     819                    # ORDER BY is required in order to use OFFSET in a query, so we use
     820                    # a noop order by, since we don't really care about the order.
     821                    # See: https://www.microsoftpressstore.com/articles/article.aspx?p=2314819
     822                    expression.order_by(exp.select(exp.null()).subquery(), copy=False)
     823
     824                limit = expression.args.get("limit")
     825                if isinstance(limit, exp.Limit):
     826                    # TOP and OFFSET can't be combined, we need use FETCH instead of TOP
     827                    # we replace here because otherwise TOP would be generated in select_sql
     828                    limit.replace(exp.Fetch(direction="FIRST", count=limit.expression))
     829
     830            return super().select_sql(expression)
     831
     832        def convert_sql(self, expression: exp.Convert) -> str:
     833            name = "TRY_CONVERT" if expression.args.get("safe") else "CONVERT"
     834            return self.func(
     835                name, expression.this, expression.expression, expression.args.get("style")
     836            )
     837
     838        def queryoption_sql(self, expression: exp.QueryOption) -> str:
     839            option = self.sql(expression, "this")
     840            value = self.sql(expression, "expression")
     841            if value:
     842                optional_equal_sign = "= " if option in OPTIONS_THAT_REQUIRE_EQUAL else ""
     843                return f"{option} {optional_equal_sign}{value}"
     844            return option
     845
     846        def lateral_op(self, expression: exp.Lateral) -> str:
     847            cross_apply = expression.args.get("cross_apply")
     848            if cross_apply is True:
     849                return "CROSS APPLY"
     850            if cross_apply is False:
     851                return "OUTER APPLY"
     852
     853            # TODO: perhaps we can check if the parent is a Join and transpile it appropriately
     854            self.unsupported("LATERAL clause is not supported.")
     855            return "LATERAL"
     856
     857        def timefromparts_sql(self, expression: exp.TimeFromParts) -> str:
     858            nano = expression.args.get("nano")
     859            if nano is not None:
     860                nano.pop()
     861                self.unsupported("Specifying nanoseconds is not supported in TIMEFROMPARTS.")
     862
     863            if expression.args.get("fractions") is None:
     864                expression.set("fractions", exp.Literal.number(0))
     865            if expression.args.get("precision") is None:
     866                expression.set("precision", exp.Literal.number(0))
     867
     868            return rename_func("TIMEFROMPARTS")(self, expression)
     869
     870        def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
     871            zone = expression.args.get("zone")
     872            if zone is not None:
     873                zone.pop()
     874                self.unsupported("Time zone is not supported in DATETIMEFROMPARTS.")
     875
     876            nano = expression.args.get("nano")
     877            if nano is not None:
     878                nano.pop()
     879                self.unsupported("Specifying nanoseconds is not supported in DATETIMEFROMPARTS.")
     880
     881            if expression.args.get("milli") is None:
     882                expression.set("milli", exp.Literal.number(0))
     883
     884            return rename_func("DATETIMEFROMPARTS")(self, expression)
     885
     886        def set_operations(self, expression: exp.Union) -> str:
     887            limit = expression.args.get("limit")
     888            if limit:
     889                return self.sql(expression.limit(limit.pop(), copy=False))
     890
     891            return super().set_operations(expression)
     892
     893        def setitem_sql(self, expression: exp.SetItem) -> str:
     894            this = expression.this
     895            if isinstance(this, exp.EQ) and not isinstance(this.left, exp.Parameter):
     896                # T-SQL does not use '=' in SET command, except when the LHS is a variable.
     897                return f"{self.sql(this.left)} {self.sql(this.right)}"
     898
     899            return super().setitem_sql(expression)
     900
     901        def boolean_sql(self, expression: exp.Boolean) -> str:
     902            if type(expression.parent) in BIT_TYPES:
     903                return "1" if expression.this else "0"
     904
     905            return "(1 = 1)" if expression.this else "(1 = 0)"
     906
     907        def is_sql(self, expression: exp.Is) -> str:
     908            if isinstance(expression.expression, exp.Boolean):
     909                return self.binary(expression, "=")
     910            return self.binary(expression, "IS")
     911
     912        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
     913            sql = self.sql(expression, "this")
     914            properties = expression.args.get("properties")
     915
     916            if sql[:1] != "#" and any(
     917                isinstance(prop, exp.TemporaryProperty)
     918                for prop in (properties.expressions if properties else [])
     919            ):
     920                sql = f"#{sql}"
     921
     922            return sql
     923
     924        def create_sql(self, expression: exp.Create) -> str:
     925            kind = expression.kind
     926            exists = expression.args.pop("exists", None)
     927            sql = super().create_sql(expression)
     928
     929            like_property = expression.find(exp.LikeProperty)
     930            if like_property:
     931                ctas_expression = like_property.this
     932            else:
     933                ctas_expression = expression.expression
     934
     935            table = expression.find(exp.Table)
     936
     937            # Convert CTAS statement to SELECT .. INTO ..
     938            if kind == "TABLE" and ctas_expression:
     939                ctas_with = ctas_expression.args.get("with")
     940                if ctas_with:
     941                    ctas_with = ctas_with.pop()
     942
     943                if isinstance(ctas_expression, exp.UNWRAPPED_QUERIES):
     944                    ctas_expression = ctas_expression.subquery()
     945
     946                select_into = exp.select("*").from_(exp.alias_(ctas_expression, "temp", table=True))
     947                select_into.set("into", exp.Into(this=table))
     948                select_into.set("with", ctas_with)
     949
     950                if like_property:
     951                    select_into.limit(0, copy=False)
     952
     953                sql = self.sql(select_into)
     954
     955            if exists:
     956                identifier = self.sql(exp.Literal.string(exp.table_name(table) if table else ""))
     957                sql = self.sql(exp.Literal.string(sql))
     958                if kind == "SCHEMA":
     959                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.schemata WHERE schema_name = {identifier}) EXEC({sql})"""
     960                elif kind == "TABLE":
     961                    assert table
     962                    where = exp.and_(
     963                        exp.column("table_name").eq(table.name),
     964                        exp.column("table_schema").eq(table.db) if table.db else None,
     965                        exp.column("table_catalog").eq(table.catalog) if table.catalog else None,
     966                    )
     967                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.tables WHERE {where}) EXEC({sql})"""
     968                elif kind == "INDEX":
     969                    index = self.sql(exp.Literal.string(expression.this.text("this")))
     970                    sql = f"""IF NOT EXISTS (SELECT * FROM sys.indexes WHERE object_id = object_id({identifier}) AND name = {index}) EXEC({sql})"""
     971            elif expression.args.get("replace"):
     972                sql = sql.replace("CREATE OR REPLACE ", "CREATE OR ALTER ", 1)
     973
     974            return self.prepend_ctes(expression, sql)
     975
     976        def offset_sql(self, expression: exp.Offset) -> str:
     977            return f"{super().offset_sql(expression)} ROWS"
     978
     979        def version_sql(self, expression: exp.Version) -> str:
     980            name = "SYSTEM_TIME" if expression.name == "TIMESTAMP" else expression.name
     981            this = f"FOR {name}"
     982            expr = expression.expression
     983            kind = expression.text("kind")
     984            if kind in ("FROM", "BETWEEN"):
     985                args = expr.expressions
     986                sep = "TO" if kind == "FROM" else "AND"
     987                expr_sql = f"{self.sql(seq_get(args, 0))} {sep} {self.sql(seq_get(args, 1))}"
     988            else:
     989                expr_sql = self.sql(expr)
     990
     991            expr_sql = f" {expr_sql}" if expr_sql else ""
     992            return f"{this} {kind}{expr_sql}"
     993
     994        def returnsproperty_sql(self, expression: exp.ReturnsProperty) -> str:
     995            table = expression.args.get("table")
     996            table = f"{table} " if table else ""
     997            return f"RETURNS {table}{self.sql(expression, 'this')}"
     998
     999        def returning_sql(self, expression: exp.Returning) -> str:
    1000            into = self.sql(expression, "into")
    1001            into = self.seg(f"INTO {into}") if into else ""
    1002            return f"{self.seg('OUTPUT')} {self.expressions(expression, flat=True)}{into}"
    1003
    1004        def transaction_sql(self, expression: exp.Transaction) -> str:
    1005            this = self.sql(expression, "this")
    1006            this = f" {this}" if this else ""
    1007            mark = self.sql(expression, "mark")
    1008            mark = f" WITH MARK {mark}" if mark else ""
    1009            return f"BEGIN TRANSACTION{this}{mark}"
    1010
    1011        def commit_sql(self, expression: exp.Commit) -> str:
    1012            this = self.sql(expression, "this")
    1013            this = f" {this}" if this else ""
    1014            durability = expression.args.get("durability")
    1015            durability = (
    1016                f" WITH (DELAYED_DURABILITY = {'ON' if durability else 'OFF'})"
    1017                if durability is not None
    1018                else ""
    1019            )
    1020            return f"COMMIT TRANSACTION{this}{durability}"
    1021
    1022        def rollback_sql(self, expression: exp.Rollback) -> str:
    1023            this = self.sql(expression, "this")
    1024            this = f" {this}" if this else ""
    1025            return f"ROLLBACK TRANSACTION{this}"
    1026
    1027        def identifier_sql(self, expression: exp.Identifier) -> str:
    1028            identifier = super().identifier_sql(expression)
    1029
    1030            if expression.args.get("global"):
    1031                identifier = f"##{identifier}"
    1032            elif expression.args.get("temporary"):
    1033                identifier = f"#{identifier}"
    1034
    1035            return identifier
    1036
    1037        def constraint_sql(self, expression: exp.Constraint) -> str:
    1038            this = self.sql(expression, "this")
    1039            expressions = self.expressions(expression, flat=True, sep=" ")
    1040            return f"CONSTRAINT {this} {expressions}"
    1041
    1042        def length_sql(self, expression: exp.Length) -> str:
    1043            return self._uncast_text(expression, "LEN")
    1044
    1045        def right_sql(self, expression: exp.Right) -> str:
    1046            return self._uncast_text(expression, "RIGHT")
    1047
    1048        def left_sql(self, expression: exp.Left) -> str:
    1049            return self._uncast_text(expression, "LEFT")
    1050
    1051        def _uncast_text(self, expression: exp.Expression, name: str) -> str:
    1052            this = expression.this
    1053            if isinstance(this, exp.Cast) and this.is_type(exp.DataType.Type.TEXT):
    1054                this_sql = self.sql(this, "this")
    1055            else:
    1056                this_sql = self.sql(this)
    1057            expression_sql = self.sql(expression, "expression")
    1058            return self.func(name, this_sql, expression_sql if expression_sql else None)
    1059
    1060        def partition_sql(self, expression: exp.Partition) -> str:
    1061            return f"WITH (PARTITIONS({self.expressions(expression, flat=True)}))"

NORMALIZATION_STRATEGY = <NormalizationStrategy.CASE_INSENSITIVE: 'CASE_INSENSITIVE'>

Specifies the strategy according to which identifiers should be normalized.

TIME_FORMAT = "'yyyy-mm-dd hh:mm:ss'"
SUPPORTS_SEMI_ANTI_JOIN = False

Whether `SEMI` or `ANTI` joins are supported.

LOG_BASE_FIRST: Optional[bool] = False

Whether the base comes first in the `LOG` function. Possible values: `True`, `False`, `None` (two arguments are not supported by `LOG`)

TYPED_DIVISION = True

Whether the behavior of `a / b` depends on the types of `a` and `b`. False means `a / b` is always float division. True means `a / b` is integer division if both `a` and `b` are integers.

CONCAT_COALESCE = True

A `NULL` arg in `CONCAT` yields `NULL` by default, but in some dialects it yields an empty string.

TIME_MAPPING: Dict[str, str] = {'year': '%Y', 'dayofyear': '%j', 'day': '%d', 'dy': '%d', 'y': '%Y', 'week': '%W', 'ww': '%W', 'wk': '%W', 'hour': '%h', 'hh': '%I', 'minute': '%M', 'mi': '%M', 'n': '%M', 'second': '%S', 'ss': '%S', 's': '%-S', 'millisecond': '%f', 'ms': '%f', 'weekday': '%W', 'dw': '%W', 'month': '%m', 'mm': '%M', 'm': '%-M', 'Y': '%Y', 'YYYY': '%Y', 'YY': '%y', 'MMMM': '%B', 'MMM': '%b', 'MM': '%m', 'M': '%-m', 'dddd': '%A', 'dd': '%d', 'd': '%-d', 'HH': '%H', 'H': '%-H', 'h': '%-I', 'S': '%f', 'yyyy': '%Y', 'yy': '%y'}

Associates this dialect's time formats with their equivalent Python `strftime` formats.

CONVERT_FORMAT_MAPPING = {'0': '%b %d %Y %-I:%M%p', '1': '%m/%d/%y', '2': '%y.%m.%d', '3': '%d/%m/%y', '4': '%d.%m.%y', '5': '%d-%m-%y', '6': '%d %b %y', '7': '%b %d, %y', '8': '%H:%M:%S', '9': '%b %d %Y %-I:%M:%S:%f%p', '10': 'mm-dd-yy', '11': 'yy/mm/dd', '12': 'yymmdd', '13': '%d %b %Y %H:%M:ss:%f', '14': '%H:%M:%S:%f', '20': '%Y-%m-%d %H:%M:%S', '21': '%Y-%m-%d %H:%M:%S.%f', '22': '%m/%d/%y %-I:%M:%S %p', '23': '%Y-%m-%d', '24': '%H:%M:%S', '25': '%Y-%m-%d %H:%M:%S.%f', '100': '%b %d %Y %-I:%M%p', '101': '%m/%d/%Y', '102': '%Y.%m.%d', '103': '%d/%m/%Y', '104': '%d.%m.%Y', '105': '%d-%m-%Y', '106': '%d %b %Y', '107': '%b %d, %Y', '108': '%H:%M:%S', '109': '%b %d %Y %-I:%M:%S:%f%p', '110': '%m-%d-%Y', '111': '%Y/%m/%d', '112': '%Y%m%d', '113': '%d %b %Y %H:%M:%S:%f', '114': '%H:%M:%S:%f', '120': '%Y-%m-%d %H:%M:%S', '121': '%Y-%m-%d %H:%M:%S.%f'}
FORMAT_TIME_MAPPING = {'y': '%B %Y', 'd': '%m/%d/%Y', 'H': '%-H', 'h': '%-I', 's': '%Y-%m-%d %H:%M:%S', 'D': '%A,%B,%Y', 'f': '%A,%B,%Y %-I:%M %p', 'F': '%A,%B,%Y %-I:%M:%S %p', 'g': '%m/%d/%Y %-I:%M %p', 'G': '%m/%d/%Y %-I:%M:%S %p', 'M': '%B %-d', 'm': '%B %-d', 'O': '%Y-%m-%dT%H:%M:%S', 'u': '%Y-%M-%D %H:%M:%S%z', 'U': '%A, %B %D, %Y %H:%M:%S%z', 'T': '%-I:%M:%S %p', 't': '%-I:%M', 'Y': '%a %Y'}
tokenizer_class = <class 'TSQL.Tokenizer'>
parser_class = <class 'TSQL.Parser'>
generator_class = <class 'TSQL.Generator'>
TIME_TRIE: Dict = {'y': {'e': {'a': {'r': {0: True}}}, 0: True, 'y': {'y': {'y': {0: True}}, 0: True}}, 'd': {'a': {'y': {'o': {'f': {'y': {'e': {'a': {'r': {0: True}}}}}}, 0: True}}, 'y': {0: True}, 'w': {0: True}, 'd': {'d': {'d': {0: True}}, 0: True}, 0: True}, 'w': {'e': {'e': {'k': {0: True, 'd': {'a': {'y': {0: True}}}}}}, 'w': {0: True}, 'k': {0: True}}, 'h': {'o': {'u': {'r': {0: True}}}, 'h': {0: True}, 0: True}, 'm': {'i': {'n': {'u': {'t': {'e': {0: True}}}}, 0: True, 'l': {'l': {'i': {'s': {'e': {'c': {'o': {'n': {'d': {0: True}}}}}}}}}}, 's': {0: True}, 'o': {'n': {'t': {'h': {0: True}}}}, 'm': {0: True}, 0: True}, 'n': {0: True}, 's': {'e': {'c': {'o': {'n': {'d': {0: True}}}}}, 's': {0: True}, 0: True}, 'Y': {0: True, 'Y': {'Y': {'Y': {0: True}}, 0: True}}, 'M': {'M': {'M': {'M': {0: True}, 0: True}, 0: True}, 0: True}, 'H': {'H': {0: True}, 0: True}, 'S': {0: True}}
FORMAT_TRIE: Dict = {'y': {'e': {'a': {'r': {0: True}}}, 0: True, 'y': {'y': {'y': {0: True}}, 0: True}}, 'd': {'a': {'y': {'o': {'f': {'y': {'e': {'a': {'r': {0: True}}}}}}, 0: True}}, 'y': {0: True}, 'w': {0: True}, 'd': {'d': {'d': {0: True}}, 0: True}, 0: True}, 'w': {'e': {'e': {'k': {0: True, 'd': {'a': {'y': {0: True}}}}}}, 'w': {0: True}, 'k': {0: True}}, 'h': {'o': {'u': {'r': {0: True}}}, 'h': {0: True}, 0: True}, 'm': {'i': {'n': {'u': {'t': {'e': {0: True}}}}, 0: True, 'l': {'l': {'i': {'s': {'e': {'c': {'o': {'n': {'d': {0: True}}}}}}}}}}, 's': {0: True}, 'o': {'n': {'t': {'h': {0: True}}}}, 'm': {0: True}, 0: True}, 'n': {0: True}, 's': {'e': {'c': {'o': {'n': {'d': {0: True}}}}}, 's': {0: True}, 0: True}, 'Y': {0: True, 'Y': {'Y': {'Y': {0: True}}, 0: True}}, 'M': {'M': {'M': {'M': {0: True}, 0: True}, 0: True}, 0: True}, 'H': {'H': {0: True}, 0: True}, 'S': {0: True}}
INVERSE_TIME_MAPPING: Dict[str, str] = {'%Y': 'yyyy', '%j': 'dayofyear', '%d': 'dd', '%W': 'dw', '%h': 'hour', '%I': 'hh', '%M': 'mm', '%S': 'ss', '%-S': 's', '%f': 'S', '%m': 'MM', '%-M': 'm', '%y': 'yy', '%B': 'MMMM', '%b': 'MMM', '%-m': 'M', '%A': 'dddd', '%-d': 'd', '%H': 'HH', '%-H': 'H', '%-I': 'h'}
INVERSE_TIME_TRIE: Dict = {'%': {'Y': {0: True}, 'j': {0: True}, 'd': {0: True}, 'W': {0: True}, 'h': {0: True}, 'I': {0: True}, 'M': {0: True}, 'S': {0: True}, '-': {'S': {0: True}, 'M': {0: True}, 'm': {0: True}, 'd': {0: True}, 'H': {0: True}, 'I': {0: True}}, 'f': {0: True}, 'm': {0: True}, 'y': {0: True}, 'B': {0: True}, 'b': {0: True}, 'A': {0: True}, 'H': {0: True}}}
INVERSE_ESCAPE_SEQUENCES: Dict[str, str] = {}
QUOTE_START = "'"
QUOTE_END = "'"
IDENTIFIER_START = '['
IDENTIFIER_END = ']'
BIT_START: Optional[str] = None
BIT_END: Optional[str] = None
HEX_START: Optional[str] = '0x'
HEX_END: Optional[str] = ''
BYTE_START: Optional[str] = None
BYTE_END: Optional[str] = None
UNICODE_START: Optional[str] = None
UNICODE_END: Optional[str] = None

##### Inherited Members

sqlglot.dialects.dialect.Dialect
  Dialect
  INDEX_OFFSET
  WEEK_OFFSET
  UNNEST_COLUMN_ONLY
  ALIAS_POST_TABLESAMPLE
  TABLESAMPLE_SIZE_IS_PERCENT
  IDENTIFIERS_CAN_START_WITH_DIGIT
  DPIPE_IS_STRING_CONCAT
  STRICT_STRING_CONCAT
  SUPPORTS_USER_DEFINED_TYPES
  NORMALIZE_FUNCTIONS
  NULL_ORDERING
  SAFE_DIVISION
  DATE_FORMAT
  DATEINT_FORMAT
  FORMAT_MAPPING
  ESCAPE_SEQUENCES
  PSEUDOCOLUMNS
  PREFER_CTE_ALIAS_COLUMN
  get_or_raise
  format_time
  normalize_identifier
  case_sensitive
  can_identify
  quote_identifier
  to_json_path
  parse
  parse_into
  generate
  transpile
  tokenize
  tokenizer
  parser
  generator

class TSQL.Tokenizer(sqlglot.tokens.Tokenizer): View Source

    445    class Tokenizer(tokens.Tokenizer):
    446        IDENTIFIERS = [("[", "]"), '"']
    447        QUOTES = ["'", '"']
    448        HEX_STRINGS = [("0x", ""), ("0X", "")]
    449        VAR_SINGLE_TOKENS = {"@", "$", "#"}
    450
    451        KEYWORDS = {
    452            **tokens.Tokenizer.KEYWORDS,
    453            "DATETIME2": TokenType.DATETIME,
    454            "DATETIMEOFFSET": TokenType.TIMESTAMPTZ,
    455            "DECLARE": TokenType.COMMAND,
    456            "EXEC": TokenType.COMMAND,
    457            "IMAGE": TokenType.IMAGE,
    458            "MONEY": TokenType.MONEY,
    459            "NTEXT": TokenType.TEXT,
    460            "PRINT": TokenType.COMMAND,
    461            "PROC": TokenType.PROCEDURE,
    462            "REAL": TokenType.FLOAT,
    463            "ROWVERSION": TokenType.ROWVERSION,
    464            "SMALLDATETIME": TokenType.DATETIME,
    465            "SMALLMONEY": TokenType.SMALLMONEY,
    466            "SQL_VARIANT": TokenType.VARIANT,
    467            "TOP": TokenType.TOP,
    468            "UNIQUEIDENTIFIER": TokenType.UNIQUEIDENTIFIER,
    469            "UPDATE STATISTICS": TokenType.COMMAND,
    470            "XML": TokenType.XML,
    471            "OUTPUT": TokenType.RETURNING,
    472            "SYSTEM_USER": TokenType.CURRENT_USER,
    473            "FOR SYSTEM_TIME": TokenType.TIMESTAMP_SNAPSHOT,
    474            "OPTION": TokenType.OPTION,
    475        }

IDENTIFIERS = [('[', ']'), '"']
QUOTES = ["'", '"']
HEX_STRINGS = [('0x', ''), ('0X', '')]
VAR_SINGLE_TOKENS = {'$', '#', '@'}
KEYWORDS = {'{%': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{%-': <TokenType.BLOCK_START: 'BLOCK_START'>, '%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '+%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-%}': <TokenType.BLOCK_END: 'BLOCK_END'>, '{{+': <TokenType.BLOCK_START: 'BLOCK_START'>, '{{-': <TokenType.BLOCK_START: 'BLOCK_START'>, '+}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '-}}': <TokenType.BLOCK_END: 'BLOCK_END'>, '/*+': <TokenType.HINT: 'HINT'>, '==': <TokenType.EQ: 'EQ'>, '::': <TokenType.DCOLON: 'DCOLON'>, '||': <TokenType.DPIPE: 'DPIPE'>, '>=': <TokenType.GTE: 'GTE'>, '<=': <TokenType.LTE: 'LTE'>, '<>': <TokenType.NEQ: 'NEQ'>, '!=': <TokenType.NEQ: 'NEQ'>, ':=': <TokenType.COLON_EQ: 'COLON_EQ'>, '<=>': <TokenType.NULLSAFE_EQ: 'NULLSAFE_EQ'>, '->': <TokenType.ARROW: 'ARROW'>, '->>': <TokenType.DARROW: 'DARROW'>, '=>': <TokenType.FARROW: 'FARROW'>, '#>': <TokenType.HASH_ARROW: 'HASH_ARROW'>, '#>>': <TokenType.DHASH_ARROW: 'DHASH_ARROW'>, '<->': <TokenType.LR_ARROW: 'LR_ARROW'>, '&&': <TokenType.DAMP: 'DAMP'>, '??': <TokenType.DQMARK: 'DQMARK'>, 'ALL': <TokenType.ALL: 'ALL'>, 'ALWAYS': <TokenType.ALWAYS: 'ALWAYS'>, 'AND': <TokenType.AND: 'AND'>, 'ANTI': <TokenType.ANTI: 'ANTI'>, 'ANY': <TokenType.ANY: 'ANY'>, 'ASC': <TokenType.ASC: 'ASC'>, 'AS': <TokenType.ALIAS: 'ALIAS'>, 'ASOF': <TokenType.ASOF: 'ASOF'>, 'AUTOINCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'AUTO_INCREMENT': <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, 'BEGIN': <TokenType.BEGIN: 'BEGIN'>, 'BETWEEN': <TokenType.BETWEEN: 'BETWEEN'>, 'CACHE': <TokenType.CACHE: 'CACHE'>, 'UNCACHE': <TokenType.UNCACHE: 'UNCACHE'>, 'CASE': <TokenType.CASE: 'CASE'>, 'CHARACTER SET': <TokenType.CHARACTER_SET: 'CHARACTER_SET'>, 'CLUSTER BY': <TokenType.CLUSTER_BY: 'CLUSTER_BY'>, 'COLLATE': <TokenType.COLLATE: 'COLLATE'>, 'COLUMN': <TokenType.COLUMN: 'COLUMN'>, 'COMMIT': <TokenType.COMMIT: 'COMMIT'>, 'CONNECT BY': <TokenType.CONNECT_BY: 'CONNECT_BY'>, 'CONSTRAINT': <TokenType.CONSTRAINT: 'CONSTRAINT'>, 'CREATE': <TokenType.CREATE: 'CREATE'>, 'CROSS': <TokenType.CROSS: 'CROSS'>, 'CUBE': <TokenType.CUBE: 'CUBE'>, 'CURRENT_DATE': <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, 'CURRENT_TIME': <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, 'CURRENT_TIMESTAMP': <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, 'CURRENT_USER': <TokenType.CURRENT_USER: 'CURRENT_USER'>, 'DATABASE': <TokenType.DATABASE: 'DATABASE'>, 'DEFAULT': <TokenType.DEFAULT: 'DEFAULT'>, 'DELETE': <TokenType.DELETE: 'DELETE'>, 'DESC': <TokenType.DESC: 'DESC'>, 'DESCRIBE': <TokenType.DESCRIBE: 'DESCRIBE'>, 'DISTINCT': <TokenType.DISTINCT: 'DISTINCT'>, 'DISTRIBUTE BY': <TokenType.DISTRIBUTE_BY: 'DISTRIBUTE_BY'>, 'DIV': <TokenType.DIV: 'DIV'>, 'DROP': <TokenType.DROP: 'DROP'>, 'ELSE': <TokenType.ELSE: 'ELSE'>, 'END': <TokenType.END: 'END'>, 'ENUM': <TokenType.ENUM: 'ENUM'>, 'ESCAPE': <TokenType.ESCAPE: 'ESCAPE'>, 'EXCEPT': <TokenType.EXCEPT: 'EXCEPT'>, 'EXECUTE': <TokenType.EXECUTE: 'EXECUTE'>, 'EXISTS': <TokenType.EXISTS: 'EXISTS'>, 'FALSE': <TokenType.FALSE: 'FALSE'>, 'FETCH': <TokenType.FETCH: 'FETCH'>, 'FILTER': <TokenType.FILTER: 'FILTER'>, 'FIRST': <TokenType.FIRST: 'FIRST'>, 'FULL': <TokenType.FULL: 'FULL'>, 'FUNCTION': <TokenType.FUNCTION: 'FUNCTION'>, 'FOR': <TokenType.FOR: 'FOR'>, 'FOREIGN KEY': <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, 'FORMAT': <TokenType.FORMAT: 'FORMAT'>, 'FROM': <TokenType.FROM: 'FROM'>, 'GEOGRAPHY': <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, 'GEOMETRY': <TokenType.GEOMETRY: 'GEOMETRY'>, 'GLOB': <TokenType.GLOB: 'GLOB'>, 'GROUP BY': <TokenType.GROUP_BY: 'GROUP_BY'>, 'GROUPING SETS': <TokenType.GROUPING_SETS: 'GROUPING_SETS'>, 'HAVING': <TokenType.HAVING: 'HAVING'>, 'ILIKE': <TokenType.ILIKE: 'ILIKE'>, 'IN': <TokenType.IN: 'IN'>, 'INDEX': <TokenType.INDEX: 'INDEX'>, 'INET': <TokenType.INET: 'INET'>, 'INNER': <TokenType.INNER: 'INNER'>, 'INSERT': <TokenType.INSERT: 'INSERT'>, 'INTERVAL': <TokenType.INTERVAL: 'INTERVAL'>, 'INTERSECT': <TokenType.INTERSECT: 'INTERSECT'>, 'INTO': <TokenType.INTO: 'INTO'>, 'IS': <TokenType.IS: 'IS'>, 'ISNULL': <TokenType.ISNULL: 'ISNULL'>, 'JOIN': <TokenType.JOIN: 'JOIN'>, 'KEEP': <TokenType.KEEP: 'KEEP'>, 'KILL': <TokenType.KILL: 'KILL'>, 'LATERAL': <TokenType.LATERAL: 'LATERAL'>, 'LEFT': <TokenType.LEFT: 'LEFT'>, 'LIKE': <TokenType.LIKE: 'LIKE'>, 'LIMIT': <TokenType.LIMIT: 'LIMIT'>, 'LOAD': <TokenType.LOAD: 'LOAD'>, 'LOCK': <TokenType.LOCK: 'LOCK'>, 'MERGE': <TokenType.MERGE: 'MERGE'>, 'NATURAL': <TokenType.NATURAL: 'NATURAL'>, 'NEXT': <TokenType.NEXT: 'NEXT'>, 'NOT': <TokenType.NOT: 'NOT'>, 'NOTNULL': <TokenType.NOTNULL: 'NOTNULL'>, 'NULL': <TokenType.NULL: 'NULL'>, 'OBJECT': <TokenType.OBJECT: 'OBJECT'>, 'OFFSET': <TokenType.OFFSET: 'OFFSET'>, 'ON': <TokenType.ON: 'ON'>, 'OR': <TokenType.OR: 'OR'>, 'XOR': <TokenType.XOR: 'XOR'>, 'ORDER BY': <TokenType.ORDER_BY: 'ORDER_BY'>, 'ORDINALITY': <TokenType.ORDINALITY: 'ORDINALITY'>, 'OUTER': <TokenType.OUTER: 'OUTER'>, 'OVER': <TokenType.OVER: 'OVER'>, 'OVERLAPS': <TokenType.OVERLAPS: 'OVERLAPS'>, 'OVERWRITE': <TokenType.OVERWRITE: 'OVERWRITE'>, 'PARTITION': <TokenType.PARTITION: 'PARTITION'>, 'PARTITION BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PARTITIONED_BY': <TokenType.PARTITION_BY: 'PARTITION_BY'>, 'PERCENT': <TokenType.PERCENT: 'PERCENT'>, 'PIVOT': <TokenType.PIVOT: 'PIVOT'>, 'PRAGMA': <TokenType.PRAGMA: 'PRAGMA'>, 'PRIMARY KEY': <TokenType.PRIMARY_KEY: 'PRIMARY_KEY'>, 'PROCEDURE': <TokenType.PROCEDURE: 'PROCEDURE'>, 'QUALIFY': <TokenType.QUALIFY: 'QUALIFY'>, 'RANGE': <TokenType.RANGE: 'RANGE'>, 'RECURSIVE': <TokenType.RECURSIVE: 'RECURSIVE'>, 'REGEXP': <TokenType.RLIKE: 'RLIKE'>, 'REPLACE': <TokenType.REPLACE: 'REPLACE'>, 'RETURNING': <TokenType.RETURNING: 'RETURNING'>, 'REFERENCES': <TokenType.REFERENCES: 'REFERENCES'>, 'RIGHT': <TokenType.RIGHT: 'RIGHT'>, 'RLIKE': <TokenType.RLIKE: 'RLIKE'>, 'ROLLBACK': <TokenType.ROLLBACK: 'ROLLBACK'>, 'ROLLUP': <TokenType.ROLLUP: 'ROLLUP'>, 'ROW': <TokenType.ROW: 'ROW'>, 'ROWS': <TokenType.ROWS: 'ROWS'>, 'SCHEMA': <TokenType.SCHEMA: 'SCHEMA'>, 'SELECT': <TokenType.SELECT: 'SELECT'>, 'SEMI': <TokenType.SEMI: 'SEMI'>, 'SET': <TokenType.SET: 'SET'>, 'SETTINGS': <TokenType.SETTINGS: 'SETTINGS'>, 'SHOW': <TokenType.SHOW: 'SHOW'>, 'SIMILAR TO': <TokenType.SIMILAR_TO: 'SIMILAR_TO'>, 'SOME': <TokenType.SOME: 'SOME'>, 'SORT BY': <TokenType.SORT_BY: 'SORT_BY'>, 'START WITH': <TokenType.START_WITH: 'START_WITH'>, 'TABLE': <TokenType.TABLE: 'TABLE'>, 'TABLESAMPLE': <TokenType.TABLE_SAMPLE: 'TABLE_SAMPLE'>, 'TEMP': <TokenType.TEMPORARY: 'TEMPORARY'>, 'TEMPORARY': <TokenType.TEMPORARY: 'TEMPORARY'>, 'THEN': <TokenType.THEN: 'THEN'>, 'TRUE': <TokenType.TRUE: 'TRUE'>, 'TRUNCATE': <TokenType.TRUNCATE: 'TRUNCATE'>, 'UNION': <TokenType.UNION: 'UNION'>, 'UNKNOWN': <TokenType.UNKNOWN: 'UNKNOWN'>, 'UNNEST': <TokenType.UNNEST: 'UNNEST'>, 'UNPIVOT': <TokenType.UNPIVOT: 'UNPIVOT'>, 'UPDATE': <TokenType.UPDATE: 'UPDATE'>, 'USE': <TokenType.USE: 'USE'>, 'USING': <TokenType.USING: 'USING'>, 'UUID': <TokenType.UUID: 'UUID'>, 'VALUES': <TokenType.VALUES: 'VALUES'>, 'VIEW': <TokenType.VIEW: 'VIEW'>, 'VOLATILE': <TokenType.VOLATILE: 'VOLATILE'>, 'WHEN': <TokenType.WHEN: 'WHEN'>, 'WHERE': <TokenType.WHERE: 'WHERE'>, 'WINDOW': <TokenType.WINDOW: 'WINDOW'>, 'WITH': <TokenType.WITH: 'WITH'>, 'APPLY': <TokenType.APPLY: 'APPLY'>, 'ARRAY': <TokenType.ARRAY: 'ARRAY'>, 'BIT': <TokenType.BIT: 'BIT'>, 'BOOL': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BOOLEAN': <TokenType.BOOLEAN: 'BOOLEAN'>, 'BYTE': <TokenType.TINYINT: 'TINYINT'>, 'MEDIUMINT': <TokenType.MEDIUMINT: 'MEDIUMINT'>, 'INT1': <TokenType.TINYINT: 'TINYINT'>, 'TINYINT': <TokenType.TINYINT: 'TINYINT'>, 'INT16': <TokenType.SMALLINT: 'SMALLINT'>, 'SHORT': <TokenType.SMALLINT: 'SMALLINT'>, 'SMALLINT': <TokenType.SMALLINT: 'SMALLINT'>, 'INT128': <TokenType.INT128: 'INT128'>, 'HUGEINT': <TokenType.INT128: 'INT128'>, 'INT2': <TokenType.SMALLINT: 'SMALLINT'>, 'INTEGER': <TokenType.INT: 'INT'>, 'INT': <TokenType.INT: 'INT'>, 'INT4': <TokenType.INT: 'INT'>, 'INT32': <TokenType.INT: 'INT'>, 'INT64': <TokenType.BIGINT: 'BIGINT'>, 'LONG': <TokenType.BIGINT: 'BIGINT'>, 'BIGINT': <TokenType.BIGINT: 'BIGINT'>, 'INT8': <TokenType.TINYINT: 'TINYINT'>, 'DEC': <TokenType.DECIMAL: 'DECIMAL'>, 'DECIMAL': <TokenType.DECIMAL: 'DECIMAL'>, 'BIGDECIMAL': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'BIGNUMERIC': <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, 'MAP': <TokenType.MAP: 'MAP'>, 'NULLABLE': <TokenType.NULLABLE: 'NULLABLE'>, 'NUMBER': <TokenType.DECIMAL: 'DECIMAL'>, 'NUMERIC': <TokenType.DECIMAL: 'DECIMAL'>, 'FIXED': <TokenType.DECIMAL: 'DECIMAL'>, 'REAL': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT4': <TokenType.FLOAT: 'FLOAT'>, 'FLOAT8': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE': <TokenType.DOUBLE: 'DOUBLE'>, 'DOUBLE PRECISION': <TokenType.DOUBLE: 'DOUBLE'>, 'JSON': <TokenType.JSON: 'JSON'>, 'CHAR': <TokenType.CHAR: 'CHAR'>, 'CHARACTER': <TokenType.CHAR: 'CHAR'>, 'NCHAR': <TokenType.NCHAR: 'NCHAR'>, 'VARCHAR': <TokenType.VARCHAR: 'VARCHAR'>, 'VARCHAR2': <TokenType.VARCHAR: 'VARCHAR'>, 'NVARCHAR': <TokenType.NVARCHAR: 'NVARCHAR'>, 'NVARCHAR2': <TokenType.NVARCHAR: 'NVARCHAR'>, 'BPCHAR': <TokenType.BPCHAR: 'BPCHAR'>, 'STR': <TokenType.TEXT: 'TEXT'>, 'STRING': <TokenType.TEXT: 'TEXT'>, 'TEXT': <TokenType.TEXT: 'TEXT'>, 'LONGTEXT': <TokenType.LONGTEXT: 'LONGTEXT'>, 'MEDIUMTEXT': <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, 'TINYTEXT': <TokenType.TINYTEXT: 'TINYTEXT'>, 'CLOB': <TokenType.TEXT: 'TEXT'>, 'LONGVARCHAR': <TokenType.TEXT: 'TEXT'>, 'BINARY': <TokenType.BINARY: 'BINARY'>, 'BLOB': <TokenType.VARBINARY: 'VARBINARY'>, 'LONGBLOB': <TokenType.LONGBLOB: 'LONGBLOB'>, 'MEDIUMBLOB': <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, 'TINYBLOB': <TokenType.TINYBLOB: 'TINYBLOB'>, 'BYTEA': <TokenType.VARBINARY: 'VARBINARY'>, 'VARBINARY': <TokenType.VARBINARY: 'VARBINARY'>, 'TIME': <TokenType.TIME: 'TIME'>, 'TIMETZ': <TokenType.TIMETZ: 'TIMETZ'>, 'TIMESTAMP': <TokenType.TIMESTAMP: 'TIMESTAMP'>, 'TIMESTAMPTZ': <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, 'TIMESTAMPLTZ': <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, 'DATE': <TokenType.DATE: 'DATE'>, 'DATETIME': <TokenType.DATETIME: 'DATETIME'>, 'INT4RANGE': <TokenType.INT4RANGE: 'INT4RANGE'>, 'INT4MULTIRANGE': <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, 'INT8RANGE': <TokenType.INT8RANGE: 'INT8RANGE'>, 'INT8MULTIRANGE': <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, 'NUMRANGE': <TokenType.NUMRANGE: 'NUMRANGE'>, 'NUMMULTIRANGE': <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, 'TSRANGE': <TokenType.TSRANGE: 'TSRANGE'>, 'TSMULTIRANGE': <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, 'TSTZRANGE': <TokenType.TSTZRANGE: 'TSTZRANGE'>, 'TSTZMULTIRANGE': <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>, 'DATERANGE': <TokenType.DATERANGE: 'DATERANGE'>, 'DATEMULTIRANGE': <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, 'UNIQUE': <TokenType.UNIQUE: 'UNIQUE'>, 'STRUCT': <TokenType.STRUCT: 'STRUCT'>, 'SEQUENCE': <TokenType.SEQUENCE: 'SEQUENCE'>, 'VARIANT': <TokenType.VARIANT: 'VARIANT'>, 'ALTER': <TokenType.ALTER: 'ALTER'>, 'ANALYZE': <TokenType.COMMAND: 'COMMAND'>, 'CALL': <TokenType.COMMAND: 'COMMAND'>, 'COMMENT': <TokenType.COMMENT: 'COMMENT'>, 'COPY': <TokenType.COMMAND: 'COMMAND'>, 'EXPLAIN': <TokenType.COMMAND: 'COMMAND'>, 'GRANT': <TokenType.COMMAND: 'COMMAND'>, 'OPTIMIZE': <TokenType.COMMAND: 'COMMAND'>, 'PREPARE': <TokenType.COMMAND: 'COMMAND'>, 'VACUUM': <TokenType.COMMAND: 'COMMAND'>, 'USER-DEFINED': <TokenType.USERDEFINED: 'USERDEFINED'>, 'FOR VERSION': <TokenType.VERSION_SNAPSHOT: 'VERSION_SNAPSHOT'>, 'FOR TIMESTAMP': <TokenType.TIMESTAMP_SNAPSHOT: 'TIMESTAMP_SNAPSHOT'>, 'DATETIME2': <TokenType.DATETIME: 'DATETIME'>, 'DATETIMEOFFSET': <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, 'DECLARE': <TokenType.COMMAND: 'COMMAND'>, 'EXEC': <TokenType.COMMAND: 'COMMAND'>, 'IMAGE': <TokenType.IMAGE: 'IMAGE'>, 'MONEY': <TokenType.MONEY: 'MONEY'>, 'NTEXT': <TokenType.TEXT: 'TEXT'>, 'PRINT': <TokenType.COMMAND: 'COMMAND'>, 'PROC': <TokenType.PROCEDURE: 'PROCEDURE'>, 'ROWVERSION': <TokenType.ROWVERSION: 'ROWVERSION'>, 'SMALLDATETIME': <TokenType.DATETIME: 'DATETIME'>, 'SMALLMONEY': <TokenType.SMALLMONEY: 'SMALLMONEY'>, 'SQL_VARIANT': <TokenType.VARIANT: 'VARIANT'>, 'TOP': <TokenType.TOP: 'TOP'>, 'UNIQUEIDENTIFIER': <TokenType.UNIQUEIDENTIFIER: 'UNIQUEIDENTIFIER'>, 'UPDATE STATISTICS': <TokenType.COMMAND: 'COMMAND'>, 'XML': <TokenType.XML: 'XML'>, 'OUTPUT': <TokenType.RETURNING: 'RETURNING'>, 'SYSTEM_USER': <TokenType.CURRENT_USER: 'CURRENT_USER'>, 'FOR SYSTEM_TIME': <TokenType.TIMESTAMP_SNAPSHOT: 'TIMESTAMP_SNAPSHOT'>, 'OPTION': <TokenType.OPTION: 'OPTION'>}

##### Inherited Members

sqlglot.tokens.Tokenizer
  Tokenizer
  SINGLE_TOKENS
  BIT_STRINGS
  BYTE_STRINGS
  RAW_STRINGS
  HEREDOC_STRINGS
  UNICODE_STRINGS
  IDENTIFIER_ESCAPES
  STRING_ESCAPES
  HEREDOC_TAG_IS_IDENTIFIER
  HEREDOC_STRING_ALTERNATIVE
  WHITE_SPACE
  COMMANDS
  COMMAND_PREFIX_TOKENS
  NUMERIC_LITERALS
  COMMENTS
  dialect
  reset
  tokenize
  tokenize_rs
  size
  sql
  tokens

class TSQL.Parser(sqlglot.parser.Parser): View Source

    477    class Parser(parser.Parser):
    478        SET_REQUIRES_ASSIGNMENT_DELIMITER = False
    479        LOG_DEFAULTS_TO_LN = True
    480        ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN = False
    481        STRING_ALIASES = True
    482        NO_PAREN_IF_COMMANDS = False
    483
    484        QUERY_MODIFIER_PARSERS = {
    485            **parser.Parser.QUERY_MODIFIER_PARSERS,
    486            TokenType.OPTION: lambda self: ("options", self._parse_options()),
    487        }
    488
    489        FUNCTIONS = {
    490            **parser.Parser.FUNCTIONS,
    491            "CHARINDEX": lambda args: exp.StrPosition(
    492                this=seq_get(args, 1),
    493                substr=seq_get(args, 0),
    494                position=seq_get(args, 2),
    495            ),
    496            "DATEADD": build_date_delta(exp.DateAdd, unit_mapping=DATE_DELTA_INTERVAL),
    497            "DATEDIFF": _build_date_delta(exp.DateDiff, unit_mapping=DATE_DELTA_INTERVAL),
    498            "DATENAME": _build_formatted_time(exp.TimeToStr, full_format_mapping=True),
    499            "DATEPART": _build_formatted_time(exp.TimeToStr),
    500            "DATETIMEFROMPARTS": _build_datetimefromparts,
    501            "EOMONTH": _build_eomonth,
    502            "FORMAT": _build_format,
    503            "GETDATE": exp.CurrentTimestamp.from_arg_list,
    504            "HASHBYTES": _build_hashbytes,
    505            "ISNULL": exp.Coalesce.from_arg_list,
    506            "JSON_QUERY": parser.build_extract_json_with_path(exp.JSONExtract),
    507            "JSON_VALUE": parser.build_extract_json_with_path(exp.JSONExtractScalar),
    508            "LEN": _build_with_arg_as_text(exp.Length),
    509            "LEFT": _build_with_arg_as_text(exp.Left),
    510            "RIGHT": _build_with_arg_as_text(exp.Right),
    511            "REPLICATE": exp.Repeat.from_arg_list,
    512            "SQUARE": lambda args: exp.Pow(this=seq_get(args, 0), expression=exp.Literal.number(2)),
    513            "SYSDATETIME": exp.CurrentTimestamp.from_arg_list,
    514            "SUSER_NAME": exp.CurrentUser.from_arg_list,
    515            "SUSER_SNAME": exp.CurrentUser.from_arg_list,
    516            "SYSTEM_USER": exp.CurrentUser.from_arg_list,
    517            "TIMEFROMPARTS": _build_timefromparts,
    518        }
    519
    520        JOIN_HINTS = {"LOOP", "HASH", "MERGE", "REMOTE"}
    521
    522        RETURNS_TABLE_TOKENS = parser.Parser.ID_VAR_TOKENS - {
    523            TokenType.TABLE,
    524            *parser.Parser.TYPE_TOKENS,
    525        }
    526
    527        STATEMENT_PARSERS = {
    528            **parser.Parser.STATEMENT_PARSERS,
    529            TokenType.END: lambda self: self._parse_command(),
    530        }
    531
    532        def _parse_options(self) -> t.Optional[t.List[exp.Expression]]:
    533            if not self._match(TokenType.OPTION):
    534                return None
    535
    536            def _parse_option() -> t.Optional[exp.Expression]:
    537                option = self._parse_var_from_options(OPTIONS)
    538                if not option:
    539                    return None
    540
    541                self._match(TokenType.EQ)
    542                return self.expression(
    543                    exp.QueryOption, this=option, expression=self._parse_primary_or_var()
    544                )
    545
    546            return self._parse_wrapped_csv(_parse_option)
    547
    548        def _parse_projections(self) -> t.List[exp.Expression]:
    549            """
    550            T-SQL supports the syntax alias = expression in the SELECT's projection list,
    551            so we transform all parsed Selects to convert their EQ projections into Aliases.
    552
    553            See: https://learn.microsoft.com/en-us/sql/t-sql/queries/select-clause-transact-sql?view=sql-server-ver16#syntax
    554            """
    555            return [
    556                (
    557                    exp.alias_(projection.expression, projection.this.this, copy=False)
    558                    if isinstance(projection, exp.EQ) and isinstance(projection.this, exp.Column)
    559                    else projection
    560                )
    561                for projection in super()._parse_projections()
    562            ]
    563
    564        def _parse_commit_or_rollback(self) -> exp.Commit | exp.Rollback:
    565            """Applies to SQL Server and Azure SQL Database
    566            COMMIT [ { TRAN | TRANSACTION }
    567                [ transaction_name | @tran_name_variable ] ]
    568                [ WITH ( DELAYED_DURABILITY = { OFF | ON } ) ]
    569
    570            ROLLBACK { TRAN | TRANSACTION }
    571                [ transaction_name | @tran_name_variable
    572                | savepoint_name | @savepoint_variable ]
    573            """
    574            rollback = self._prev.token_type == TokenType.ROLLBACK
    575
    576            self._match_texts(("TRAN", "TRANSACTION"))
    577            this = self._parse_id_var()
    578
    579            if rollback:
    580                return self.expression(exp.Rollback, this=this)
    581
    582            durability = None
    583            if self._match_pair(TokenType.WITH, TokenType.L_PAREN):
    584                self._match_text_seq("DELAYED_DURABILITY")
    585                self._match(TokenType.EQ)
    586
    587                if self._match_text_seq("OFF"):
    588                    durability = False
    589                else:
    590                    self._match(TokenType.ON)
    591                    durability = True
    592
    593                self._match_r_paren()
    594
    595            return self.expression(exp.Commit, this=this, durability=durability)
    596
    597        def _parse_transaction(self) -> exp.Transaction | exp.Command:
    598            """Applies to SQL Server and Azure SQL Database
    599            BEGIN { TRAN | TRANSACTION }
    600            [ { transaction_name | @tran_name_variable }
    601            [ WITH MARK [ 'description' ] ]
    602            ]
    603            """
    604            if self._match_texts(("TRAN", "TRANSACTION")):
    605                transaction = self.expression(exp.Transaction, this=self._parse_id_var())
    606                if self._match_text_seq("WITH", "MARK"):
    607                    transaction.set("mark", self._parse_string())
    608
    609                return transaction
    610
    611            return self._parse_as_command(self._prev)
    612
    613        def _parse_returns(self) -> exp.ReturnsProperty:
    614            table = self._parse_id_var(any_token=False, tokens=self.RETURNS_TABLE_TOKENS)
    615            returns = super()._parse_returns()
    616            returns.set("table", table)
    617            return returns
    618
    619        def _parse_convert(
    620            self, strict: bool, safe: t.Optional[bool] = None
    621        ) -> t.Optional[exp.Expression]:
    622            this = self._parse_types()
    623            self._match(TokenType.COMMA)
    624            args = [this, *self._parse_csv(self._parse_conjunction)]
    625            convert = exp.Convert.from_arg_list(args)
    626            convert.set("safe", safe)
    627            convert.set("strict", strict)
    628            return convert
    629
    630        def _parse_user_defined_function(
    631            self, kind: t.Optional[TokenType] = None
    632        ) -> t.Optional[exp.Expression]:
    633            this = super()._parse_user_defined_function(kind=kind)
    634
    635            if (
    636                kind == TokenType.FUNCTION
    637                or isinstance(this, exp.UserDefinedFunction)
    638                or self._match(TokenType.ALIAS, advance=False)
    639            ):
    640                return this
    641
    642            expressions = self._parse_csv(self._parse_function_parameter)
    643            return self.expression(exp.UserDefinedFunction, this=this, expressions=expressions)
    644
    645        def _parse_id_var(
    646            self,
    647            any_token: bool = True,
    648            tokens: t.Optional[t.Collection[TokenType]] = None,
    649        ) -> t.Optional[exp.Expression]:
    650            is_temporary = self._match(TokenType.HASH)
    651            is_global = is_temporary and self._match(TokenType.HASH)
    652
    653            this = super()._parse_id_var(any_token=any_token, tokens=tokens)
    654            if this:
    655                if is_global:
    656                    this.set("global", True)
    657                elif is_temporary:
    658                    this.set("temporary", True)
    659
    660            return this
    661
    662        def _parse_create(self) -> exp.Create | exp.Command:
    663            create = super()._parse_create()
    664
    665            if isinstance(create, exp.Create):
    666                table = create.this.this if isinstance(create.this, exp.Schema) else create.this
    667                if isinstance(table, exp.Table) and table.this.args.get("temporary"):
    668                    if not create.args.get("properties"):
    669                        create.set("properties", exp.Properties(expressions=[]))
    670
    671                    create.args["properties"].append("expressions", exp.TemporaryProperty())
    672
    673            return create
    674
    675        def _parse_if(self) -> t.Optional[exp.Expression]:
    676            index = self._index
    677
    678            if self._match_text_seq("OBJECT_ID"):
    679                self._parse_wrapped_csv(self._parse_string)
    680                if self._match_text_seq("IS", "NOT", "NULL") and self._match(TokenType.DROP):
    681                    return self._parse_drop(exists=True)
    682                self._retreat(index)
    683
    684            return super()._parse_if()
    685
    686        def _parse_unique(self) -> exp.UniqueColumnConstraint:
    687            if self._match_texts(("CLUSTERED", "NONCLUSTERED")):
    688                this = self.CONSTRAINT_PARSERS[self._prev.text.upper()](self)
    689            else:
    690                this = self._parse_schema(self._parse_id_var(any_token=False))
    691
    692            return self.expression(exp.UniqueColumnConstraint, this=this)
    693
    694        def _parse_partition(self) -> t.Optional[exp.Partition]:
    695            if not self._match_text_seq("WITH", "(", "PARTITIONS"):
    696                return None
    697
    698            def parse_range():
    699                low = self._parse_bitwise()
    700                high = self._parse_bitwise() if self._match_text_seq("TO") else None
    701
    702                return (
    703                    self.expression(exp.PartitionRange, this=low, expression=high) if high else low
    704                )
    705
    706            partition = self.expression(
    707                exp.Partition, expressions=self._parse_wrapped_csv(parse_range)
    708            )
    709
    710            self._match_r_paren()
    711
    712            return partition

Parser consumes a list of tokens produced by the Tokenizer and produces a parsed syntax tree.

###### Arguments:

- error_level: The desired error level. Default: ErrorLevel.IMMEDIATE
- error_message_context: The amount of context to capture from a query string when displaying the error message (in number of characters). Default: 100
- max_errors: Maximum number of error messages to include in a raised ParseError. This is only relevant if error_level is ErrorLevel.RAISE. Default: 3

SET_REQUIRES_ASSIGNMENT_DELIMITER = False
LOG_DEFAULTS_TO_LN = True
ALTER_TABLE_ADD_REQUIRED_FOR_EACH_COLUMN = False
STRING_ALIASES = True
NO_PAREN_IF_COMMANDS = False
QUERY_MODIFIER_PARSERS = {<TokenType.MATCH_RECOGNIZE: 'MATCH_RECOGNIZE'>: <function Parser.<lambda>>, <TokenType.PREWHERE: 'PREWHERE'>: <function Parser.<lambda>>, <TokenType.WHERE: 'WHERE'>: <function Parser.<lambda>>, <TokenType.GROUP_BY: 'GROUP_BY'>: <function Parser.<lambda>>, <TokenType.HAVING: 'HAVING'>: <function Parser.<lambda>>, <TokenType.QUALIFY: 'QUALIFY'>: <function Parser.<lambda>>, <TokenType.WINDOW: 'WINDOW'>: <function Parser.<lambda>>, <TokenType.ORDER_BY: 'ORDER_BY'>: <function Parser.<lambda>>, <TokenType.LIMIT: 'LIMIT'>: <function Parser.<lambda>>, <TokenType.FETCH: 'FETCH'>: <function Parser.<lambda>>, <TokenType.OFFSET: 'OFFSET'>: <function Parser.<lambda>>, <TokenType.FOR: 'FOR'>: <function Parser.<lambda>>, <TokenType.LOCK: 'LOCK'>: <function Parser.<lambda>>, <TokenType.TABLE_SAMPLE: 'TABLE_SAMPLE'>: <function Parser.<lambda>>, <TokenType.USING: 'USING'>: <function Parser.<lambda>>, <TokenType.CLUSTER_BY: 'CLUSTER_BY'>: <function Parser.<lambda>>, <TokenType.DISTRIBUTE_BY: 'DISTRIBUTE_BY'>: <function Parser.<lambda>>, <TokenType.SORT_BY: 'SORT_BY'>: <function Parser.<lambda>>, <TokenType.CONNECT_BY: 'CONNECT_BY'>: <function Parser.<lambda>>, <TokenType.START_WITH: 'START_WITH'>: <function Parser.<lambda>>, <TokenType.OPTION: 'OPTION'>: <function TSQL.Parser.<lambda>>}
FUNCTIONS = {'ABS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Abs'>>, 'ADD_MONTHS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AddMonths'>>, 'ANONYMOUS_AGG_FUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnonymousAggFunc'>>, 'ANY_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.AnyValue'>>, 'APPROX_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_COUNT_DISTINCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxDistinct'>>, 'APPROX_QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxQuantile'>>, 'APPROX_TOP_K': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ApproxTopK'>>, 'ARG_MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARGMAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'MAX_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMax'>>, 'ARG_MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARGMIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'MIN_BY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArgMin'>>, 'ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Array'>>, 'ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAgg'>>, 'ARRAY_ALL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAll'>>, 'ARRAY_ANY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayAny'>>, 'ARRAY_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayConcat'>>, 'ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayContains'>>, 'FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_FILTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayFilter'>>, 'ARRAY_OVERLAPS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayOverlaps'>>, 'ARRAY_SIZE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'ARRAY_LENGTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySize'>>, 'ARRAY_SORT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySort'>>, 'ARRAY_SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArraySum'>>, 'ARRAY_TO_STRING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayToString'>>, 'ARRAY_JOIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayToString'>>, 'ARRAY_UNION_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUnionAgg'>>, 'ARRAY_UNIQUE_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ArrayUniqueAgg'>>, 'AVG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Avg'>>, 'CASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Case'>>, 'CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Cast'>>, 'CAST_TO_STR_TYPE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CastToStrType'>>, 'CBRT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Cbrt'>>, 'CEIL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CEILING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ceil'>>, 'CHR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'CHAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Chr'>>, 'COALESCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'IFNULL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'NVL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'COLLATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Collate'>>, 'COMBINED_AGG_FUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CombinedAggFunc'>>, 'COMBINED_PARAMETERIZED_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CombinedParameterizedAgg'>>, 'CONCAT': <function Parser.<lambda>>, 'CONCAT_WS': <function Parser.<lambda>>, 'CONNECT_BY_ROOT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ConnectByRoot'>>, 'CONVERT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Convert'>>, 'CORR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Corr'>>, 'COUNT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Count'>>, 'COUNT_IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CountIf'>>, 'COUNTIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CountIf'>>, 'COVAR_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CovarPop'>>, 'COVAR_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CovarSamp'>>, 'CURRENT_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDate'>>, 'CURRENT_DATETIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentDatetime'>>, 'CURRENT_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTime'>>, 'CURRENT_TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'CURRENT_USER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>, 'DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Date'>>, 'DATE_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateAdd'>>, 'DATEDIFF': <function _build_date_delta.<locals>._builder>, 'DATE_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateDiff'>>, 'DATE_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateFromParts'>>, 'DATEFROMPARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateFromParts'>>, 'DATE_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateStrToDate'>>, 'DATE_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateSub'>>, 'DATE_TO_DATE_STR': <function Parser.<lambda>>, 'DATE_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateToDi'>>, 'DATE_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DateTrunc'>>, 'DATETIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeAdd'>>, 'DATETIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeDiff'>>, 'DATETIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeSub'>>, 'DATETIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DatetimeTrunc'>>, 'DAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Day'>>, 'DAY_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAYOFMONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfMonth'>>, 'DAY_OF_WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAYOFWEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfWeek'>>, 'DAY_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DAYOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DayOfYear'>>, 'DECODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Decode'>>, 'DI_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.DiToDate'>>, 'ENCODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Encode'>>, 'EXP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Exp'>>, 'EXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Explode'>>, 'EXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ExplodeOuter'>>, 'EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Extract'>>, 'FIRST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.First'>>, 'FIRST_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FirstValue'>>, 'FLATTEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Flatten'>>, 'FLOOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Floor'>>, 'FROM_BASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase'>>, 'FROM_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.FromBase64'>>, 'GENERATE_DATE_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GenerateDateArray'>>, 'GENERATE_SERIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GenerateSeries'>>, 'GREATEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Greatest'>>, 'GROUP_CONCAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.GroupConcat'>>, 'HEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Hex'>>, 'HLL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Hll'>>, 'IF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.If'>>, 'IIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.If'>>, 'INITCAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Initcap'>>, 'IS_INF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'ISINF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsInf'>>, 'IS_NAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'ISNAN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.IsNan'>>, 'J_S_O_N_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArray'>>, 'J_S_O_N_ARRAY_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayAgg'>>, 'JSON_ARRAY_CONTAINS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONArrayContains'>>, 'JSONB_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtract'>>, 'JSONB_EXTRACT_SCALAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONBExtractScalar'>>, 'JSON_EXTRACT': <function build_extract_json_with_path.<locals>._builder>, 'JSON_EXTRACT_SCALAR': <function build_extract_json_with_path.<locals>._builder>, 'JSON_FORMAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONFormat'>>, 'J_S_O_N_OBJECT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONObject'>>, 'J_S_O_N_OBJECT_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONObjectAgg'>>, 'J_S_O_N_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.JSONTable'>>, 'LAG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lag'>>, 'LAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Last'>>, 'LAST_DAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastDay'>>, 'LAST_DAY_OF_MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastDay'>>, 'LAST_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LastValue'>>, 'LEAD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lead'>>, 'LEAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Least'>>, 'LEFT': <function _build_with_arg_as_text.<locals>._parse>, 'LENGTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Length'>>, 'LEN': <function _build_with_arg_as_text.<locals>._parse>, 'LEVENSHTEIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Levenshtein'>>, 'LN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Ln'>>, 'LOG': <function build_logarithm>, 'LOGICAL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOL_AND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'BOOLAND_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalAnd'>>, 'LOGICAL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOL_OR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'BOOLOR_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.LogicalOr'>>, 'LOWER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lower'>>, 'LCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Lower'>>, 'MD5': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5'>>, 'MD5_DIGEST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MD5Digest'>>, 'MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Map'>>, 'MAP_FROM_ENTRIES': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MapFromEntries'>>, 'MATCH_AGAINST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MatchAgainst'>>, 'MAX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Max'>>, 'MIN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Min'>>, 'MONTH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Month'>>, 'MONTHS_BETWEEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.MonthsBetween'>>, 'NEXT_VALUE_FOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NextValueFor'>>, 'NTH_VALUE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NthValue'>>, 'NULLIF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nullif'>>, 'NUMBER_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.NumberToStr'>>, 'NVL2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Nvl2'>>, 'OPEN_J_S_O_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.OpenJSON'>>, 'PARAMETERIZED_AGG': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParameterizedAgg'>>, 'PARSE_JSON': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'JSON_PARSE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ParseJSON'>>, 'PERCENTILE_CONT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileCont'>>, 'PERCENTILE_DISC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PercentileDisc'>>, 'POSEXPLODE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Posexplode'>>, 'POSEXPLODE_OUTER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.PosexplodeOuter'>>, 'POWER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'POW': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Pow'>>, 'PREDICT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Predict'>>, 'QUANTILE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Quantile'>>, 'RAND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Rand'>>, 'RANDOM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Rand'>>, 'RANDN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Randn'>>, 'RANGE_N': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RangeN'>>, 'READ_CSV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ReadCSV'>>, 'REDUCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Reduce'>>, 'REGEXP_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpExtract'>>, 'REGEXP_I_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpILike'>>, 'REGEXP_LIKE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpLike'>>, 'REGEXP_REPLACE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpReplace'>>, 'REGEXP_SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RegexpSplit'>>, 'REPEAT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Repeat'>>, 'RIGHT': <function _build_with_arg_as_text.<locals>._parse>, 'ROUND': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Round'>>, 'ROW_NUMBER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.RowNumber'>>, 'SHA': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA1': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA'>>, 'SHA2': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SHA2'>>, 'SAFE_DIVIDE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SafeDivide'>>, 'SIGN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sign'>>, 'SIGNUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sign'>>, 'SORT_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.SortArray'>>, 'SPLIT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Split'>>, 'SQRT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sqrt'>>, 'STANDARD_HASH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StandardHash'>>, 'STAR_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StarMap'>>, 'STARTS_WITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STARTSWITH': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StartsWith'>>, 'STDDEV': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stddev'>>, 'STDDEV_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevPop'>>, 'STDDEV_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StddevSamp'>>, 'STR_POSITION': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrPosition'>>, 'STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToDate'>>, 'STR_TO_MAP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToMap'>>, 'STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToTime'>>, 'STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StrToUnix'>>, 'STRUCT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Struct'>>, 'STRUCT_EXTRACT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.StructExtract'>>, 'STUFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'INSERT': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Stuff'>>, 'SUBSTRING': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Substring'>>, 'SUM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Sum'>>, 'TIME_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeAdd'>>, 'TIME_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeDiff'>>, 'TIME_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeFromParts'>>, 'TIMEFROMPARTS': <function _build_timefromparts>, 'TIME_STR_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToDate'>>, 'TIME_STR_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToTime'>>, 'TIME_STR_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeStrToUnix'>>, 'TIME_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeSub'>>, 'TIME_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToStr'>>, 'TIME_TO_TIME_STR': <function Parser.<lambda>>, 'TIME_TO_UNIX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeToUnix'>>, 'TIME_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimeTrunc'>>, 'TIMESTAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Timestamp'>>, 'TIMESTAMP_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampAdd'>>, 'TIMESTAMPDIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampDiff'>>, 'TIMESTAMP_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampDiff'>>, 'TIMESTAMP_FROM_PARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampFromParts'>>, 'TIMESTAMPFROMPARTS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampFromParts'>>, 'TIMESTAMP_SUB': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampSub'>>, 'TIMESTAMP_TRUNC': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TimestampTrunc'>>, 'TO_ARRAY': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToArray'>>, 'TO_BASE64': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToBase64'>>, 'TO_CHAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToChar'>>, 'TO_DAYS': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToDays'>>, 'TO_NUMBER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.ToNumber'>>, 'TRANSFORM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Transform'>>, 'TRIM': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Trim'>>, 'TRY_CAST': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TryCast'>>, 'TS_OR_DI_TO_DI': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDiToDi'>>, 'TS_OR_DS_ADD': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsAdd'>>, 'TS_OR_DS_DIFF': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsDiff'>>, 'TS_OR_DS_TO_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToDate'>>, 'TS_OR_DS_TO_DATE_STR': <function Parser.<lambda>>, 'TS_OR_DS_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.TsOrDsToTime'>>, 'UNHEX': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Unhex'>>, 'UNIX_DATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixDate'>>, 'UNIX_TO_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToStr'>>, 'UNIX_TO_TIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTime'>>, 'UNIX_TO_TIME_STR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.UnixToTimeStr'>>, 'UPPER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Upper'>>, 'UCASE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Upper'>>, 'VAR_MAP': <function build_var_map>, 'VARIANCE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VAR_SAMP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Variance'>>, 'VARIANCE_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'VAR_POP': <bound method Func.from_arg_list of <class 'sqlglot.expressions.VariancePop'>>, 'WEEK': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Week'>>, 'WEEK_OF_YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WEEKOFYEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.WeekOfYear'>>, 'WHEN': <bound method Func.from_arg_list of <class 'sqlglot.expressions.When'>>, 'X_M_L_TABLE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.XMLTable'>>, 'XOR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Xor'>>, 'YEAR': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Year'>>, 'GLOB': <function Parser.<lambda>>, 'JSON_EXTRACT_PATH_TEXT': <function build_extract_json_with_path.<locals>._builder>, 'LIKE': <function build_like>, 'LOG2': <function Parser.<lambda>>, 'LOG10': <function Parser.<lambda>>, 'MOD': <function Parser.<lambda>>, 'CHARINDEX': <function TSQL.Parser.<lambda>>, 'DATEADD': <function build_date_delta.<locals>._builder>, 'DATENAME': <function _build_formatted_time.<locals>._builder>, 'DATEPART': <function _build_formatted_time.<locals>._builder>, 'DATETIMEFROMPARTS': <function _build_datetimefromparts>, 'EOMONTH': <function _build_eomonth>, 'FORMAT': <function _build_format>, 'GETDATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'HASHBYTES': <function _build_hashbytes>, 'ISNULL': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Coalesce'>>, 'JSON_QUERY': <function build_extract_json_with_path.<locals>._builder>, 'JSON_VALUE': <function build_extract_json_with_path.<locals>._builder>, 'REPLICATE': <bound method Func.from_arg_list of <class 'sqlglot.expressions.Repeat'>>, 'SQUARE': <function TSQL.Parser.<lambda>>, 'SYSDATETIME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentTimestamp'>>, 'SUSER_NAME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>, 'SUSER_SNAME': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>, 'SYSTEM_USER': <bound method Func.from_arg_list of <class 'sqlglot.expressions.CurrentUser'>>}
JOIN_HINTS = {'MERGE', 'LOOP', 'REMOTE', 'HASH'}
RETURNS_TABLE_TOKENS = {<TokenType.ASC: 'ASC'>, <TokenType.DATABASE: 'DATABASE'>, <TokenType.INDEX: 'INDEX'>, <TokenType.SHOW: 'SHOW'>, <TokenType.FULL: 'FULL'>, <TokenType.ANY: 'ANY'>, <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, <TokenType.EXISTS: 'EXISTS'>, <TokenType.FORMAT: 'FORMAT'>, <TokenType.SCHEMA: 'SCHEMA'>, <TokenType.SEMI: 'SEMI'>, <TokenType.SETTINGS: 'SETTINGS'>, <TokenType.FALSE: 'FALSE'>, <TokenType.ORDINALITY: 'ORDINALITY'>, <TokenType.VAR: 'VAR'>, <TokenType.PRAGMA: 'PRAGMA'>, <TokenType.ANTI: 'ANTI'>, <TokenType.FINAL: 'FINAL'>, <TokenType.VOLATILE: 'VOLATILE'>, <TokenType.PROCEDURE: 'PROCEDURE'>, <TokenType.UNPIVOT: 'UNPIVOT'>, <TokenType.ASOF: 'ASOF'>, <TokenType.COLLATE: 'COLLATE'>, <TokenType.FIRST: 'FIRST'>, <TokenType.COMMIT: 'COMMIT'>, <TokenType.KILL: 'KILL'>, <TokenType.REPLACE: 'REPLACE'>, <TokenType.SET: 'SET'>, <TokenType.OVERLAPS: 'OVERLAPS'>, <TokenType.OFFSET: 'OFFSET'>, <TokenType.USE: 'USE'>, <TokenType.REFRESH: 'REFRESH'>, <TokenType.CURRENT_USER: 'CURRENT_USER'>, <TokenType.DEFAULT: 'DEFAULT'>, <TokenType.EXECUTE: 'EXECUTE'>, <TokenType.MERGE: 'MERGE'>, <TokenType.CURRENT_DATETIME: 'CURRENT_DATETIME'>, <TokenType.DICTIONARY: 'DICTIONARY'>, <TokenType.VIEW: 'VIEW'>, <TokenType.UPDATE: 'UPDATE'>, <TokenType.TRUE: 'TRUE'>, <TokenType.CASE: 'CASE'>, <TokenType.OPERATOR: 'OPERATOR'>, <TokenType.BEGIN: 'BEGIN'>, <TokenType.END: 'END'>, <TokenType.DELETE: 'DELETE'>, <TokenType.MODEL: 'MODEL'>, <TokenType.REFERENCES: 'REFERENCES'>, <TokenType.ISNULL: 'ISNULL'>, <TokenType.RIGHT: 'RIGHT'>, <TokenType.IS: 'IS'>, <TokenType.PARTITION: 'PARTITION'>, <TokenType.FILTER: 'FILTER'>, <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, <TokenType.WINDOW: 'WINDOW'>, <TokenType.SEQUENCE: 'SEQUENCE'>, <TokenType.UNIQUE: 'UNIQUE'>, <TokenType.TEMPORARY: 'TEMPORARY'>, <TokenType.TRUNCATE: 'TRUNCATE'>, <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, <TokenType.LOAD: 'LOAD'>, <TokenType.PERCENT: 'PERCENT'>, <TokenType.PIVOT: 'PIVOT'>, <TokenType.ROWS: 'ROWS'>, <TokenType.FUNCTION: 'FUNCTION'>, <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, <TokenType.KEEP: 'KEEP'>, <TokenType.LEFT: 'LEFT'>, <TokenType.ESCAPE: 'ESCAPE'>, <TokenType.DESCRIBE: 'DESCRIBE'>, <TokenType.COLUMN: 'COLUMN'>, <TokenType.DIV: 'DIV'>, <TokenType.OVERWRITE: 'OVERWRITE'>, <TokenType.RANGE: 'RANGE'>, <TokenType.TOP: 'TOP'>, <TokenType.DESC: 'DESC'>, <TokenType.NATURAL: 'NATURAL'>, <TokenType.CONSTRAINT: 'CONSTRAINT'>, <TokenType.CACHE: 'CACHE'>, <TokenType.ALL: 'ALL'>, <TokenType.SOME: 'SOME'>, <TokenType.APPLY: 'APPLY'>, <TokenType.ROW: 'ROW'>, <TokenType.COMMAND: 'COMMAND'>, <TokenType.STORAGE_INTEGRATION: 'STORAGE_INTEGRATION'>, <TokenType.NEXT: 'NEXT'>, <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, <TokenType.RECURSIVE: 'RECURSIVE'>, <TokenType.COMMENT: 'COMMENT'>}
STATEMENT_PARSERS = {<TokenType.ALTER: 'ALTER'>: <function Parser.<lambda>>, <TokenType.BEGIN: 'BEGIN'>: <function Parser.<lambda>>, <TokenType.CACHE: 'CACHE'>: <function Parser.<lambda>>, <TokenType.COMMENT: 'COMMENT'>: <function Parser.<lambda>>, <TokenType.COMMIT: 'COMMIT'>: <function Parser.<lambda>>, <TokenType.CREATE: 'CREATE'>: <function Parser.<lambda>>, <TokenType.DELETE: 'DELETE'>: <function Parser.<lambda>>, <TokenType.DESC: 'DESC'>: <function Parser.<lambda>>, <TokenType.DESCRIBE: 'DESCRIBE'>: <function Parser.<lambda>>, <TokenType.DROP: 'DROP'>: <function Parser.<lambda>>, <TokenType.INSERT: 'INSERT'>: <function Parser.<lambda>>, <TokenType.KILL: 'KILL'>: <function Parser.<lambda>>, <TokenType.LOAD: 'LOAD'>: <function Parser.<lambda>>, <TokenType.MERGE: 'MERGE'>: <function Parser.<lambda>>, <TokenType.PIVOT: 'PIVOT'>: <function Parser.<lambda>>, <TokenType.PRAGMA: 'PRAGMA'>: <function Parser.<lambda>>, <TokenType.REFRESH: 'REFRESH'>: <function Parser.<lambda>>, <TokenType.ROLLBACK: 'ROLLBACK'>: <function Parser.<lambda>>, <TokenType.SET: 'SET'>: <function Parser.<lambda>>, <TokenType.TRUNCATE: 'TRUNCATE'>: <function Parser.<lambda>>, <TokenType.UNCACHE: 'UNCACHE'>: <function Parser.<lambda>>, <TokenType.UPDATE: 'UPDATE'>: <function Parser.<lambda>>, <TokenType.USE: 'USE'>: <function Parser.<lambda>>, <TokenType.END: 'END'>: <function TSQL.Parser.<lambda>>}
TABLE_ALIAS_TOKENS = {<TokenType.DECIMAL: 'DECIMAL'>, <TokenType.ASC: 'ASC'>, <TokenType.BPCHAR: 'BPCHAR'>, <TokenType.DATABASE: 'DATABASE'>, <TokenType.MEDIUMINT: 'MEDIUMINT'>, <TokenType.INDEX: 'INDEX'>, <TokenType.LOWCARDINALITY: 'LOWCARDINALITY'>, <TokenType.DATEMULTIRANGE: 'DATEMULTIRANGE'>, <TokenType.SHOW: 'SHOW'>, <TokenType.BIGINT: 'BIGINT'>, <TokenType.UUID: 'UUID'>, <TokenType.TIMESTAMP_S: 'TIMESTAMP_S'>, <TokenType.ANY: 'ANY'>, <TokenType.INT8RANGE: 'INT8RANGE'>, <TokenType.SMALLINT: 'SMALLINT'>, <TokenType.TINYINT: 'TINYINT'>, <TokenType.CURRENT_TIME: 'CURRENT_TIME'>, <TokenType.EXISTS: 'EXISTS'>, <TokenType.BOOLEAN: 'BOOLEAN'>, <TokenType.LONGTEXT: 'LONGTEXT'>, <TokenType.FORMAT: 'FORMAT'>, <TokenType.UNIQUEIDENTIFIER: 'UNIQUEIDENTIFIER'>, <TokenType.UINT128: 'UINT128'>, <TokenType.SCHEMA: 'SCHEMA'>, <TokenType.HLLSKETCH: 'HLLSKETCH'>, <TokenType.SEMI: 'SEMI'>, <TokenType.SETTINGS: 'SETTINGS'>, <TokenType.FALSE: 'FALSE'>, <TokenType.ORDINALITY: 'ORDINALITY'>, <TokenType.TIMETZ: 'TIMETZ'>, <TokenType.VAR: 'VAR'>, <TokenType.PRAGMA: 'PRAGMA'>, <TokenType.VOLATILE: 'VOLATILE'>, <TokenType.FINAL: 'FINAL'>, <TokenType.INET: 'INET'>, <TokenType.ANTI: 'ANTI'>, <TokenType.PROCEDURE: 'PROCEDURE'>, <TokenType.INT: 'INT'>, <TokenType.UDECIMAL: 'UDECIMAL'>, <TokenType.UNPIVOT: 'UNPIVOT'>, <TokenType.NUMMULTIRANGE: 'NUMMULTIRANGE'>, <TokenType.DOUBLE: 'DOUBLE'>, <TokenType.COLLATE: 'COLLATE'>, <TokenType.FIRST: 'FIRST'>, <TokenType.COMMIT: 'COMMIT'>, <TokenType.KILL: 'KILL'>, <TokenType.REPLACE: 'REPLACE'>, <TokenType.UINT256: 'UINT256'>, <TokenType.TSRANGE: 'TSRANGE'>, <TokenType.SERIAL: 'SERIAL'>, <TokenType.SET: 'SET'>, <TokenType.MAP: 'MAP'>, <TokenType.OBJECT: 'OBJECT'>, <TokenType.OVERLAPS: 'OVERLAPS'>, <TokenType.USE: 'USE'>, <TokenType.ROWVERSION: 'ROWVERSION'>, <TokenType.TIMESTAMP: 'TIMESTAMP'>, <TokenType.REFRESH: 'REFRESH'>, <TokenType.NCHAR: 'NCHAR'>, <TokenType.INT128: 'INT128'>, <TokenType.DATE: 'DATE'>, <TokenType.ENUM: 'ENUM'>, <TokenType.BINARY: 'BINARY'>, <TokenType.CURRENT_USER: 'CURRENT_USER'>, <TokenType.DEFAULT: 'DEFAULT'>, <TokenType.EXECUTE: 'EXECUTE'>, <TokenType.FLOAT: 'FLOAT'>, <TokenType.TIMESTAMP_NS: 'TIMESTAMP_NS'>, <TokenType.STRUCT: 'STRUCT'>, <TokenType.MERGE: 'MERGE'>, <TokenType.CURRENT_DATETIME: 'CURRENT_DATETIME'>, <TokenType.DICTIONARY: 'DICTIONARY'>, <TokenType.VIEW: 'VIEW'>, <TokenType.VARBINARY: 'VARBINARY'>, <TokenType.FIXEDSTRING: 'FIXEDSTRING'>, <TokenType.IPV6: 'IPV6'>, <TokenType.HSTORE: 'HSTORE'>, <TokenType.ENUM8: 'ENUM8'>, <TokenType.USERDEFINED: 'USERDEFINED'>, <TokenType.TIMESTAMPLTZ: 'TIMESTAMPLTZ'>, <TokenType.UINT: 'UINT'>, <TokenType.UBIGINT: 'UBIGINT'>, <TokenType.IPADDRESS: 'IPADDRESS'>, <TokenType.UPDATE: 'UPDATE'>, <TokenType.TRUE: 'TRUE'>, <TokenType.CASE: 'CASE'>, <TokenType.VARCHAR: 'VARCHAR'>, <TokenType.NUMRANGE: 'NUMRANGE'>, <TokenType.BEGIN: 'BEGIN'>, <TokenType.OPERATOR: 'OPERATOR'>, <TokenType.END: 'END'>, <TokenType.DELETE: 'DELETE'>, <TokenType.TIMESTAMP_MS: 'TIMESTAMP_MS'>, <TokenType.MODEL: 'MODEL'>, <TokenType.REFERENCES: 'REFERENCES'>, <TokenType.BIT: 'BIT'>, <TokenType.USMALLINT: 'USMALLINT'>, <TokenType.TSMULTIRANGE: 'TSMULTIRANGE'>, <TokenType.ARRAY: 'ARRAY'>, <TokenType.INTERVAL: 'INTERVAL'>, <TokenType.BIGSERIAL: 'BIGSERIAL'>, <TokenType.TIMESTAMPTZ: 'TIMESTAMPTZ'>, <TokenType.ISNULL: 'ISNULL'>, <TokenType.SUPER: 'SUPER'>, <TokenType.YEAR: 'YEAR'>, <TokenType.ENUM16: 'ENUM16'>, <TokenType.IS: 'IS'>, <TokenType.TINYBLOB: 'TINYBLOB'>, <TokenType.PARTITION: 'PARTITION'>, <TokenType.INT4RANGE: 'INT4RANGE'>, <TokenType.FILTER: 'FILTER'>, <TokenType.SMALLSERIAL: 'SMALLSERIAL'>, <TokenType.AUTO_INCREMENT: 'AUTO_INCREMENT'>, <TokenType.PSEUDO_TYPE: 'PSEUDO_TYPE'>, <TokenType.IMAGE: 'IMAGE'>, <TokenType.DATETIME64: 'DATETIME64'>, <TokenType.MEDIUMTEXT: 'MEDIUMTEXT'>, <TokenType.NESTED: 'NESTED'>, <TokenType.SEQUENCE: 'SEQUENCE'>, <TokenType.UNIQUE: 'UNIQUE'>, <TokenType.TEMPORARY: 'TEMPORARY'>, <TokenType.TRUNCATE: 'TRUNCATE'>, <TokenType.SMALLMONEY: 'SMALLMONEY'>, <TokenType.UTINYINT: 'UTINYINT'>, <TokenType.CURRENT_DATE: 'CURRENT_DATE'>, <TokenType.LOAD: 'LOAD'>, <TokenType.PERCENT: 'PERCENT'>, <TokenType.PIVOT: 'PIVOT'>, <TokenType.MONEY: 'MONEY'>, <TokenType.TIME: 'TIME'>, <TokenType.ROWS: 'ROWS'>, <TokenType.INT4MULTIRANGE: 'INT4MULTIRANGE'>, <TokenType.FUNCTION: 'FUNCTION'>, <TokenType.GEOGRAPHY: 'GEOGRAPHY'>, <TokenType.AGGREGATEFUNCTION: 'AGGREGATEFUNCTION'>, <TokenType.CURRENT_TIMESTAMP: 'CURRENT_TIMESTAMP'>, <TokenType.TABLE: 'TABLE'>, <TokenType.KEEP: 'KEEP'>, <TokenType.BIGDECIMAL: 'BIGDECIMAL'>, <TokenType.INT8MULTIRANGE: 'INT8MULTIRANGE'>, <TokenType.UMEDIUMINT: 'UMEDIUMINT'>, <TokenType.IPV4: 'IPV4'>, <TokenType.NULL: 'NULL'>, <TokenType.ESCAPE: 'ESCAPE'>, <TokenType.NAME: 'NAME'>, <TokenType.XML: 'XML'>, <TokenType.LONGBLOB: 'LONGBLOB'>, <TokenType.JSON: 'JSON'>, <TokenType.DESCRIBE: 'DESCRIBE'>, <TokenType.DATERANGE: 'DATERANGE'>, <TokenType.INT256: 'INT256'>, <TokenType.COLUMN: 'COLUMN'>, <TokenType.DIV: 'DIV'>, <TokenType.GEOMETRY: 'GEOMETRY'>, <TokenType.OVERWRITE: 'OVERWRITE'>, <TokenType.TINYTEXT: 'TINYTEXT'>, <TokenType.RANGE: 'RANGE'>, <TokenType.OBJECT_IDENTIFIER: 'OBJECT_IDENTIFIER'>, <TokenType.TOP: 'TOP'>, <TokenType.TEXT: 'TEXT'>, <TokenType.DESC: 'DESC'>, <TokenType.UNKNOWN: 'UNKNOWN'>, <TokenType.DATETIME: 'DATETIME'>, <TokenType.NVARCHAR: 'NVARCHAR'>, <TokenType.CONSTRAINT: 'CONSTRAINT'>, <TokenType.MEDIUMBLOB: 'MEDIUMBLOB'>, <TokenType.CACHE: 'CACHE'>, <TokenType.JSONB: 'JSONB'>, <TokenType.DATE32: 'DATE32'>, <TokenType.ALL: 'ALL'>, <TokenType.SOME: 'SOME'>, <TokenType.ROW: 'ROW'>, <TokenType.STORAGE_INTEGRATION: 'STORAGE_INTEGRATION'>, <TokenType.COMMAND: 'COMMAND'>, <TokenType.NEXT: 'NEXT'>, <TokenType.FOREIGN_KEY: 'FOREIGN_KEY'>, <TokenType.IPPREFIX: 'IPPREFIX'>, <TokenType.RECURSIVE: 'RECURSIVE'>, <TokenType.COMMENT: 'COMMENT'>, <TokenType.CHAR: 'CHAR'>, <TokenType.SIMPLEAGGREGATEFUNCTION: 'SIMPLEAGGREGATEFUNCTION'>, <TokenType.TSTZRANGE: 'TSTZRANGE'>, <TokenType.VARIANT: 'VARIANT'>, <TokenType.NULLABLE: 'NULLABLE'>, <TokenType.TSTZMULTIRANGE: 'TSTZMULTIRANGE'>}
SHOW_TRIE: Dict = {}
SET_TRIE: Dict = {'GLOBAL': {0: True}, 'LOCAL': {0: True}, 'SESSION': {0: True}, 'TRANSACTION': {0: True}}

##### Inherited Members

sqlglot.parser.Parser
  Parser
  NO_PAREN_FUNCTIONS
  STRUCT_TYPE_TOKENS
  NESTED_TYPE_TOKENS
  ENUM_TYPE_TOKENS
  AGGREGATE_TYPE_TOKENS
  TYPE_TOKENS
  SIGNED_TO_UNSIGNED_TYPE_TOKEN
  SUBQUERY_PREDICATES
  RESERVED_TOKENS
  DB_CREATABLES
  CREATABLES
  ID_VAR_TOKENS
  INTERVAL_VARS
  ALIAS_TOKENS
  COMMENT_TABLE_ALIAS_TOKENS
  UPDATE_ALIAS_TOKENS
  TRIM_TYPES
  FUNC_TOKENS
  CONJUNCTION
  EQUALITY
  COMPARISON
  BITWISE
  TERM
  FACTOR
  EXPONENT
  TIMES
  TIMESTAMPS
  SET_OPERATIONS
  JOIN_METHODS
  JOIN_SIDES
  JOIN_KINDS
  LAMBDAS
  COLUMN_OPERATORS
  EXPRESSION_PARSERS
  UNARY_PARSERS
  STRING_PARSERS
  NUMERIC_PARSERS
  PRIMARY_PARSERS
  PLACEHOLDER_PARSERS
  RANGE_PARSERS
  PROPERTY_PARSERS
  CONSTRAINT_PARSERS
  ALTER_PARSERS
  SCHEMA_UNNAMED_CONSTRAINTS
  NO_PAREN_FUNCTION_PARSERS
  INVALID_FUNC_NAME_TOKENS
  FUNCTIONS_WITH_ALIASED_ARGS
  KEY_VALUE_DEFINITIONS
  FUNCTION_PARSERS
  SET_PARSERS
  SHOW_PARSERS
  TYPE_LITERAL_PARSERS
  DDL_SELECT_TOKENS
  PRE_VOLATILE_TOKENS
  TRANSACTION_KIND
  TRANSACTION_CHARACTERISTICS
  CONFLICT_ACTIONS
  CREATE_SEQUENCE
  ISOLATED_LOADING_OPTIONS
  USABLES
  CAST_ACTIONS
  INSERT_ALTERNATIVES
  CLONE_KEYWORDS
  HISTORICAL_DATA_KIND
  OPCLASS_FOLLOW_KEYWORDS
  OPTYPE_FOLLOW_TOKENS
  TABLE_INDEX_HINT_TOKENS
  VIEW_ATTRIBUTES
  WINDOW_ALIAS_TOKENS
  WINDOW_BEFORE_PAREN_TOKENS
  WINDOW_SIDES
  JSON_KEY_VALUE_SEPARATOR_TOKENS
  FETCH_TOKENS
  ADD_CONSTRAINT_TOKENS
  DISTINCT_TOKENS
  NULL_TOKENS
  UNNEST_OFFSET_ALIAS_TOKENS
  SELECT_START_TOKENS
  STRICT_CAST
  PREFIXED_PIVOT_COLUMNS
  IDENTIFY_PIVOT_STRINGS
  TABLESAMPLE_CSV
  TRIM_PATTERN_FIRST
  MODIFIERS_ATTACHED_TO_UNION
  UNION_MODIFIERS
  JSON_ARROWS_REQUIRE_JSON_TYPE
  VALUES_FOLLOWED_BY_PAREN
  SUPPORTS_IMPLICIT_UNNEST
  error_level
  error_message_context
  max_errors
  dialect
  reset
  parse
  parse_into
  check_errors
  raise_error
  expression
  validate_expression
  errors
  sql

class TSQL.Generator(sqlglot.generator.Generator): View Source

     714    class Generator(generator.Generator):
     715        LIMIT_IS_TOP = True
     716        QUERY_HINTS = False
     717        RETURNING_END = False
     718        NVL2_SUPPORTED = False
     719        ALTER_TABLE_INCLUDE_COLUMN_KEYWORD = False
     720        LIMIT_FETCH = "FETCH"
     721        COMPUTED_COLUMN_WITH_TYPE = False
     722        CTE_RECURSIVE_KEYWORD_REQUIRED = False
     723        ENSURE_BOOLS = True
     724        NULL_ORDERING_SUPPORTED = None
     725        SUPPORTS_SINGLE_ARG_CONCAT = False
     726        TABLESAMPLE_SEED_KEYWORD = "REPEATABLE"
     727        SUPPORTS_SELECT_INTO = True
     728        JSON_PATH_BRACKETED_KEY_SUPPORTED = False
     729        SUPPORTS_TO_NUMBER = False
     730
     731        EXPRESSIONS_WITHOUT_NESTED_CTES = {
     732            exp.Delete,
     733            exp.Insert,
     734            exp.Merge,
     735            exp.Select,
     736            exp.Subquery,
     737            exp.Union,
     738            exp.Update,
     739        }
     740
     741        SUPPORTED_JSON_PATH_PARTS = {
     742            exp.JSONPathKey,
     743            exp.JSONPathRoot,
     744            exp.JSONPathSubscript,
     745        }
     746
     747        TYPE_MAPPING = {
     748            **generator.Generator.TYPE_MAPPING,
     749            exp.DataType.Type.BOOLEAN: "BIT",
     750            exp.DataType.Type.DECIMAL: "NUMERIC",
     751            exp.DataType.Type.DATETIME: "DATETIME2",
     752            exp.DataType.Type.DOUBLE: "FLOAT",
     753            exp.DataType.Type.INT: "INTEGER",
     754            exp.DataType.Type.TEXT: "VARCHAR(MAX)",
     755            exp.DataType.Type.TIMESTAMP: "DATETIME2",
     756            exp.DataType.Type.TIMESTAMPTZ: "DATETIMEOFFSET",
     757            exp.DataType.Type.VARIANT: "SQL_VARIANT",
     758        }
     759
     760        TYPE_MAPPING.pop(exp.DataType.Type.NCHAR)
     761        TYPE_MAPPING.pop(exp.DataType.Type.NVARCHAR)
     762
     763        TRANSFORMS = {
     764            **generator.Generator.TRANSFORMS,
     765            exp.AnyValue: any_value_to_max_sql,
     766            exp.ArrayToString: rename_func("STRING_AGG"),
     767            exp.AutoIncrementColumnConstraint: lambda *_: "IDENTITY",
     768            exp.DateAdd: date_delta_sql("DATEADD"),
     769            exp.DateDiff: date_delta_sql("DATEDIFF"),
     770            exp.CTE: transforms.preprocess([qualify_derived_table_outputs]),
     771            exp.CurrentDate: rename_func("GETDATE"),
     772            exp.CurrentTimestamp: rename_func("GETDATE"),
     773            exp.DateStrToDate: datestrtodate_sql,
     774            exp.Extract: rename_func("DATEPART"),
     775            exp.GeneratedAsIdentityColumnConstraint: generatedasidentitycolumnconstraint_sql,
     776            exp.GroupConcat: _string_agg_sql,
     777            exp.If: rename_func("IIF"),
     778            exp.JSONExtract: _json_extract_sql,
     779            exp.JSONExtractScalar: _json_extract_sql,
     780            exp.LastDay: lambda self, e: self.func("EOMONTH", e.this),
     781            exp.Max: max_or_greatest,
     782            exp.MD5: lambda self, e: self.func("HASHBYTES", exp.Literal.string("MD5"), e.this),
     783            exp.Min: min_or_least,
     784            exp.NumberToStr: _format_sql,
     785            exp.ParseJSON: lambda self, e: self.sql(e, "this"),
     786            exp.Select: transforms.preprocess(
     787                [
     788                    transforms.eliminate_distinct_on,
     789                    transforms.eliminate_semi_and_anti_joins,
     790                    transforms.eliminate_qualify,
     791                ]
     792            ),
     793            exp.StrPosition: lambda self, e: self.func(
     794                "CHARINDEX", e.args.get("substr"), e.this, e.args.get("position")
     795            ),
     796            exp.Subquery: transforms.preprocess([qualify_derived_table_outputs]),
     797            exp.SHA: lambda self, e: self.func("HASHBYTES", exp.Literal.string("SHA1"), e.this),
     798            exp.SHA2: lambda self, e: self.func(
     799                "HASHBYTES", exp.Literal.string(f"SHA2_{e.args.get('length', 256)}"), e.this
     800            ),
     801            exp.TemporaryProperty: lambda self, e: "",
     802            exp.TimeStrToTime: timestrtotime_sql,
     803            exp.TimeToStr: _format_sql,
     804            exp.Trim: trim_sql,
     805            exp.TsOrDsAdd: date_delta_sql("DATEADD", cast=True),
     806            exp.TsOrDsDiff: date_delta_sql("DATEDIFF"),
     807        }
     808
     809        TRANSFORMS.pop(exp.ReturnsProperty)
     810
     811        PROPERTIES_LOCATION = {
     812            **generator.Generator.PROPERTIES_LOCATION,
     813            exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
     814        }
     815
     816        def select_sql(self, expression: exp.Select) -> str:
     817            if expression.args.get("offset"):
     818                if not expression.args.get("order"):
     819                    # ORDER BY is required in order to use OFFSET in a query, so we use
     820                    # a noop order by, since we don't really care about the order.
     821                    # See: https://www.microsoftpressstore.com/articles/article.aspx?p=2314819
     822                    expression.order_by(exp.select(exp.null()).subquery(), copy=False)
     823
     824                limit = expression.args.get("limit")
     825                if isinstance(limit, exp.Limit):
     826                    # TOP and OFFSET can't be combined, we need use FETCH instead of TOP
     827                    # we replace here because otherwise TOP would be generated in select_sql
     828                    limit.replace(exp.Fetch(direction="FIRST", count=limit.expression))
     829
     830            return super().select_sql(expression)
     831
     832        def convert_sql(self, expression: exp.Convert) -> str:
     833            name = "TRY_CONVERT" if expression.args.get("safe") else "CONVERT"
     834            return self.func(
     835                name, expression.this, expression.expression, expression.args.get("style")
     836            )
     837
     838        def queryoption_sql(self, expression: exp.QueryOption) -> str:
     839            option = self.sql(expression, "this")
     840            value = self.sql(expression, "expression")
     841            if value:
     842                optional_equal_sign = "= " if option in OPTIONS_THAT_REQUIRE_EQUAL else ""
     843                return f"{option} {optional_equal_sign}{value}"
     844            return option
     845
     846        def lateral_op(self, expression: exp.Lateral) -> str:
     847            cross_apply = expression.args.get("cross_apply")
     848            if cross_apply is True:
     849                return "CROSS APPLY"
     850            if cross_apply is False:
     851                return "OUTER APPLY"
     852
     853            # TODO: perhaps we can check if the parent is a Join and transpile it appropriately
     854            self.unsupported("LATERAL clause is not supported.")
     855            return "LATERAL"
     856
     857        def timefromparts_sql(self, expression: exp.TimeFromParts) -> str:
     858            nano = expression.args.get("nano")
     859            if nano is not None:
     860                nano.pop()
     861                self.unsupported("Specifying nanoseconds is not supported in TIMEFROMPARTS.")
     862
     863            if expression.args.get("fractions") is None:
     864                expression.set("fractions", exp.Literal.number(0))
     865            if expression.args.get("precision") is None:
     866                expression.set("precision", exp.Literal.number(0))
     867
     868            return rename_func("TIMEFROMPARTS")(self, expression)
     869
     870        def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
     871            zone = expression.args.get("zone")
     872            if zone is not None:
     873                zone.pop()
     874                self.unsupported("Time zone is not supported in DATETIMEFROMPARTS.")
     875
     876            nano = expression.args.get("nano")
     877            if nano is not None:
     878                nano.pop()
     879                self.unsupported("Specifying nanoseconds is not supported in DATETIMEFROMPARTS.")
     880
     881            if expression.args.get("milli") is None:
     882                expression.set("milli", exp.Literal.number(0))
     883
     884            return rename_func("DATETIMEFROMPARTS")(self, expression)
     885
     886        def set_operations(self, expression: exp.Union) -> str:
     887            limit = expression.args.get("limit")
     888            if limit:
     889                return self.sql(expression.limit(limit.pop(), copy=False))
     890
     891            return super().set_operations(expression)
     892
     893        def setitem_sql(self, expression: exp.SetItem) -> str:
     894            this = expression.this
     895            if isinstance(this, exp.EQ) and not isinstance(this.left, exp.Parameter):
     896                # T-SQL does not use '=' in SET command, except when the LHS is a variable.
     897                return f"{self.sql(this.left)} {self.sql(this.right)}"
     898
     899            return super().setitem_sql(expression)
     900
     901        def boolean_sql(self, expression: exp.Boolean) -> str:
     902            if type(expression.parent) in BIT_TYPES:
     903                return "1" if expression.this else "0"
     904
     905            return "(1 = 1)" if expression.this else "(1 = 0)"
     906
     907        def is_sql(self, expression: exp.Is) -> str:
     908            if isinstance(expression.expression, exp.Boolean):
     909                return self.binary(expression, "=")
     910            return self.binary(expression, "IS")
     911
     912        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
     913            sql = self.sql(expression, "this")
     914            properties = expression.args.get("properties")
     915
     916            if sql[:1] != "#" and any(
     917                isinstance(prop, exp.TemporaryProperty)
     918                for prop in (properties.expressions if properties else [])
     919            ):
     920                sql = f"#{sql}"
     921
     922            return sql
     923
     924        def create_sql(self, expression: exp.Create) -> str:
     925            kind = expression.kind
     926            exists = expression.args.pop("exists", None)
     927            sql = super().create_sql(expression)
     928
     929            like_property = expression.find(exp.LikeProperty)
     930            if like_property:
     931                ctas_expression = like_property.this
     932            else:
     933                ctas_expression = expression.expression
     934
     935            table = expression.find(exp.Table)
     936
     937            # Convert CTAS statement to SELECT .. INTO ..
     938            if kind == "TABLE" and ctas_expression:
     939                ctas_with = ctas_expression.args.get("with")
     940                if ctas_with:
     941                    ctas_with = ctas_with.pop()
     942
     943                if isinstance(ctas_expression, exp.UNWRAPPED_QUERIES):
     944                    ctas_expression = ctas_expression.subquery()
     945
     946                select_into = exp.select("*").from_(exp.alias_(ctas_expression, "temp", table=True))
     947                select_into.set("into", exp.Into(this=table))
     948                select_into.set("with", ctas_with)
     949
     950                if like_property:
     951                    select_into.limit(0, copy=False)
     952
     953                sql = self.sql(select_into)
     954
     955            if exists:
     956                identifier = self.sql(exp.Literal.string(exp.table_name(table) if table else ""))
     957                sql = self.sql(exp.Literal.string(sql))
     958                if kind == "SCHEMA":
     959                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.schemata WHERE schema_name = {identifier}) EXEC({sql})"""
     960                elif kind == "TABLE":
     961                    assert table
     962                    where = exp.and_(
     963                        exp.column("table_name").eq(table.name),
     964                        exp.column("table_schema").eq(table.db) if table.db else None,
     965                        exp.column("table_catalog").eq(table.catalog) if table.catalog else None,
     966                    )
     967                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.tables WHERE {where}) EXEC({sql})"""
     968                elif kind == "INDEX":
     969                    index = self.sql(exp.Literal.string(expression.this.text("this")))
     970                    sql = f"""IF NOT EXISTS (SELECT * FROM sys.indexes WHERE object_id = object_id({identifier}) AND name = {index}) EXEC({sql})"""
     971            elif expression.args.get("replace"):
     972                sql = sql.replace("CREATE OR REPLACE ", "CREATE OR ALTER ", 1)
     973
     974            return self.prepend_ctes(expression, sql)
     975
     976        def offset_sql(self, expression: exp.Offset) -> str:
     977            return f"{super().offset_sql(expression)} ROWS"
     978
     979        def version_sql(self, expression: exp.Version) -> str:
     980            name = "SYSTEM_TIME" if expression.name == "TIMESTAMP" else expression.name
     981            this = f"FOR {name}"
     982            expr = expression.expression
     983            kind = expression.text("kind")
     984            if kind in ("FROM", "BETWEEN"):
     985                args = expr.expressions
     986                sep = "TO" if kind == "FROM" else "AND"
     987                expr_sql = f"{self.sql(seq_get(args, 0))} {sep} {self.sql(seq_get(args, 1))}"
     988            else:
     989                expr_sql = self.sql(expr)
     990
     991            expr_sql = f" {expr_sql}" if expr_sql else ""
     992            return f"{this} {kind}{expr_sql}"
     993
     994        def returnsproperty_sql(self, expression: exp.ReturnsProperty) -> str:
     995            table = expression.args.get("table")
     996            table = f"{table} " if table else ""
     997            return f"RETURNS {table}{self.sql(expression, 'this')}"
     998
     999        def returning_sql(self, expression: exp.Returning) -> str:
    1000            into = self.sql(expression, "into")
    1001            into = self.seg(f"INTO {into}") if into else ""
    1002            return f"{self.seg('OUTPUT')} {self.expressions(expression, flat=True)}{into}"
    1003
    1004        def transaction_sql(self, expression: exp.Transaction) -> str:
    1005            this = self.sql(expression, "this")
    1006            this = f" {this}" if this else ""
    1007            mark = self.sql(expression, "mark")
    1008            mark = f" WITH MARK {mark}" if mark else ""
    1009            return f"BEGIN TRANSACTION{this}{mark}"
    1010
    1011        def commit_sql(self, expression: exp.Commit) -> str:
    1012            this = self.sql(expression, "this")
    1013            this = f" {this}" if this else ""
    1014            durability = expression.args.get("durability")
    1015            durability = (
    1016                f" WITH (DELAYED_DURABILITY = {'ON' if durability else 'OFF'})"
    1017                if durability is not None
    1018                else ""
    1019            )
    1020            return f"COMMIT TRANSACTION{this}{durability}"
    1021
    1022        def rollback_sql(self, expression: exp.Rollback) -> str:
    1023            this = self.sql(expression, "this")
    1024            this = f" {this}" if this else ""
    1025            return f"ROLLBACK TRANSACTION{this}"
    1026
    1027        def identifier_sql(self, expression: exp.Identifier) -> str:
    1028            identifier = super().identifier_sql(expression)
    1029
    1030            if expression.args.get("global"):
    1031                identifier = f"##{identifier}"
    1032            elif expression.args.get("temporary"):
    1033                identifier = f"#{identifier}"
    1034
    1035            return identifier
    1036
    1037        def constraint_sql(self, expression: exp.Constraint) -> str:
    1038            this = self.sql(expression, "this")
    1039            expressions = self.expressions(expression, flat=True, sep=" ")
    1040            return f"CONSTRAINT {this} {expressions}"
    1041
    1042        def length_sql(self, expression: exp.Length) -> str:
    1043            return self._uncast_text(expression, "LEN")
    1044
    1045        def right_sql(self, expression: exp.Right) -> str:
    1046            return self._uncast_text(expression, "RIGHT")
    1047
    1048        def left_sql(self, expression: exp.Left) -> str:
    1049            return self._uncast_text(expression, "LEFT")
    1050
    1051        def _uncast_text(self, expression: exp.Expression, name: str) -> str:
    1052            this = expression.this
    1053            if isinstance(this, exp.Cast) and this.is_type(exp.DataType.Type.TEXT):
    1054                this_sql = self.sql(this, "this")
    1055            else:
    1056                this_sql = self.sql(this)
    1057            expression_sql = self.sql(expression, "expression")
    1058            return self.func(name, this_sql, expression_sql if expression_sql else None)
    1059
    1060        def partition_sql(self, expression: exp.Partition) -> str:
    1061            return f"WITH (PARTITIONS({self.expressions(expression, flat=True)}))"

Generator converts a given syntax tree to the corresponding SQL string.

###### Arguments:

- pretty: Whether to format the produced SQL string. Default: False.
- identify: Determines when an identifier should be quoted. Possible values are: False (default): Never quote, except in cases where it's mandatory by the dialect. True or 'always': Always quote. 'safe': Only quote identifiers that are case insensitive.
- normalize: Whether to normalize identifiers to lowercase. Default: False.
- pad: The pad size in a formatted string. For example, this affects the indentation of a projection in a query, relative to its nesting level. Default: 2.
- indent: The indentation size in a formatted string. For example, this affects the indentation of subqueries and filters under a `WHERE` clause. Default: 2.
- normalize_functions: How to normalize function names. Possible values are: "upper" or True (default): Convert names to uppercase. "lower": Convert names to lowercase. False: Disables function name normalization.
- unsupported_level: Determines the generator's behavior when it encounters unsupported expressions. Default ErrorLevel.WARN.
- max_unsupported: Maximum number of unsupported messages to include in a raised UnsupportedError. This is only relevant if unsupported_level is ErrorLevel.RAISE. Default: 3
- leading_comma: Whether the comma is leading or trailing in select expressions. This is only relevant when generating in pretty mode. Default: False
- max_text_width: The max number of characters in a segment before creating new lines in pretty mode. The default is on the smaller end because the length only represents a segment and not the true line length. Default: 80
- comments: Whether to preserve comments in the output SQL code. Default: True

LIMIT_IS_TOP = True
QUERY_HINTS = False
RETURNING_END = False
NVL2_SUPPORTED = False
ALTER_TABLE_INCLUDE_COLUMN_KEYWORD = False
LIMIT_FETCH = 'FETCH'
COMPUTED_COLUMN_WITH_TYPE = False
CTE_RECURSIVE_KEYWORD_REQUIRED = False
ENSURE_BOOLS = True
NULL_ORDERING_SUPPORTED = None
SUPPORTS_SINGLE_ARG_CONCAT = False
TABLESAMPLE_SEED_KEYWORD = 'REPEATABLE'
SUPPORTS_SELECT_INTO = True
JSON_PATH_BRACKETED_KEY_SUPPORTED = False
SUPPORTS_TO_NUMBER = False
EXPRESSIONS_WITHOUT_NESTED_CTES = {<class 'sqlglot.expressions.Subquery'>, <class 'sqlglot.expressions.Delete'>, <class 'sqlglot.expressions.Select'>, <class 'sqlglot.expressions.Update'>, <class 'sqlglot.expressions.Merge'>, <class 'sqlglot.expressions.Union'>, <class 'sqlglot.expressions.Insert'>}
SUPPORTED_JSON_PATH_PARTS = {<class 'sqlglot.expressions.JSONPathRoot'>, <class 'sqlglot.expressions.JSONPathSubscript'>, <class 'sqlglot.expressions.JSONPathKey'>}
TYPE_MAPPING = {<Type.MEDIUMTEXT: 'MEDIUMTEXT'>: 'TEXT', <Type.LONGTEXT: 'LONGTEXT'>: 'TEXT', <Type.TINYTEXT: 'TINYTEXT'>: 'TEXT', <Type.MEDIUMBLOB: 'MEDIUMBLOB'>: 'BLOB', <Type.LONGBLOB: 'LONGBLOB'>: 'BLOB', <Type.TINYBLOB: 'TINYBLOB'>: 'BLOB', <Type.INET: 'INET'>: 'INET', <Type.BOOLEAN: 'BOOLEAN'>: 'BIT', <Type.DECIMAL: 'DECIMAL'>: 'NUMERIC', <Type.DATETIME: 'DATETIME'>: 'DATETIME2', <Type.DOUBLE: 'DOUBLE'>: 'FLOAT', <Type.INT: 'INT'>: 'INTEGER', <Type.TEXT: 'TEXT'>: 'VARCHAR(MAX)', <Type.TIMESTAMP: 'TIMESTAMP'>: 'DATETIME2', <Type.TIMESTAMPTZ: 'TIMESTAMPTZ'>: 'DATETIMEOFFSET', <Type.VARIANT: 'VARIANT'>: 'SQL_VARIANT'}
TRANSFORMS = {<class 'sqlglot.expressions.JSONPathKey'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathRoot'>: <function <lambda>>, <class 'sqlglot.expressions.JSONPathSubscript'>: <function <lambda>>, <class 'sqlglot.expressions.AutoRefreshProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.BackupProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CaseSpecificColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CharacterSetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CollateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CommentColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DateAdd'>: <function date_delta_sql.<locals>._delta_sql>, <class 'sqlglot.expressions.DateFormatColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.DefaultColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.EncodeColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExcludeColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ExternalProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.GlobalProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.HeapProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.IcebergProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InheritsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InlineLengthColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.InputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.IntervalSpan'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.JSONExtract'>: <function _json_extract_sql>, <class 'sqlglot.expressions.JSONExtractScalar'>: <function _json_extract_sql>, <class 'sqlglot.expressions.LanguageProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LocationProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.LogProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.MaterializedProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NonClusteredColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.NotForReplicationColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnCommitProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OnUpdateColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.OutputModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.PathColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SampleProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SetConfigProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SetProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SettingsProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SharingProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SqlReadWriteProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.StabilityProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TemporaryProperty'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.TitleColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.Timestamp'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ToTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransformModelProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.TransientProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.UppercaseColumnConstraint'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.UnloggedProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.VarMap'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.ViewAttributeProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.VolatileProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.WithOperator'>: <function Generator.<lambda>>, <class 'sqlglot.expressions.AnyValue'>: <function any_value_to_max_sql>, <class 'sqlglot.expressions.ArrayToString'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.AutoIncrementColumnConstraint'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.DateDiff'>: <function date_delta_sql.<locals>._delta_sql>, <class 'sqlglot.expressions.CTE'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.CurrentDate'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.CurrentTimestamp'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.DateStrToDate'>: <function datestrtodate_sql>, <class 'sqlglot.expressions.Extract'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.GeneratedAsIdentityColumnConstraint'>: <function generatedasidentitycolumnconstraint_sql>, <class 'sqlglot.expressions.GroupConcat'>: <function _string_agg_sql>, <class 'sqlglot.expressions.If'>: <function rename_func.<locals>.<lambda>>, <class 'sqlglot.expressions.LastDay'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.Max'>: <function max_or_greatest>, <class 'sqlglot.expressions.MD5'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.Min'>: <function min_or_least>, <class 'sqlglot.expressions.NumberToStr'>: <function _format_sql>, <class 'sqlglot.expressions.ParseJSON'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.Select'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.StrPosition'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.Subquery'>: <function preprocess.<locals>._to_sql>, <class 'sqlglot.expressions.SHA'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.SHA2'>: <function TSQL.Generator.<lambda>>, <class 'sqlglot.expressions.TimeStrToTime'>: <function timestrtotime_sql>, <class 'sqlglot.expressions.TimeToStr'>: <function _format_sql>, <class 'sqlglot.expressions.Trim'>: <function trim_sql>, <class 'sqlglot.expressions.TsOrDsAdd'>: <function date_delta_sql.<locals>._delta_sql>, <class 'sqlglot.expressions.TsOrDsDiff'>: <function date_delta_sql.<locals>._delta_sql>}
PROPERTIES_LOCATION = {<class 'sqlglot.expressions.AlgorithmProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.AutoIncrementProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.AutoRefreshProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.BackupProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.BlockCompressionProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CharacterSetProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ChecksumProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.CollateProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.CopyGrantsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Cluster'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ClusteredByProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DataBlocksizeProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.DefinerProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.DictRange'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DictProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DistKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.DistStyleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.EngineProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExecuteAsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ExternalProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.FallbackProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.FileFormatProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.FreespaceProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.GlobalProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.HeapProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.InheritsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.IcebergProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.InputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.IsolatedLoadingProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.JournalProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.LanguageProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LikeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LocationProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LockProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.LockingProperty'>: <Location.POST_ALIAS: 'POST_ALIAS'>, <class 'sqlglot.expressions.LogProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.MaterializedProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.MergeBlockRatioProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.NoPrimaryIndexProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.OnProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OnCommitProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.Order'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.OutputModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PartitionedByProperty'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.PartitionedOfProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.PrimaryKey'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Property'>: <Location.POST_WITH: 'POST_WITH'>, <class 'sqlglot.expressions.RemoteWithConnectionModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.ReturnsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatDelimitedProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.RowFormatSerdeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SampleProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SchemaCommentProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SerdeProperties'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.Set'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SettingsProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SetProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.SetConfigProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SharingProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.SequenceProperties'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.SortKeyProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SqlReadWriteProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.SqlSecurityProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.StabilityProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TemporaryProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.ToTableProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.TransientProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.TransformModelProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.MergeTreeTTL'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.UnloggedProperty'>: <Location.POST_CREATE: 'POST_CREATE'>, <class 'sqlglot.expressions.ViewAttributeProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>, <class 'sqlglot.expressions.VolatileProperty'>: <Location.UNSUPPORTED: 'UNSUPPORTED'>, <class 'sqlglot.expressions.WithDataProperty'>: <Location.POST_EXPRESSION: 'POST_EXPRESSION'>, <class 'sqlglot.expressions.WithJournalTableProperty'>: <Location.POST_NAME: 'POST_NAME'>, <class 'sqlglot.expressions.WithSystemVersioningProperty'>: <Location.POST_SCHEMA: 'POST_SCHEMA'>}
def select_sql(self, expression: sqlglot.expressions.Select) -> str: View Source

    816        def select_sql(self, expression: exp.Select) -> str:
    817            if expression.args.get("offset"):
    818                if not expression.args.get("order"):
    819                    # ORDER BY is required in order to use OFFSET in a query, so we use
    820                    # a noop order by, since we don't really care about the order.
    821                    # See: https://www.microsoftpressstore.com/articles/article.aspx?p=2314819
    822                    expression.order_by(exp.select(exp.null()).subquery(), copy=False)
    823
    824                limit = expression.args.get("limit")
    825                if isinstance(limit, exp.Limit):
    826                    # TOP and OFFSET can't be combined, we need use FETCH instead of TOP
    827                    # we replace here because otherwise TOP would be generated in select_sql
    828                    limit.replace(exp.Fetch(direction="FIRST", count=limit.expression))
    829
    830            return super().select_sql(expression)

def convert_sql(self, expression: sqlglot.expressions.Convert) -> str: View Source

    832        def convert_sql(self, expression: exp.Convert) -> str:
    833            name = "TRY_CONVERT" if expression.args.get("safe") else "CONVERT"
    834            return self.func(
    835                name, expression.this, expression.expression, expression.args.get("style")
    836            )

def queryoption_sql(self, expression: sqlglot.expressions.QueryOption) -> str: View Source

    838        def queryoption_sql(self, expression: exp.QueryOption) -> str:
    839            option = self.sql(expression, "this")
    840            value = self.sql(expression, "expression")
    841            if value:
    842                optional_equal_sign = "= " if option in OPTIONS_THAT_REQUIRE_EQUAL else ""
    843                return f"{option} {optional_equal_sign}{value}"
    844            return option

def lateral_op(self, expression: sqlglot.expressions.Lateral) -> str: View Source

    846        def lateral_op(self, expression: exp.Lateral) -> str:
    847            cross_apply = expression.args.get("cross_apply")
    848            if cross_apply is True:
    849                return "CROSS APPLY"
    850            if cross_apply is False:
    851                return "OUTER APPLY"
    852
    853            # TODO: perhaps we can check if the parent is a Join and transpile it appropriately
    854            self.unsupported("LATERAL clause is not supported.")
    855            return "LATERAL"

def timefromparts_sql(self, expression: sqlglot.expressions.TimeFromParts) -> str: View Source

    857        def timefromparts_sql(self, expression: exp.TimeFromParts) -> str:
    858            nano = expression.args.get("nano")
    859            if nano is not None:
    860                nano.pop()
    861                self.unsupported("Specifying nanoseconds is not supported in TIMEFROMPARTS.")
    862
    863            if expression.args.get("fractions") is None:
    864                expression.set("fractions", exp.Literal.number(0))
    865            if expression.args.get("precision") is None:
    866                expression.set("precision", exp.Literal.number(0))
    867
    868            return rename_func("TIMEFROMPARTS")(self, expression)

def timestampfromparts_sql(self, expression: sqlglot.expressions.TimestampFromParts) -> str: View Source

    870        def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
    871            zone = expression.args.get("zone")
    872            if zone is not None:
    873                zone.pop()
    874                self.unsupported("Time zone is not supported in DATETIMEFROMPARTS.")
    875
    876            nano = expression.args.get("nano")
    877            if nano is not None:
    878                nano.pop()
    879                self.unsupported("Specifying nanoseconds is not supported in DATETIMEFROMPARTS.")
    880
    881            if expression.args.get("milli") is None:
    882                expression.set("milli", exp.Literal.number(0))
    883
    884            return rename_func("DATETIMEFROMPARTS")(self, expression)

def set_operations(self, expression: sqlglot.expressions.Union) -> str: View Source

    886        def set_operations(self, expression: exp.Union) -> str:
    887            limit = expression.args.get("limit")
    888            if limit:
    889                return self.sql(expression.limit(limit.pop(), copy=False))
    890
    891            return super().set_operations(expression)

def setitem_sql(self, expression: sqlglot.expressions.SetItem) -> str: View Source

    893        def setitem_sql(self, expression: exp.SetItem) -> str:
    894            this = expression.this
    895            if isinstance(this, exp.EQ) and not isinstance(this.left, exp.Parameter):
    896                # T-SQL does not use '=' in SET command, except when the LHS is a variable.
    897                return f"{self.sql(this.left)} {self.sql(this.right)}"
    898
    899            return super().setitem_sql(expression)

def boolean_sql(self, expression: sqlglot.expressions.Boolean) -> str: View Source

    901        def boolean_sql(self, expression: exp.Boolean) -> str:
    902            if type(expression.parent) in BIT_TYPES:
    903                return "1" if expression.this else "0"
    904
    905            return "(1 = 1)" if expression.this else "(1 = 0)"

def is_sql(self, expression: sqlglot.expressions.Is) -> str: View Source

    907        def is_sql(self, expression: exp.Is) -> str:
    908            if isinstance(expression.expression, exp.Boolean):
    909                return self.binary(expression, "=")
    910            return self.binary(expression, "IS")

def createable_sql( self, expression: sqlglot.expressions.Create, locations: DefaultDict) -> str: View Source

    912        def createable_sql(self, expression: exp.Create, locations: t.DefaultDict) -> str:
    913            sql = self.sql(expression, "this")
    914            properties = expression.args.get("properties")
    915
    916            if sql[:1] != "#" and any(
    917                isinstance(prop, exp.TemporaryProperty)
    918                for prop in (properties.expressions if properties else [])
    919            ):
    920                sql = f"#{sql}"
    921
    922            return sql

def create_sql(self, expression: sqlglot.expressions.Create) -> str: View Source

    924        def create_sql(self, expression: exp.Create) -> str:
    925            kind = expression.kind
    926            exists = expression.args.pop("exists", None)
    927            sql = super().create_sql(expression)
    928
    929            like_property = expression.find(exp.LikeProperty)
    930            if like_property:
    931                ctas_expression = like_property.this
    932            else:
    933                ctas_expression = expression.expression
    934
    935            table = expression.find(exp.Table)
    936
    937            # Convert CTAS statement to SELECT .. INTO ..
    938            if kind == "TABLE" and ctas_expression:
    939                ctas_with = ctas_expression.args.get("with")
    940                if ctas_with:
    941                    ctas_with = ctas_with.pop()
    942
    943                if isinstance(ctas_expression, exp.UNWRAPPED_QUERIES):
    944                    ctas_expression = ctas_expression.subquery()
    945
    946                select_into = exp.select("*").from_(exp.alias_(ctas_expression, "temp", table=True))
    947                select_into.set("into", exp.Into(this=table))
    948                select_into.set("with", ctas_with)
    949
    950                if like_property:
    951                    select_into.limit(0, copy=False)
    952
    953                sql = self.sql(select_into)
    954
    955            if exists:
    956                identifier = self.sql(exp.Literal.string(exp.table_name(table) if table else ""))
    957                sql = self.sql(exp.Literal.string(sql))
    958                if kind == "SCHEMA":
    959                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.schemata WHERE schema_name = {identifier}) EXEC({sql})"""
    960                elif kind == "TABLE":
    961                    assert table
    962                    where = exp.and_(
    963                        exp.column("table_name").eq(table.name),
    964                        exp.column("table_schema").eq(table.db) if table.db else None,
    965                        exp.column("table_catalog").eq(table.catalog) if table.catalog else None,
    966                    )
    967                    sql = f"""IF NOT EXISTS (SELECT * FROM information_schema.tables WHERE {where}) EXEC({sql})"""
    968                elif kind == "INDEX":
    969                    index = self.sql(exp.Literal.string(expression.this.text("this")))
    970                    sql = f"""IF NOT EXISTS (SELECT * FROM sys.indexes WHERE object_id = object_id({identifier}) AND name = {index}) EXEC({sql})"""
    971            elif expression.args.get("replace"):
    972                sql = sql.replace("CREATE OR REPLACE ", "CREATE OR ALTER ", 1)
    973
    974            return self.prepend_ctes(expression, sql)

def offset_sql(self, expression: sqlglot.expressions.Offset) -> str: View Source

    976        def offset_sql(self, expression: exp.Offset) -> str:
    977            return f"{super().offset_sql(expression)} ROWS"

def version_sql(self, expression: sqlglot.expressions.Version) -> str: View Source

    979        def version_sql(self, expression: exp.Version) -> str:
    980            name = "SYSTEM_TIME" if expression.name == "TIMESTAMP" else expression.name
    981            this = f"FOR {name}"
    982            expr = expression.expression
    983            kind = expression.text("kind")
    984            if kind in ("FROM", "BETWEEN"):
    985                args = expr.expressions
    986                sep = "TO" if kind == "FROM" else "AND"
    987                expr_sql = f"{self.sql(seq_get(args, 0))} {sep} {self.sql(seq_get(args, 1))}"
    988            else:
    989                expr_sql = self.sql(expr)
    990
    991            expr_sql = f" {expr_sql}" if expr_sql else ""
    992            return f"{this} {kind}{expr_sql}"

def returnsproperty_sql(self, expression: sqlglot.expressions.ReturnsProperty) -> str: View Source

    994        def returnsproperty_sql(self, expression: exp.ReturnsProperty) -> str:
    995            table = expression.args.get("table")
    996            table = f"{table} " if table else ""
    997            return f"RETURNS {table}{self.sql(expression, 'this')}"

def returning_sql(self, expression: sqlglot.expressions.Returning) -> str: View Source

     999        def returning_sql(self, expression: exp.Returning) -> str:
    1000            into = self.sql(expression, "into")
    1001            into = self.seg(f"INTO {into}") if into else ""
    1002            return f"{self.seg('OUTPUT')} {self.expressions(expression, flat=True)}{into}"

def transaction_sql(self, expression: sqlglot.expressions.Transaction) -> str: View Source

    1004        def transaction_sql(self, expression: exp.Transaction) -> str:
    1005            this = self.sql(expression, "this")
    1006            this = f" {this}" if this else ""
    1007            mark = self.sql(expression, "mark")
    1008            mark = f" WITH MARK {mark}" if mark else ""
    1009            return f"BEGIN TRANSACTION{this}{mark}"

def commit_sql(self, expression: sqlglot.expressions.Commit) -> str: View Source

    1011        def commit_sql(self, expression: exp.Commit) -> str:
    1012            this = self.sql(expression, "this")
    1013            this = f" {this}" if this else ""
    1014            durability = expression.args.get("durability")
    1015            durability = (
    1016                f" WITH (DELAYED_DURABILITY = {'ON' if durability else 'OFF'})"
    1017                if durability is not None
    1018                else ""
    1019            )
    1020            return f"COMMIT TRANSACTION{this}{durability}"

def rollback_sql(self, expression: sqlglot.expressions.Rollback) -> str: View Source

    1022        def rollback_sql(self, expression: exp.Rollback) -> str:
    1023            this = self.sql(expression, "this")
    1024            this = f" {this}" if this else ""
    1025            return f"ROLLBACK TRANSACTION{this}"

def identifier_sql(self, expression: sqlglot.expressions.Identifier) -> str: View Source

    1027        def identifier_sql(self, expression: exp.Identifier) -> str:
    1028            identifier = super().identifier_sql(expression)
    1029
    1030            if expression.args.get("global"):
    1031                identifier = f"##{identifier}"
    1032            elif expression.args.get("temporary"):
    1033                identifier = f"#{identifier}"
    1034
    1035            return identifier

def constraint_sql(self, expression: sqlglot.expressions.Constraint) -> str: View Source

    1037        def constraint_sql(self, expression: exp.Constraint) -> str:
    1038            this = self.sql(expression, "this")
    1039            expressions = self.expressions(expression, flat=True, sep=" ")
    1040            return f"CONSTRAINT {this} {expressions}"

def length_sql(self, expression: sqlglot.expressions.Length) -> str: View Source

    1042        def length_sql(self, expression: exp.Length) -> str:
    1043            return self._uncast_text(expression, "LEN")

def right_sql(self, expression: sqlglot.expressions.Right) -> str: View Source

    1045        def right_sql(self, expression: exp.Right) -> str:
    1046            return self._uncast_text(expression, "RIGHT")

def left_sql(self, expression: sqlglot.expressions.Left) -> str: View Source

    1048        def left_sql(self, expression: exp.Left) -> str:
    1049            return self._uncast_text(expression, "LEFT")

def partition_sql(self, expression: sqlglot.expressions.Partition) -> str: View Source

    1060        def partition_sql(self, expression: exp.Partition) -> str:
    1061            return f"WITH (PARTITIONS({self.expressions(expression, flat=True)}))"

SELECT_KINDS: Tuple[str, ...] = ()
AFTER_HAVING_MODIFIER_TRANSFORMS = {'qualify': <function Generator.<lambda>>, 'windows': <function Generator.<lambda>>}

##### Inherited Members

sqlglot.generator.Generator
  Generator
  IGNORE_NULLS_IN_FUNC
  LOCKING_READS_SUPPORTED
  EXPLICIT_UNION
  WRAP_DERIVED_VALUES
  CREATE_FUNCTION_RETURN_AS
  MATCHED_BY_SOURCE
  SINGLE_STRING_INTERVAL
  INTERVAL_ALLOWS_PLURAL_FORM
  LIMIT_ONLY_LITERALS
  RENAME_TABLE_WITH_DB
  GROUPINGS_SEP
  INDEX_ON
  JOIN_HINTS
  TABLE_HINTS
  QUERY_HINT_SEP
  IS_BOOL_ALLOWED
  DUPLICATE_KEY_UPDATE_WITH_SET
  COLUMN_JOIN_MARKS_SUPPORTED
  EXTRACT_ALLOWS_QUOTES
  TZ_TO_WITH_TIME_ZONE
  VALUES_AS_TABLE
  UNNEST_WITH_ORDINALITY
  AGGREGATE_FILTER_SUPPORTED
  SEMI_ANTI_JOIN_WITH_SIDE
  SUPPORTS_TABLE_COPY
  TABLESAMPLE_REQUIRES_PARENS
  TABLESAMPLE_SIZE_IS_ROWS
  TABLESAMPLE_KEYWORDS
  TABLESAMPLE_WITH_METHOD
  COLLATE_IS_FUNC
  DATA_TYPE_SPECIFIERS_ALLOWED
  LAST_DAY_SUPPORTS_DATE_PART
  SUPPORTS_TABLE_ALIAS_COLUMNS
  UNPIVOT_ALIASES_ARE_IDENTIFIERS
  JSON_KEY_VALUE_PAIR_SEP
  INSERT_OVERWRITE
  SUPPORTS_UNLOGGED_TABLES
  SUPPORTS_CREATE_TABLE_LIKE
  LIKE_PROPERTY_INSIDE_SCHEMA
  MULTI_ARG_DISTINCT
  JSON_TYPE_REQUIRED_FOR_EXTRACTION
  JSON_PATH_SINGLE_QUOTE_ESCAPE
  CAN_IMPLEMENT_ARRAY_ANY
  STAR_MAPPING
  TIME_PART_SINGULARS
  TOKEN_MAPPING
  STRUCT_DELIMITER
  PARAMETER_TOKEN
  NAMED_PLACEHOLDER_TOKEN
  RESERVED_KEYWORDS
  WITH_SEPARATED_COMMENTS
  EXCLUDE_COMMENTS
  UNWRAPPED_INTERVAL_VALUES
  PARAMETERIZABLE_TEXT_TYPES
  SENTINEL_LINE_BREAK
  pretty
  identify
  normalize
  pad
  unsupported_level
  max_unsupported
  leading_comma
  max_text_width
  comments
  dialect
  normalize_functions
  unsupported_messages
  generate
  preprocess
  unsupported
  sep
  seg
  pad_comment
  maybe_comment
  wrap
  no_identify
  normalize_func
  indent
  sql
  uncache_sql
  cache_sql
  characterset_sql
  column_sql
  columnposition_sql
  columndef_sql
  columnconstraint_sql
  computedcolumnconstraint_sql
  autoincrementcolumnconstraint_sql
  compresscolumnconstraint_sql
  generatedasidentitycolumnconstraint_sql
  generatedasrowcolumnconstraint_sql
  periodforsystemtimeconstraint_sql
  notnullcolumnconstraint_sql
  transformcolumnconstraint_sql
  primarykeycolumnconstraint_sql
  uniquecolumnconstraint_sql
  sequenceproperties_sql
  clone_sql
  describe_sql
  heredoc_sql
  prepend_ctes
  with_sql
  cte_sql
  tablealias_sql
  bitstring_sql
  hexstring_sql
  bytestring_sql
  unicodestring_sql
  rawstring_sql
  datatypeparam_sql
  datatype_sql
  directory_sql
  delete_sql
  drop_sql
  except_sql
  except_op
  fetch_sql
  filter_sql
  hint_sql
  indexparameters_sql
  index_sql
  inputoutputformat_sql
  national_sql
  properties_sql
  root_properties
  properties
  with_properties
  locate_properties
  property_name
  property_sql
  likeproperty_sql
  fallbackproperty_sql
  journalproperty_sql
  freespaceproperty_sql
  checksumproperty_sql
  mergeblockratioproperty_sql
  datablocksizeproperty_sql
  blockcompressionproperty_sql
  isolatedloadingproperty_sql
  partitionboundspec_sql
  partitionedofproperty_sql
  lockingproperty_sql
  withdataproperty_sql
  withsystemversioningproperty_sql
  insert_sql
  intersect_sql
  intersect_op
  introducer_sql
  kill_sql
  pseudotype_sql
  objectidentifier_sql
  onconflict_sql
  rowformatdelimitedproperty_sql
  withtablehint_sql
  indextablehint_sql
  historicaldata_sql
  table_parts
  table_sql
  tablesample_sql
  pivot_sql
  tuple_sql
  update_sql
  values_sql
  var_sql
  into_sql
  from_sql
  group_sql
  having_sql
  connect_sql
  prior_sql
  join_sql
  lambda_sql
  lateral_sql
  limit_sql
  set_sql
  pragma_sql
  lock_sql
  literal_sql
  escape_str
  loaddata_sql
  null_sql
  order_sql
  withfill_sql
  cluster_sql
  distribute_sql
  sort_sql
  ordered_sql
  matchrecognize_sql
  query_modifiers
  offset_limit_modifiers
  after_limit_modifiers
  schema_sql
  schema_columns_sql
  star_sql
  parameter_sql
  sessionparameter_sql
  placeholder_sql
  subquery_sql
  qualify_sql
  union_sql
  union_op
  unnest_sql
  prewhere_sql
  where_sql
  window_sql
  partition_by_sql
  windowspec_sql
  withingroup_sql
  between_sql
  bracket_sql
  all_sql
  any_sql
  exists_sql
  case_sql
  nextvaluefor_sql
  extract_sql
  trim_sql
  convert_concat_args
  concat_sql
  concatws_sql
  check_sql
  foreignkey_sql
  primarykey_sql
  if_sql
  matchagainst_sql
  jsonkeyvalue_sql
  jsonpath_sql
  json_path_part
  formatjson_sql
  jsonobject_sql
  jsonobjectagg_sql
  jsonarray_sql
  jsonarrayagg_sql
  jsoncolumndef_sql
  jsonschema_sql
  jsontable_sql
  openjsoncolumndef_sql
  openjson_sql
  in_sql
  in_unnest_op
  interval_sql
  return_sql
  reference_sql
  anonymous_sql
  paren_sql
  neg_sql
  not_sql
  alias_sql
  pivotalias_sql
  aliases_sql
  atindex_sql
  attimezone_sql
  fromtimezone_sql
  add_sql
  and_sql
  or_sql
  xor_sql
  connector_sql
  bitwiseand_sql
  bitwiseleftshift_sql
  bitwisenot_sql
  bitwiseor_sql
  bitwiserightshift_sql
  bitwisexor_sql
  cast_sql
  currentdate_sql
  currenttimestamp_sql
  collate_sql
  command_sql
  comment_sql
  mergetreettlaction_sql
  mergetreettl_sql
  altercolumn_sql
  renametable_sql
  renamecolumn_sql
  altertable_sql
  add_column_sql
  droppartition_sql
  addconstraint_sql
  distinct_sql
  ignorenulls_sql
  respectnulls_sql
  havingmax_sql
  intdiv_sql
  dpipe_sql
  div_sql
  overlaps_sql
  distance_sql
  dot_sql
  eq_sql
  propertyeq_sql
  escape_sql
  glob_sql
  gt_sql
  gte_sql
  ilike_sql
  ilikeany_sql
  like_sql
  likeany_sql
  similarto_sql
  lt_sql
  lte_sql
  mod_sql
  mul_sql
  neq_sql
  nullsafeeq_sql
  nullsafeneq_sql
  slice_sql
  sub_sql
  trycast_sql
  log_sql
  use_sql
  binary
  function_fallback_sql
  func
  format_args
  text_width
  format_time
  expressions
  op_expressions
  naked_property
  tag_sql
  token_sql
  userdefinedfunction_sql
  joinhint_sql
  kwarg_sql
  when_sql
  merge_sql
  tochar_sql
  tonumber_sql
  dictproperty_sql
  dictrange_sql
  dictsubproperty_sql
  oncluster_sql
  clusteredbyproperty_sql
  anyvalue_sql
  querytransform_sql
  indexconstraintoption_sql
  checkcolumnconstraint_sql
  indexcolumnconstraint_sql
  nvl2_sql
  comprehension_sql
  columnprefix_sql
  opclass_sql
  predict_sql
  forin_sql
  refresh_sql
  operator_sql
  toarray_sql
  tsordstotime_sql
  tsordstodate_sql
  unixdate_sql
  lastday_sql
  arrayany_sql
  generateseries_sql
  struct_sql
  partitionrange_sql
  truncatetable_sql